package.preload['minidriver'] = (function (...)
--Copyright 2018 Control4 Corporation. All rights reserved.

--[[
	In general, you should merge
]]

function ON_DRIVER_INIT.minidriver_init()
	PASSTHROUGH_PROXY = NET_PROXY_BINDINGID		-- set this to the proxy ID that should handle all passthrough commands from minidrivers
	SWITCHER_PROXY = 5001					-- set this to the proxy ID of the SET_INPUT capable device that has the RF_MINI_APP connections (may be the same as PASSTHROUGH_PROXY)
	USES_DEDICATED_SWITCHER = false			-- set this to false if the driver did not need the dedicated avswitch proxy (e.g. this is a TV/receiver)
	MINIAPP_BINDING_START = 3051				-- set this to the first binding ID in the XML for the RF_MINI_APP connections
	MINIAPP_BINDING_END = 3075				-- set this to the last binding ID in the XML for the RF_MINI_APP connections
	MINIAPP_TYPE = 'UM_INTEGRA'				-- set this to your unique name as defined in the minidriver SERVICE_IDS table
	
	C4:RegisterSystemEvent(C4SystemEvents["OnPIP"], 0)
end

function ON_DRIVER_LATEINIT.minidriver_init()
	if (USES_DEDICATED_SWITCHER) then
		HideProxyInAllRooms (SWITCHER_PROXY)
	end
	RegisterRooms ()
end

function OnSystemEvent (event)
	local eventname = string.match (event, '.-name="(.-)"')
	if (eventname == 'OnPIP') then
		RegisterRooms()
		GetMACAddress()
		GetIPAddress()
	end
end

function OnWatchedVariableChanged (idDevice, idVariable, strValue)
	if (RoomIDs and RoomIDs [idDevice]) then
		local roomId = tonumber (idDevice)
		if (idVariable == 1000) then
			local deviceId = tonumber (strValue) or 0
			RoomIDSources [roomId] = deviceId
		end
	end
end

--[[
function ReceivedFromProxy (idBinding, strCommand, tParams)
	if (idBinding == nil or strCommand == nil) then
		return	-- this shouldn't happen, but is a good defensive programming practice
	end

	if (tParams == nil) then
		tParams = {}	-- this often happens, especially with basic transport commands.  This protects against nil table referencing.
	end

	if (idBinding == SWITCHER_PROXY and strCommand == 'PASSTHROUGH') then
		idBinding = PASSTHROUGH_PROXY
		strCommand = tParams.PASSTHROUGH_COMMAND
	end

	if (idBinding == SWITCHER_PROXY) then
		if (strCommand == 'SET_INPUT') then
			local input = tonumber (tParams.INPUT)
			if (input >= MINIAPP_BINDING_START and input <= MINIAPP_BINDING_END) then
				-- Get the device ID of the proxy handling the miniapp switch on this driver
				local proxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

				-- Get the device ID of the minidriver proxy connected to the requested input on this driver
				local appProxyId = C4:GetBoundProviderDevice (proxyDeviceId, input)

				-- Get the device ID of the minidriver protocol connected to the minidriver proxy
				local appDeviceId = C4:GetBoundProviderDevice (appProxyId, 5001)

				-- get the details for the app for this kind of universal-minidriver-compatible type
				local appId = GetRelevantUniversalAppId (appDeviceId, MINIAPP_TYPE)
				local appName = GetRelevantUniversalAppId (appDeviceId, 'APP_NAME')

				-- there is now enough information to launch the application using the protocol for this device

				-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
				-- =-=-=-=-=-=-=-= YOUR DEVICE-SPECIFIC APP LAUNCHING CODE GOES HERE... -=-=-=-=-=-=-=
				-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

				-- there is now enough information to launch the application using the protocol for this device

				if ((Properties ['Passthrough Mode'] or 'On') ~= 'On') then
					local passthroughProxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), PASSTHROUGH_PROXY))
					local _timer = function (timer)
						print ('Looking for ' .. appProxyId)
						for roomId, deviceId in pairs (RoomIDSources) do
							if (deviceId == appProxyId) then
								C4:SendToDevice (roomId, 'SELECT_VIDEO_DEVICE', {deviceid = passthroughProxyDeviceId})
							end
						end
					end
					C4:SetTimer (500, _timer)
				end
			end
		end
	end
end
--]]

function GetRelevantUniversalAppId (deviceId, source)
	local vars = C4:GetDeviceVariables (deviceId)
	for _, var in pairs (vars) do
		if (var.name == source) then
			return (var.value)
		end
	end
	if (source ~= 'APP_ID') then
		-- try getting pre-universal minidriver app ID to launch.
		return (GetRelevantUniversalAppId (deviceId, 'APP_ID'))
	end
end

function HideProxyInAllRooms (idBinding)
	idBinding = idBinding or 0
	if (idBinding == 0) then return end -- silently fail if no binding passed in

	-- Get Bound Proxy's Device ID / Name.
	local id, name = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

	dbg ('Hiding ' .. name .. ' in all rooms')

	-- Send hide command to all rooms, for 'ALL' Navigator groups.
	for roomId, roomName in pairs (C4:GetDevicesByC4iName ('roomdevice.c4i') or {}) do
		dbg ('Hiding ' .. name .. ' in ' .. roomName)
		C4:SendToDevice (roomId, 'SET_DEVICE_HIDDEN_STATE', {PROXY_GROUP = 'ALL', DEVICE_ID = id, IS_HIDDEN = true})
	end
end

function RegisterRooms ()
	RoomIDs = C4:GetDevicesByC4iName ('roomdevice.c4i')
	RoomIDSources = {}
	for roomId, _ in pairs (RoomIDs) do
		RoomIDSources [roomId] = tonumber (C4:GetDeviceVariable (roomId, 1000)) or 0
		C4:UnregisterVariableListener (roomId, 1000)
		C4:RegisterVariableListener (roomId, 1000)
	end
end
 end)
package.preload['receiver.receiver_proxy_notifies'] = (function (...)
--[[=============================================================================
    Received Proxy Notification Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--[[
	Notify: INPUT_OUTPUT_CHANGED
	Sent to the proxy to indicate the currently selected Input has changed on the specified Output
	Parameters
		bindingID - proxy id of proxy bound to input connection
		input - mod 1000 value of Input Connection ID
		output - mod 1000 value of Output Connection ID
--]]
function NOTIFY.INPUT_OUTPUT_CHANGED(bindingID, input, output)
	--NOTE: Need to handle cases where 7000 range output values are sent (when Room Binding is next in the path)
	--so out "output" is a mod 1000 integer. Also, need to handle audio and video legs of the input...
    local tParams = {INPUT = input + 3000, OUTPUT = 4000 + output}
    SendNotify("INPUT_OUTPUT_CHANGED", tParams, bindingID)
    tParams = {INPUT = input + 1000, OUTPUT = 2000 + output}
    SendNotify("INPUT_OUTPUT_CHANGED", tParams, bindingID)	
end
------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
	SendSimpleNotify("ON")
end

function NOTIFY.OFF()
	SendSimpleNotify("OFF")
end
------------------------------ Volume Notification Code ------------------------------
--[[
	Notify: VOLUME_LEVEL_CHANGED
	Sent to the Receiver proxy to indicate that the Volume level has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		level - C4 level uses a percentage scale: 0 - 100
--]]
function NOTIFY.VOLUME_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("VOLUME_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BASS_LEVEL_CHANGED
	Sent to the Receiver proxy to indicate that the Bass level has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BASS_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BASS_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: TREBLE_LEVEL_CHANGED
	Sent to the Receiver proxy to indicate that the Treble level has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.TREBLE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("TREBLE_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BALANCE_LEVEL_CHANGED
	Sent to the Receiver proxy to indicate that the Balance level has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BALANCE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BALANCE_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: LOUNDENSS_CHANGED
	Sent to the Receiver proxy to indicate that the Loudness state has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.LOUDNESS_CHANGED(bindingID, output, state)
    local tParams = {LOUDNESS = state, OUTPUT = output + 4000}
    SendNotify("LOUDNESS_CHANGED", tParams, bindingID)		
end

--[[
	Notify: MUTE_LEVEL_CHANGED
	Sent to the Receiver proxy to indicate that the Mute state has changed on the specified Output
	Parameters
		bindingID - proxy id of Receiver proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.MUTE_CHANGED(bindingID, output, state)
    local tParams = {MUTE = state, OUTPUT = output + 4000}
    SendNotify("MUTE_CHANGED", tParams, bindingID)		
end

--[[
	Notify: SURROUND_MODE_CHANGED
	Sent to the Receiver proxy to indicate the current Surround Mode has changed.
	Parameters
		bindingID - proxy id of Tuner proxy
		mode_id - ID of Surround Mode (as declared in Capabilities XML)
--]]
function NOTIFY.SURROUND(bindingID, output, mode_id)
    local tParams = {SURROUND_MODE = mode_id, OUTPUT = output + 4000}
    SendNotify("SURROUND_MODE_CHANGED", tParams, bindingID)	    
end

--[[
	Notify: CHANNEL_CHANGED & INPUT_CHANGED
	Sent to the Tuner proxy to indicate the current station frequency has changed. The INPUT_CHANGED must also be sent even if it has not changed.
	Parameters
		bindingID - proxy id of Tuner proxy
		input - mod 1000 value of Input Connection ID (this is the Antenna input)
		channel - station frequency
--]]
function NOTIFY.TUNER(bindingID, input, channel)
    local tParams
    if (gReceiverProxy._TunerBand == "AMBand") then
	   tParams = {INPUT=input,BANDTYPE=gReceiverProxy._TunerBand,MINCHANNEL=tFreqStepInfo[gFreqStep].AmMin,MAXCHANNEL=tFreqStepInfo[gFreqStep].AmMax,CHANNELSPACING=tFreqStepInfo[gFreqStep].AmSpacing}
    else
	   tParams = {INPUT=input,BANDTYPE=gReceiverProxy._TunerBand,MINCHANNEL=tFreqStepInfo[gFreqStep].FmMin,MAXCHANNEL=tFreqStepInfo[gFreqStep].FmMax,CHANNELSPACING=tFreqStepInfo[gFreqStep].FmSpacing}
    end 
    SendNotify("INPUT_CHANGED", tParams, bindingID)	 
    
    tParams = {CHANNEL = channel}
    SendNotify("CHANNEL_CHANGED", tParams, bindingID)	 
end	 end)
package.preload['receiver.receiver_proxy_commands'] = (function (...)
--[[=============================================================================
    ReceivedFromProxy Code for the Receiver/Tuner Proxies

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.ON(idBinding, tParams)
	--Handled by CONNECT_OUTPUT
	--gReceiverProxy:prx_ON(tParams)
end

function PRX_CMD.OFF(idBinding, tParams)
	--Handled by DISCONNECT_OUTPUT
	--gReceiverProxy:prx_OFF(tParams)
end

function PRX_CMD.CONNECT_OUTPUT(idBinding, tParams)
	gReceiverProxy:prx_CONNECT_OUTPUT(tParams)
end

function PRX_CMD.DISCONNECT_OUTPUT(idBinding, tParams)
	gReceiverProxy:prx_DISCONNECT_OUTPUT(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gReceiverProxy:prx_SET_INPUT(idBinding, tParams)
end

function PRX_CMD.INPUT_TOGGLE(idBinding, tParams)
	gReceiverProxy:prx_INPUT_TOGGLE(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	gReceiverProxy:prx_MUTE_OFF(tParams)
end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	gReceiverProxy:prx_MUTE_ON(tParams)	
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	gReceiverProxy:prx_MUTE_TOGGLE(tParams)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	gReceiverProxy:prx_SET_VOLUME_LEVEL(tParams)
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	gReceiverProxy:prx_PULSE_VOL_DOWN(tParams)
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	gReceiverProxy:prx_PULSE_VOL_UP(tParams)
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	gReceiverProxy:prx_START_VOL_DOWN(tParams)
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	gReceiverProxy:prx_START_VOL_UP(tParams)
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	gReceiverProxy:prx_STOP_VOL_DOWN(tParams)
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	gReceiverProxy:prx_STOP_VOL_UP(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.INFO(idBinding, tParams)
	gReceiverProxy:prx_INFO(idBinding, tParams)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	gReceiverProxy:prx_GUIDE(idBinding, tParams)
end

function PRX_CMD.MENU(idBinding, tParams)
	gReceiverProxy:prx_MENU(idBinding, tParams)
end

function PRX_CMD.CANCEL(idBinding, tParams)
	gReceiverProxy:prx_CANCEL(idBinding, tParams)
end

function PRX_CMD.UP(idBinding, tParams)
	gReceiverProxy:prx_UP(idBinding, tParams)
end

function PRX_CMD.DOWN(idBinding, tParams)
	gReceiverProxy:prx_DOWN(idBinding, tParams)
end

function PRX_CMD.LEFT(idBinding, tParams)
	gReceiverProxy:prx_LEFT(idBinding, tParams)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	gReceiverProxy:prx_RIGHT(idBinding, tParams)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	gReceiverProxy:prx_START_DOWN(idBinding, tParams)
end

function PRX_CMD.START_UP(idBinding, tParams)
	gReceiverProxy:prx_START_UP(idBinding, tParams)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	gReceiverProxy:prx_START_LEFT(idBinding, tParams)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	gReceiverProxy:prx_START_RIGHT(idBinding, tParams)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	gReceiverProxy:prx_STOP_DOWN(idBinding, tParams)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	gReceiverProxy:prx_STOP_UP(idBinding, tParams)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	gReceiverProxy:prx_STOP_LEFT(idBinding, tParams)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	gReceiverProxy:prx_STOP_RIGHT(idBinding, tParams)
end

function PRX_CMD.ENTER(idBinding, tParams)
	gReceiverProxy:prx_ENTER(idBinding, tParams)
end

function PRX_CMD.RECALL(idBinding, tParams)
	gReceiverProxy:prx_RECALL(idBinding, tParams)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
	gReceiverProxy:prx_OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	gReceiverProxy:prx_PROGRAM_A(idBinding, tParams)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	gReceiverProxy:prx_PROGRAM_B(idBinding, tParams)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	gReceiverProxy:prx_PROGRAM_C(idBinding, tParams)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	gReceiverProxy:prx_PROGRAM_D(idBinding, tParams)
end

function PRX_CMD.PAGE_UP(idBinding, tParams)
	gReceiverProxy:prx_PAGE_UP(idBinding, tParams)
end

function PRX_CMD.PAGE_DOWN(idBinding, tParams)
	gReceiverProxy:prx_PAGE_DOWN(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.NUMBER_0(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_0(idBinding, tParams)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_1(idBinding, tParams)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_2(idBinding, tParams)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_3(idBinding, tParams)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_4(idBinding, tParams)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_5(idBinding, tParams)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_6(idBinding, tParams)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_7(idBinding, tParams)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_8(idBinding, tParams)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	gReceiverProxy:prx_NUMBER_9(idBinding, tParams)
end

function PRX_CMD.STAR(idBinding, tParams)
	gReceiverProxy:prx_STAR(idBinding, tParams)
end

function PRX_CMD.POUND(idBinding, tParams)
	gReceiverProxy:prx_POUND(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Surround Mode Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_SURROUND_MODE(idBinding, tParams)
	gReceiverProxy:prx_SET_SURROUND_MODE(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.PULSE_CH_UP(idBinding, tParams)
	gReceiverProxy:prx_PULSE_CH_UP(idBinding, tParams)
end

function PRX_CMD.START_CH_UP(idBinding, tParams)
	gReceiverProxy:prx_START_CH_UP(idBinding, tParams)
end

function PRX_CMD.PULSE_CH_DOWN(idBinding, tParams)
	gReceiverProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
end

function PRX_CMD.START_CH_DOWN(idBinding, tParams)
	gReceiverProxy:prx_START_CH_DOWN(idBinding, tParams)
end

function PRX_CMD.PRESET_UP(idBinding, tParams)
	gReceiverProxy:prx_PRESET_UP(idBinding, tParams)
end

function PRX_CMD.PRESET_DOWN(idBinding, tParams)
	gReceiverProxy:prx_PRESET_DOWN(idBinding, tParams)
end

function PRX_CMD.SET_CHANNEL(idBinding, tParams)
	gReceiverProxy:prx_SET_CHANNEL(idBinding, tParams)
end

function PRX_CMD.SET_PRESET(idBinding, tParams)
	gReceiverProxy:prx_SET_PRESET(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Transport Navigation Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.PLAY(idBinding, tParams)
	gReceiverProxy:prx_PLAY(idBinding, tParams)
end
	
function PRX_CMD.PAUSE(idBinding, tParams)
	gReceiverProxy:prx_PAUSE(idBinding, tParams)
end

function PRX_CMD.STOP(idBinding, tParams)
	gReceiverProxy:prx_STOP(idBinding, tParams)
end

function PRX_CMD.SKIP_REV(idBinding, tParams)
	gReceiverProxy:prx_SKIP_REV(idBinding, tParams)
end

function PRX_CMD.SKIP_FWD(idBinding, tParams)
	gReceiverProxy:prx_SKIP_FWD(idBinding, tParams)
end

function PRX_CMD.SCAN_REV(idBinding, tParams)
	gReceiverProxy:prx_SCAN_REV(idBinding, tParams)
end

function PRX_CMD.SCAN_FWD(idBinding, tParams)
	gReceiverProxy:prx_SCAN_FWD(idBinding, tParams)
end
 end)
package.preload['receiver.receiver_proxy_class'] = (function (...)
--[[=============================================================================
    Receiver Proxy Class Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

ReceiverProxy = inheritsFrom(nil)

function ReceiverProxy:construct(receiverBindingID, tunerBindingID, bProcessesDeviceMessages, tVolumeRampingTracking, bUsePulseCommandsForVolumeRamping)
	-- member variables
	self._ReceiverBindingID = receiverBindingID
	self._TunerBindingID = tunerBindingID
	self._PowerState = {}						--Valid Values: "ON", "OFF", "POWER_ON_SEQUENCE", "POWER_OFF_SEQUENCE"
	self._TunerBand = ""
	self._VolumeIsRamping = false
	self._VolumeRamping = tVolumeRampingTracking		--[0] = {state = false,mode = "",} ||	"state" is boolean, "mode" values: "VOLUME_UP" & "VOLUME_DOWN"
	self._UsePulseCommandsForVolumeRamping = bUsePulseCommandsForVolumeRamping or false
	self._LastVolumeStatusValue = {}	
	self._MenuNavigationInProgress = false
     self._MenuNavigationMode = ""
	self._MenuNavigationProxyID = ""
	self._MenuNavigationOutput = ""	
	self._CurrentlySelectedInput = {}
	self._ProcessesDeviceMessages = bProcessesDeviceMessages
	self._ControlMethod = ""						--Valid Values: "NETWORK", "SERIAL", "IR" 
end

------------------------------------------------------------------------
-- Receiver Proxy Commands (PRX_CMD)
------------------------------------------------------------------------
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_ON(tParams)
	--Handled by CONNECT_OUTPUT
	--ON()
end

function ReceiverProxy:prx_OFF(tParams)
	--Handled by DISCONNECT_OUTPUT
	--OFF()
end

function ReceiverProxy:prx_CONNECT_OUTPUT(tParams)
	local output = tonumber(tParams.OUTPUT) % 1000 
	if (self._PowerState[output] ~= nil) then
		if (self._PowerState[output] == 'ON') or (self._PowerState[output] == 'POWER_ON_SEQUENCE') then 
			LogInfo("Power State is '" .. self._PowerState[output] .. "', exiting CONNECT_OUTPUT")
			return 
		end	
	end	
	if (self._ProcessesDeviceMessages == false) then
		self._PowerState[output] = 'ON'
	else	
		self._PowerState[output] = 'POWER_ON_SEQUENCE'
	end	
	CONNECT_OUTPUT(output)
end

function ReceiverProxy:prx_DISCONNECT_OUTPUT(tParams)
    local output = tonumber(tParams.OUTPUT) % 1000
    if (self._ProcessesDeviceMessages == false) then
	   self._PowerState[output] = 'OFF'
    else	
	   self._PowerState[output] = 'POWER_OFF_SEQUENCE'
    end	
    DISCONNECT_OUTPUT(output)
  
    self._CurrentlySelectedInput[output] = -1
    C4:SendToProxy(self._ReceiverBindingID, 'INPUT_OUTPUT_CHANGED', {INPUT = -1, OUTPUT = 4000 + output})
    C4:SendToProxy(self._ReceiverBindingID, 'INPUT_OUTPUT_CHANGED', {INPUT = -1, OUTPUT = 2000 + output})
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_SET_INPUT(idBinding, tParams)
	--SET_INPUT called from Tune Up/Down from Navigator does not have INPUT parameter
	if (tParams.INPUT==nil) then
		local id =  tInputConnMapByName[tParams.BANDTYPE .. " Antenna"].ID 
		tParams.INPUT = 3000 + id
	end	
	if (tParams.OUTPUT==nil) then
		tParams.OUTPUT = 0	--Control Control Antenna input selection does not return OUTPUT paramter
	end		
	local input = tonumber(tParams["INPUT"] % 1000)
	local output = tonumber(tParams["OUTPUT"] % 1000)
	if (idBinding == self._ReceiverBindingID) then
		self._CurrentlySelectedInput[output] = input
		if (gControlMethod == "IR") then			
			NOTIFY.INPUT_OUTPUT_CHANGED(self._ReceiverBindingID, input, output)				
		end 		
	elseif (idBinding == self._TunerBindingID) then
		if (tInputConnMapByID[input].Name == 'FM Antenna') then 
			self._TunerBand = 'FMBand' 
		elseif (tInputConnMapByID[input].Name == 'AM Antenna') then 	
			self._TunerBand = 'AMBand'
		else
			LogWarn(tInputConnMapByID[input].Name .. " is not a valid Antenna Input for the AM/FM Tuner")
			return
		end	
	end
	SET_INPUT(idBinding, output, input)
end

function ReceiverProxy:prx_INPUT_TOGGLE(idBinding, tParams)
    local output = tonumber(tParams["OUTPUT"] % 1000)
    INPUT_TOGGLE(idBinding, output)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_MUTE_OFF(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) or DEFAULT_OUTPUT_ID
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ReceiverBindingID, output, "False")
	end 		
	MUTE_OFF(output)
end

function ReceiverProxy:prx_MUTE_ON(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) or DEFAULT_OUTPUT_ID
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ReceiverBindingID, output, "True")
	end 		
	MUTE_ON(output)
end

function ReceiverProxy:prx_MUTE_TOGGLE(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) 
	MUTE_TOGGLE(output)
end

function ReceiverProxy:prx_SET_VOLUME_LEVEL(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) 
	local c4VolumeLevel = tonumber(tParams['LEVEL'])
	SET_VOLUME_LEVEL(output, c4VolumeLevel)
end

function ReceiverProxy:prx_PULSE_VOL_DOWN(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) 
	PULSE_VOL_DOWN(output)
end

function ReceiverProxy:prx_PULSE_VOL_UP(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) 
	PULSE_VOL_UP(output)
end

function ReceiverProxy:prx_START_VOL_DOWN(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000) 
	self:ChangeVolume(output, "START_VOL_DOWN")
end

function ReceiverProxy:prx_START_VOL_UP(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000)
	self:ChangeVolume(output, "START_VOL_UP")
end

function ReceiverProxy:prx_STOP_VOL_DOWN(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000)
	self:ChangeVolume(output, "STOP_VOL_DOWN")
end

function ReceiverProxy:prx_STOP_VOL_UP(tParams)
	local output = tonumber(tParams["OUTPUT"] % 1000)
	self:ChangeVolume(output, "STOP_VOL_UP")
end

---------------------- Volume Helper Functions ----------------------
function ReceiverProxy:ChangeVolume(output, command_name)
	if (command_name == "STOP_VOL_UP") or (command_name == "STOP_VOL_DOWN") then
		self._VolumeIsRamping = false
		self._VolumeRamping[output].state = false
		self._VolumeRamping[output].mode = ""
	elseif (command_name == "START_VOL_UP") then 
		self._VolumeIsRamping = true
		self._VolumeRamping[output].state = true
		self._VolumeRamping[output].mode = "VOLUME_UP" 
		PULSE_VOL_UP(output)	
	elseif (command_name == "START_VOL_DOWN") then 	
		self._VolumeIsRamping = true
		self._VolumeRamping[output].state = true
		self._VolumeRamping[output].mode = "VOLUME_DOWN"	
		PULSE_VOL_DOWN(output)		
	else
		LogWarn(command_name .. " not handled in ChangeVolume()")
	end
end

function ReceiverProxy:ContinueVolumeRamping(output)
	local command
	if (gControlMethod == "IR") then   
		if (self._VolumeRamping[output].mode == "VOLUME_UP") then
			PULSE_VOL_UP(output)	
		elseif (self._VolumeRamping[output].mode == "VOLUME_DOWN") then
			PULSE_VOL_DOWN(output)	
		else
			LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		end	
	else
	     if (self._UsePulseCommandsForVolumeRamping) then
		    if (self._VolumeRamping[output].mode == "VOLUME_UP") then
			 PULSE_VOL_UP(output)	
		    elseif (self._VolumeRamping[output].mode == "VOLUME_DOWN") then
			 PULSE_VOL_DOWN(output)	
		    else
			 LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		    end	
		else
		    local volume = self._LastVolumeStatusValue[output]
		    local deviceVolumeLevel = self:GetNextVolumeCurveValue(output, volume)
		    if (deviceVolumeLevel ~= nil) then
			    self._LastVolumeStatusValue[output] = deviceVolumeLevel
			    SET_VOLUME_LEVEL_DEVICE(output, deviceVolumeLevel)                                 
		    else
			    LogWarn("ContinueVolumeRamping() next value is nil")
			    return
		    end
	     end
	end 
end

function ReceiverProxy:GetNextVolumeCurveValue(output, volume)
	local i, point
	volume=tonumber(volume)
	if (self._VolumeRamping[output].mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (self._VolumeRamping[output].mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		LogWarn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ReceiverProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    local curveV = {}
    curveV.__index = curveV

    function curveV:new(min, max, base)
	 local len = max-min
	 local logBase = math.log(base)
	 local baseM1 = base-1

	 local instance = {
	   min = min,
	   max = max,
	   len = len,
	   base = base,
	   logBase = logBase,
	   baseM1 = baseM1,
	   toNormal = function(x)
		return (x-min)/len
	   end,
	   fromNormal = function(x)
		return min+x*len
	   end,
	   value = function(x)
		return math.log(x*baseM1+1)/logBase
	   end,
	   invert = function(x)
		return (math.exp(x*logBase)-1)/baseM1
	   end
	 }
	 return setmetatable(instance, self)
    end

    function curveV:list(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))

	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	 end
    end
    
    function curveV:create_curve(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))
	 
      local tCurve = {}
	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	   --print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	    
	    local x = math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    
	    if (has_value(tCurve, x) == false) then
		  table.insert(tCurve, x)
		  print(i, x)
	    end
	 end
	 
	 return tCurve
    end
    
    

    -- min, max, base of log (must be > 1), try some for the best results
    local a = curveV:new(minDeviceLevel, maxDeviceLevel, slope)

    -- from, to, steps
    --a:list(minDeviceLevel, maxDeviceLevel, steps)
    
    local t = a:create_curve(minDeviceLevel, maxDeviceLevel, steps)
    
    return t

end

function has_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return true
        end
    end

    return false
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	  local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --LogInfo("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --LogInfo("vl = " .. vl)
	  vl= tonumber_loc(("%.".."0".."f"):format(vl))
	  --LogInfo("vl new = " .. vl)
	  LogInfo("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_PULSE_CH_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PULSE_CH_UP")
end

function ReceiverProxy:prx_START_CH_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	--Proxy does not support channel ramping so we send a pulse command...
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PULSE_CH_UP")
end

function ReceiverProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PULSE_CH_DOWN")
end

function ReceiverProxy:prx_START_CH_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	--Proxy does not support channel ramping so we send a pulse command...
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PULSE_CH_DOWN")
end

function ReceiverProxy:prx_PRESET_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PRESET_UP")
end

function ReceiverProxy:prx_PRESET_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PRESET_DOWN")
end

function ReceiverProxy:prx_SET_CHANNEL(idBinding, tParams)
    local sMajor, sMinor, channel, command
    channel= tParams['CHANNEL']

    local output
    if (tParams.OUTPUT ~= nil) then
		output = tParams.OUTPUT % 1000
    else
		--Composer UI (Control Control) does not return an OUTPUT parameter
		output = 0 
    end

    local band
    if tParams['BANDTYPE'] == nil then
	     --This case is to handle SET_CHANNEL intitiated from Composer UI (Control Control)
		if tonumber(tParams["CHANNEL"]) < tFreqStepInfo[gFreqStep].AmMin then	
			band = "FM"
		else
			band = "AM"
		end
    else
	   band = tParams["BANDTYPE"]
    end   
	
	SET_CHANNEL(idBinding, output, band, channel)
	
end

function ReceiverProxy:prx_SET_PRESET(idBinding, tParams)
    if (idBinding ~= self._TunerBindingID) then return end
	
    --tuner proxy is zero based so add one to value returned in PRESET parameter
    local preset = tParams['PRESET'] + 1
	
    local output
    if (tParams.OUTPUT ~= nil) then
		output = tParams.OUTPUT % 1000
    else
		--Composer UI (Control Control) does not return an OUTPUT parameter
		output = 0 
    end
	
    SET_PRESET(idBinding, output, preset)
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Surround Mode Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_SET_SURROUND_MODE(tParams)
  local mode = tonumber(tParams.SURROUND_MODE)
  SET_SURROUND_MODE(mode)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_INFO(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "INFO")
end

function ReceiverProxy:prx_GUIDE(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "GUIDE")
end

function ReceiverProxy:prx_MENU(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "MENU")
end

function ReceiverProxy:prx_CANCEL(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "CANCEL")
end

function ReceiverProxy:prx_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "UP")
end

function ReceiverProxy:prx_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "DOWN")
end

function ReceiverProxy:prx_LEFT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "LEFT")
end

function ReceiverProxy:prx_RIGHT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "RIGHT")
end

function ReceiverProxy:prx_START_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "START_DOWN")
end

function ReceiverProxy:prx_START_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output,  "START_UP")
end

function ReceiverProxy:prx_START_LEFT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "START_LEFT")
end

function ReceiverProxy:prx_START_RIGHT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "START_RIGHT")
end

function ReceiverProxy:prx_STOP_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "STOP_DOWN")
end

function ReceiverProxy:prx_STOP_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "STOP_UP")
end

function ReceiverProxy:prx_STOP_LEFT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "STOP_LEFT")
end

function ReceiverProxy:prx_STOP_RIGHT(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	self:NavigateMenu(idBinding, output, "STOP_RIGHT")
end

function ReceiverProxy:prx_ENTER(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "ENTER")
end

function ReceiverProxy:prx_RECALL(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "RECALL")
end

function ReceiverProxy:prx_OPEN_CLOSE(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "OPEN_CLOSE")
end

function ReceiverProxy:prx_PROGRAM_A(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PROGRAM_A")
end

function ReceiverProxy:prx_PROGRAM_B(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PROGRAM_B")
end

function ReceiverProxy:prx_PROGRAM_C(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PROGRAM_C")
end

function ReceiverProxy:prx_PROGRAM_D(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PROGRAM_D")
end

function ReceiverProxy:prx_PAGE_UP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PAGE_UP")
end

function ReceiverProxy:prx_PAGE_DOWN(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PAGE_DOWN")
end

---------------------- Menu Navigation Helper Functions ----------------------
function ReceiverProxy:NavigateMenu(idBinding, output, command_name)
	if (command_name == "STOP_UP") 
				or (command_name == "STOP_DOWN") 
				or (command_name == "STOP_LEFT") 
				or (command_name == "STOP_RIGHT") then
		self._MenuNavigationInProgress = false
		self._MenuNavigationMode = ""
		self._MenuNavigationProxyID = ""
		return
	elseif (command_name == "START_UP") then 
		self._MenuNavigationMode = "UP"
	elseif (command_name == "START_DOWN") then 	
		self._MenuNavigationMode = "DOWN"	
	elseif (command_name == "START_LEFT") then 
		self._MenuNavigationMode = "LEFT"	
     elseif (command_name == "START_RIGHT") then 
		self._MenuNavigationMode = "RIGHT"
	else
		LogWarn(command_name .. " not handled in NavigateMenu()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, self._MenuNavigationMode)
	self._MenuNavigationInProgress = true
	self._MenuNavigationProxyID = idBinding
	self._MenuNavigationOutput = output
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_NUMBER_0(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_0")
end

function ReceiverProxy:prx_NUMBER_1(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_1")
end

function ReceiverProxy:prx_NUMBER_2(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_2")
end

function ReceiverProxy:prx_NUMBER_3(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_3")
end

function ReceiverProxy:prx_NUMBER_4(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_4")
end

function ReceiverProxy:prx_NUMBER_5(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_5")
end

function ReceiverProxy:prx_NUMBER_6(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_6")
end

function ReceiverProxy:prx_NUMBER_7(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_7")
end

function ReceiverProxy:prx_NUMBER_8(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_8")
end

function ReceiverProxy:prx_NUMBER_9(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_9")
end

function ReceiverProxy:prx_STAR(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "STAR")
end

function ReceiverProxy:prx_POUND(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "POUND")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Transport Navigation Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ReceiverProxy:prx_PLAY(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PLAY")
end
	
function ReceiverProxy:prx_PAUSE(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "PAUSE")
end

function ReceiverProxy:prx_STOP(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "STOP")
end

function ReceiverProxy:prx_SKIP_REV(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "SKIP_REV")
end

function ReceiverProxy:prx_SKIP_FWD(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "SKIP_FWD")
end

function ReceiverProxy:prx_SCAN_REV(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "SCAN_REV")
end

function ReceiverProxy:prx_SCAN_FWD(idBinding, tParams)
	local output = get_output_with_nil_test(tParams["OUTPUT"])
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "SCAN_FWD")
end


function get_output_with_nil_test(output)
	local id = nil
	if (output ~= nil) then
		id = tonumber(output % 1000)
	end
	return id
end

------------------------------------------------------------------------
-- Receiver Proxy Notifies
------------------------------------------------------------------------

function ReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
	NOTIFY.INPUT_OUTPUT_CHANGED(bindingID, input, output)
end

function ReceiverProxy:dev_PowerOn(output)
	self._PowerState[output] = "ON"
	NOTIFY.ON()	
end

function ReceiverProxy:dev_PowerOff(output)
	self._PowerState[output] = "OFF"
	NOTIFY.OFF()
end

function ReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	NOTIFY.VOLUME_LEVEL_CHANGED(self._ReceiverBindingID, output, c4Level)	
	
	if (self._VolumeIsRamping) then
		--do nothing
		--during volume ramping, LastVolumeStatusValue is set in ContinueVolumeRamping()
	else
		self._LastVolumeStatusValue[output] = deviceLevel
	end	
end

function ReceiverProxy:dev_MuteChanged(output, state)
	NOTIFY.MUTE_CHANGED(self._ReceiverBindingID, output, state)
end		

function ReceiverProxy:dev_BassLevelChanged(output, level)
	NOTIFY.BASS_LEVEL_CHANGED(self._ReceiverBindingID, output, level)
end	

function ReceiverProxy:dev_TrebleLevelChanged(output, level)
	NOTIFY.TREBLE_LEVEL_CHANGED(self._ReceiverBindingID, output, level)
end	

function ReceiverProxy:dev_BalanceLevelChanged(output, level)
	NOTIFY.BALANCE_LEVEL_CHANGED(self._ReceiverBindingID, output, level)
end	

function ReceiverProxy:dev_LoudnessChanged(output, state)
	NOTIFY.LOUDNESS_CHANGED(self._ReceiverBindingID, output, state)
end

function ReceiverProxy:dev_SurroundModeChanged(output, mode_id)
	NOTIFY.SURROUND(self._ReceiverBindingID, output, mode_id)
end

------------------------------------------------------------------------
-- Tuner Proxy Notifies
------------------------------------------------------------------------
function ReceiverProxy:dev_TunerStatus(input, channel)
     if (self._TunerBand == "") then
	   local id = input % 1000
	   if (tInputConnMapByID[id] == "AM Antenna") then
		  self._TunerBand = "AMBand"
	   else
		  self._TunerBand = "FMBand"
	   end
	end
	NOTIFY.TUNER(self._TunerBindingID, input, channel)
end end)
package.preload['proxy_commands'] = (function (...)
-- This macro is utilized to identify the version string of the driver template version used.
TEMPLATE_VERSION.driver = "2014.08.14"

--[[
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    Commands received from the receiver proxy (ReceivedFromProxy)
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--]]

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: CONNECT_OUTPUT
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function CONNECT_OUTPUT(output)
    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
    local command
    if (gControlMethod == "IR") then
	   command = CMDS_IR_ZONES[output]["CONNECT_OUTPUT"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("CONNECT_OUTPUT", command, command_delay, delay_units)
    else
	   command = tPowerCommandMap[output] .. "01"
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("CONNECT_OUTPUT", command, command_delay, delay_units)
		
	   GetDeviceVolumeStatus(output)
    end
end

--[[
	Proxy Command: DISCONNECT_OUTPUT
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function DISCONNECT_OUTPUT(output)
    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    local command 
    --gPowerOffDelayTimer:StartTimer(tonumber(Properties["Power Off Delay Seconds"]))
    if (gControlMethod == "IR") then
	   command = CMDS_IR_ZONES[output]["DISCONNECT_OUTPUT"]
    else
	   command = tPowerCommandMap[output] .. '00' 
    end  
    LogTrace("command = " .. command)
    PackAndQueueCommand("DISCONNECT_OUTPUT", command, command_delay, delay_units)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: SET_INPUT
	Parameters:
		idBinding: proxybindingid of proxy bound to input connection
		output: mod 1000 value of Output Connection id	
		input: mod 1000 value of Input Connection id
--]]
function SET_INPUT(idBinding, output, input)
	local command
	
	if (idBinding == TUNER_PROXY_BINDINGID) then
	   if (tInputConnMapByID[input].Name == "FM Antenna") then 
		  gReceiverProxy._TunerBand = "FMBand" 
	   elseif (tInputConnMapByID[input].Name == "AM Antenna") then 	
		  gReceiverProxy._TunerBand = "AMBand"
	   else
		  LOG:Warn(tInputConnMapByID[input].Name .. " is not a valid Antenna Input for the AM/FM Tuner")
		  return
	   end	
		
	   if (gControlMethod == "IR") then			
			command = tInputCommandMap_IR[output][tInputConnMapByID[input].Name]
	   else
			command =  tOutputCommandMap[output] .. tInputCommandMap[tInputConnMapByID[input].Name]
	   end
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("SET_INPUT", command)
	   
	   return
	end

	if (input >= MINIAPP_BINDING_START % 1000 and input <= MINIAPP_BINDING_END % 1000) then
		-- Get the device ID of the proxy handling the miniapp switch on this driver
		local proxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

		-- Get the device ID of the minidriver proxy connected to the requested input on this driver
		local appProxyId = C4:GetBoundProviderDevice (proxyDeviceId, input + 3000)

		-- Get the device ID of the minidriver protocol connected to the minidriver proxy
		local appDeviceId = C4:GetBoundProviderDevice (appProxyId, 5001)

		-- get the details for the app for this kind of universal-minidriver-compatible type
		local appId = GetRelevantUniversalAppId (appDeviceId, MINIAPP_TYPE)
		local appName = GetRelevantUniversalAppId (appDeviceId, 'APP_NAME')

		-- there is now enough information to launch the application using the protocol for this device

		-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
		-- =-=-=-=-=-=-=-= YOUR DEVICE-SPECIFIC APP LAUNCHING CODE GOES HERE... -=-=-=-=-=-=-=
		-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
		

		if (gControlMethod == "IR") then
		    command = tInputCommandMap_IR[output]["NET"]	
		    PackAndQueueCommand("SET_INPUT", command)
		    LogInfo("Discrete selection of Network Services not supported via IR control.")
		    
	     elseif (gControlMethod == "Serial") then
		
		    --handle DAB which has discrete selection code
		    if (appName == "DAB") then
			   command =  tOutputCommandMap[output] .. appId
		    else
		        command =  tOutputCommandMap[output] .. tInputCommandMap["NET"]
			   LogInfo("Discrete selection of Network Services not supported via Serial control.")
		    end
		    
		    PackAndQueueCommand("SET_INPUT", command)
		    
		    gLastSelectedNetService[output] = input 
		    
		elseif (gControlMethod == "Network") then
		
		    --handle DAB which has discrete selection code
		    if (appName == "DAB") then 
			   command =  tOutputCommandMap[output] .. appId
			   PackAndQueueCommand("SET_INPUT", command)
		  
		    else
			   --increasing delay, occasionally the NET Service Selection below was failing with the standard delay
			   local command_delay = tonumber(Properties["Command Delay Milliseconds"]) * 2	
			   local delay_units = "MILLISECONDS"
			   command =  tOutputCommandMap[output] .. "2B"
			   PackAndQueueCommand("SET_INPUT", command, command_delay, delay_units)
		    
			   --check Airplay version...
			   if (appName == 'Airplay') then 
				local version = GetAirplayVersion()
				if (version == "Airplay2") then appId = "NSV440" end
			   end
			   
			   command_delay = 1500
			   PackAndQueueCommand("SET_INPUT-Net Service Selection", appId, command_delay, delay_units)	
		    end
		    
		    gLastSelectedNetService[output] = input
		
		else
		
		  LogWarn("gControlMethod not handled: " .. tostring(gControlMethod))
		  
		end		

		if ((Properties ['Passthrough Mode'] or 'On') ~= 'On') then
			local passthroughProxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), PASSTHROUGH_PROXY))
			local _timer = function (timer)
				print ('Looking for ' .. appProxyId)
				for roomId, deviceId in pairs (RoomIDSources) do
					if (deviceId == appProxyId) then
						C4:SendToDevice (roomId, 'SELECT_VIDEO_DEVICE', {deviceid = passthroughProxyDeviceId})
					end
				end
			end
			C4:SetTimer (500, _timer)
		end
		
	     --update IO mapping tables
	     tInputResponseMap[appId] = appName
	     tInputConnMapByID[input] = {Name = appName,BindingID = 5001,}
	     tInputConnMapByName[appName] = {ID = input,BindingID = 5001,}

	else
	
		--TUNER Selection command acts as a Band toggle if TUNER is already selected
		--So we will use the discreet Band selection command for the currently selected Band
		if (tInputConnMapByID[input].Name == "AM/FM TUNER") then
			if (gReceiverProxy._TunerBand == "AMBand") then
				input = tInputConnMapByName["AM Antenna"].ID
			else
				--default to FM if gTunerBand is not set
				input = tInputConnMapByName["FM Antenna"].ID
				gReceiverProxy._TunerBand = "FMBand"
			end
		end		
		
		if (gControlMethod == "IR") then	
			if (DRIVER_NAME == "Pioneer SX-N30") then
				--Pioneer SX-N30 does not support discrete band selection or USB selection
				if (input == tInputConnMapByName["AM Antenna"].ID or input == tInputConnMapByName["FM Antenna"].ID) then
					input = tInputConnMapByName["AM/FM TUNER"].ID
				elseif (input == tInputConnMapByName["USB (FRONT)"].ID) then
					input = tInputConnMapByName["NET"].ID
				end
			end	
			if (DRIVER_NAME == "Pioneer VSX-1131" or DRIVER_NAME == "Pioneer VSX-LX301") then
				if (output == 1) then
					--These models do not support discrete band selection in zone 2...
					if (input == tInputConnMapByName["AM Antenna"].ID or input == tInputConnMapByName["FM Antenna"].ID) then
						input = tInputConnMapByName["AM/FM TUNER"].ID
					end			 
				end
			end

			command = tInputCommandMap_IR[output][tInputConnMapByID[input].Name]	
		
		else
		
			command =  tOutputCommandMap[output] .. tInputCommandMap[tInputConnMapByID[input].Name]
			
		end
		
		LogTrace("command = " .. command)
		PackAndQueueCommand("SET_INPUT", command)	
		
		gLastSelectedNetService[output] = input
		
	end

end

function INPUT_TOGGLE(idBinding, output)
	if (idBinding == TUNER_PROXY_BINDINGID) then
	   local connection_name = ""
	   if (gTunerBand == 'FMBand') then
		 connection_name = "AM Antenna"
		 gTunerBand = 'AMBand'
	   else 
		 connection_name = "FM Antenna"
		 gTunerBand = 'FMBand'
	   end

	   if (gControlMethod == "IR") then			
		  command = tInputCommandMap_IR[output][connection_name]
	   else
		  command =  tOutputCommandMap[output] .. tInputCommandMap[connection_name]
	   end	
	   
	   PackAndQueueCommand("INPUT_TOGGLE", command)
    
	end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: MUTE_OFF
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_OFF(output)
	local command
	if (gControlMethod == "IR") then		
		command = CMDS_IR_ZONES[output]["MUTE_OFF"]	
	else
		command = tMuteCommandMap[output] .. '00'
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_OFF", command)
end

--[[
	Proxy Command: MUTE_ON
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_ON(output)
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR_ZONES[output]["MUTE_ON"]	
	else
		command = tMuteCommandMap[output] .. '01'
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_ON", command)	
end

--[[
	Proxy Command: MUTE_TOGGLE
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_TOGGLE(output)
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR_ZONES[output]["MUTE_TOGGLE"]	
	else
		command = tMuteCommandMap[output] .. 'TG'
	end 
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_TOGGLE", command)
end

--[[
	Proxy Command: SET_VOLUME_LEVEL
	Parameters:
		output: mod 1000 value of Output Connection id	
		c4VolumeLevel: volume level to be set represented in Control4 scale (0-100)
--]]
function SET_VOLUME_LEVEL(output, c4VolumeLevel)
    if (gControlMethod == "IR") then return end	--discrete volume not supported with IR 
    
    local minDeviceLevel = 0
    local maxDeviceLevel = MAX_DEVICE_LEVEL
	
    local deviceVolumeLevel
    if (MAX_DEVICE_LEVEL ~= 100) then
		deviceVolumeLevel = ConvertVolumeToDevice(c4VolumeLevel,  minDeviceLevel, maxDeviceLevel)
    else
	   deviceVolumeLevel = c4VolumeLevel
    end

    local command = tVolumeSetCommandMap[output] .. string.format('%02X', deviceVolumeLevel)
    LogTrace("command = " .. command)
    PackAndQueueCommand("SET_VOLUME_LEVEL", command)
	
end

--[[
	Helper Function: SET_VOLUME_LEVEL_DEVICE
	Parameters:
		output: mod 1000 value of Output Connection id	
		deviceVolumeLevel: volume level to be set represented in device scale (as sepcified in the device's control protocol)
--]]
function SET_VOLUME_LEVEL_DEVICE(output, deviceVolumeLevel)
	--Called from ContinueVolumeRamping()
	
	-- TODO: create packet/command to send to the device
	local command = tVolumeSetCommandMap[output] .. string.format('%02X', deviceVolumeLevel) 
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("SET_VOLUME_LEVEL_DEVICE", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_DOWN
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_DOWN(output)
	local command
	local command_delay
	if (gControlMethod == "IR") then		
		command = CMDS_IR_ZONES[output]["VOLUME_DOWN"]	
		command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"]) * 2
	else
		command = tVolumeCommandMap[output] .. "DOWN"
		command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	end 		
	LogTrace("command = " .. command)
	PackAndQueueCommand("PULSE_VOL_DOWN", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_UP
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_UP(output)
	local command
	local command_delay
	if (gControlMethod == "IR") then	
		command = CMDS_IR_ZONES[output]["VOLUME_UP"]	
		command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"]) * 2
	else
		command = tVolumeCommandMap[output] .. 'UP'
		command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	end 		
	LogTrace("command = " .. command)
	PackAndQueueCommand("PULSE_VOL_UP", command, command_delay)
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

function SET_CHANNEL(idBinding, output, band, channel)
	if (band == "FM") then
		_, _, sMajor =string.find(channel,'(.-)%D')
		_, _, sMinor =string.find(channel,'%D(.*)')
		if (sMajor == nil) or (sMinor == nil) then return end
		--INT'L TUNING: depending on tFreqStepInfo[gFreqStep].FmSpacing, 'frequency' may need to be adjusted for proper command syntax
		channel = string.format('%03d', sMajor) .. sMinor .. string.rep("0", 2 - string.len(sMinor))
	elseif (band == "AM") then
	     --INT'L TUNING: depending on tFreqStepInfo[gFreqStep].AmSpacing, 'frequency' may need to be adjusted for proper command syntax
		channel = string.rep("0", 5 - string.len(channel)) .. channel
	end 

	if (gControlMethod == "IR") then	
		-- TODO: create packet/command to send to the device (if a Direct Tuner command is needed, map it to a remote button like "STAR")
	     if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
		  SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "STAR")
	     else	--"Pioneer"
		  SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "DIRECT_ACCESS")
	     end		
		
		for i=1,string.len(channel) do
			SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_" .. string.sub(channel,i,i))
		end
	else
		-- TODO: create packet/command to send to the device
		local command = tTunerTuningCommandMap[output] .. channel
		LogTrace("command = " .. command)
		PackAndQueueCommand("SET_CHANNEL", command)	
	end
end

function SET_PRESET(idBinding, output, preset)
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end    
    if (gControlMethod == "IR") then	
	   for i=1,string.len(preset) do
		  --GetDigitCommand(idBinding, tParams, "NUMBER_" .. string.sub(preset,i,i))
		  SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, "NUMBER_" .. string.sub(preset,i,i))
	   end
    else
	   local numPreset = string.format('%02X', preset)
	   local command = tTunerPresetCommandMap[output] .. numPreset
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("SET_PRESET", command)			
    end	
		  
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Surround Mode Function
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function SET_SURROUND_MODE(mode)
     if (gControlMethod == "IR") then return end	--discrete surround selection not supported with IR 
	
	local command = tSurroundModes[mode]
	LogTrace("command = " .. command)
	PackAndQueueCommand("SET_SURROUND_MODE", command, 1, "SECONDS")	
	
	local id = string.sub(command,4)
	local bToggleCommand = string.find(tSurroundResponses[id].Name, "Toggle") or false
	if bToggleCommand then
	   PackAndQueueCommand("GetCurrentListeningMode", "LMDQSTN")
	end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	Helper Function: SEND_COMMAND_FROM_COMMAND_TABLE
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be sent
--]]
function SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, output, command_name)
    local output_for_log = output or "nil"
    LogTrace("SEND_COMMAND_FROM_COMMAND_TABLE(), idBinding=" .. idBinding .. ", output=" .. output_for_log .. ", command_name=" .. command_name)
    local command
    if (command_name == "PULSE_CH_UP") or 
	   (command_name == "START_CH_UP") or 
	   (command_name == "PULSE_CH_DOWN") or 
	   (command_name == "START_CH_DOWN") or 
	   (command_name == "DIRECT_ACCESS") then
		  command = GetTuningCommand(idBinding, output, command_name)
    elseif (command_name == "PRESET_UP") or 
	      (command_name == "PRESET_DOWN") or
	      (command_name == "PAGE_UP") or
	   (command_name == "PAGE_DOWN") then	
		  command = GetPresetTuningCommand(idBinding, output, command_name)	 
    elseif (command_name == "NUMBER_0") or 
	   (command_name == "NUMBER_1") or 
	   (command_name == "NUMBER_2") or 
	   (command_name == "NUMBER_3") or 
	   (command_name == "NUMBER_4") or 
	   (command_name == "NUMBER_5") or 
	   (command_name == "NUMBER_6") or 
	   (command_name == "NUMBER_7") or 
	   (command_name == "NUMBER_8") or 
	   (command_name == "NUMBER_9") or 
	   (command_name == "POUND") or 
	   (command_name == "STAR") then
		  command = GetDigitCommand(idBinding, output, command_name)
    elseif (command_name == "INFO") or
	   (command_name == "GUIDE") or
	   (command_name == "MENU") or
	   (command_name == "CANCEL") or
	   (command_name == "UP") or
	   (command_name == "DOWN") or
	   (command_name == "LEFT") or
	   (command_name == "RIGHT") or
	   (command_name == "ENTER") or
	   (command_name == "PLAY") or
	   (command_name == "PAUSE") or
	   (command_name == "STOP") or
	   (command_name == "SKIP_REV") or
	   (command_name == "SKIP_FWD") or
	   (command_name == "SCAN_REV") or
	   (command_name == "SCAN_FWD") then
	   if (idBinding == RECEIVER_PROXY_BINDINGID) then
		 command = GetCommandFromCommandTable(idBinding, output, command_name)
	   else
		 command = GetNetTuneCommand(idBinding, output, command_name)
	   end 	 
    else
	   command = GetCommandFromCommandTable(idBinding, output, command_name)
    end
	
	if (command == nil) then
		LogTrace("command is nil")
	else
		LogTrace("command = " .. command)
		PackAndQueueCommand(command_name, command)
	end	
end

--[[
	Helper Function: GetCommandFromCommandTable
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be returned
--]]
function GetCommandFromCommandTable(idBinding, output, command_name)
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end
    LogTrace("GetCommandFromCommandTable(), idBinding=" .. idBinding .. ", output=" .. output .. ", command_name=" .. command_name)
    
    local t = {}
    local command
    if (gControlMethod == "IR") then
	   t = CMDS_IR_ZONES
	   if (output == nil) then
		  output = findOutputFromBindingID(idBinding)
	   end	
	   if (t[output][command_name] ~= nil) then
		  command = t[output][command_name]
	   elseif (t[command_name] ~= nil) then
		  command = t[command_name]
	   else
		  LOG:Warn('GetCommand(): Command not defined - '.. command_name)
	   end			
    else
	   t = CMDS

	   if (t[idBinding][command_name] ~= nil) then
		  command = t[idBinding][command_name]
	   elseif (t[command_name] ~= nil) then
		  command = t[command_name]
	   else
		  LOG:Warn('GetCommand(): Command not defined - '.. command_name)
	   end		
    end	
    return command
	
end

--[[
	Helper Function: findOutputFromBindingID
	Parameters:
		idBinding: proxy id	
	Called from GetCommandFromCommandTable() when the deviceprotocol has multi-zone 
	command structrues for commands which Control4 does not include an OUTPUT paramter
--]]
function findOutputFromBindingID(idBinding)
	LogTrace('findOutputFromBindingID: ' .. idBinding)
	local name = tProxyInputNameMapByProxyID[idBinding] 
	local output = findOutputIDFromInputName(name)
	return output or 0
end

--[[
	Helper Function: findOutputIDFromInputName
	Parameters:
		inputName: input connection  name (this will be the input for the AV proxy connection)	
	Called from findOutputFromBindingID() when the device protocol has multi-zone 
	command structrues for commands which Control4 does not include an OUTPUT paramter
--]]
function findOutputIDFromInputName(inputName)
    LogTrace('findOutputIDFromInputName: ' .. tostring(inputName))
    local output
    for j,k in pairs(gReceiverProxy._CurrentlySelectedInput) do
		if (tInputConnMapByID[k] ~= nil) then
			if (inputName==tInputConnMapByID[k].Name) then
				if (j==0) then 
					return j 
				else
					output = j
				end
			end	
		end
	end
    return output or 0
end

function GetNetTuneCommand(idBinding, output, command_name)
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end
    LogTrace("GetNetTuneCommand(), idBinding=" .. idBinding .. ", output=" .. output .. ", command_name=" .. command_name)
    
    local command
    if (idBinding == TUNER_PROXY_BINDINGID) then 
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[command_name]
		  end		  
	   else
		  if (command_name == "SKIP_FWD") or (command_name == "SKIP_REV") then
			 command = tTunerTuningCommandMap[output] .. CMDS[idBinding][command_name]
		  end	
	   end	
    else
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[command_name]
		  end			  
	   else
		  command = tNetTuneCommandMap[output] .. CMDS[idBinding][command_name]
	   end
    end
	
    if (command == nil) then
	   LOG:Warn('GetNetTuneCommand(): Command not defined - '.. command_name)
    end	
    return command
	
end

function GetDigitCommand(idBinding, output, command_name)
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end	
    LogTrace("GetDigitCommand(), idBinding=" .. idBinding .. ", output=" .. output .. ", command_name=" .. command_name)
    
    local command
    if (idBinding == RECEIVER_PROXY_BINDINGID) then 
	   --digits not applicable with receiver proxy
    elseif (idBinding == TUNER_PROXY_BINDINGID) then 
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[idBinding][command_name]
		  end		  
	   else
		  command = tTunerTuningCommandMap[output] .. CMDS[idBinding][command_name]
	   end	
    else
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[idBinding][command_name]
		  end			  
	   else
		  if (output == 0) then
			 command = tNetTuneCommandMap[output] .. CMDS[idBinding][command_name]
		  else
			 --process digits for Internet Radio Preset Command
			 local digit = command_name:match("NUMBER_(.)")		--NUMBER_1
			 gNetTunePresetCommand.prefix = tNetTunePresetCommandMap[output]
			 if (gNetTunePresetCommand.preset == nil) then
				gNetTunePresetCommand.preset = digit
				gNetTunePresetCommand.prefix = tNetTunePresetCommandMap[output]
				gNetTunePresetCommandTimer:StartTimer()
			 else
				gNetTunePresetCommand.preset = gNetTunePresetCommand.preset .. digit
			 end
		  end 
	   end
    end
	
    if (command == nil) then
	   LOG:Warn('GetDigitCommand(): Command not defined - '.. command_name)
    end	
    return command
end

function OnNetTunePresetCommandExpired()
    LogTrace("OnNetTunePresetCommandExpired(), Prefix=" .. gNetTunePresetCommand.prefix .. ", Preset=" .. gNetTunePresetCommand.preset)
    gNetTunePresetCommandTimer:KillTimer()
    local command = gNetTunePresetCommand.prefix .. string.format('%02X', gNetTunePresetCommand.preset)
    PackAndQueueCommand("NetTunePresetCommand", command)
    gNetTunePresetCommand.preset = nil
end

function GetTuningCommand(idBinding, output, command_name)
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end	
    LogTrace("GetTuningCommand(), idBinding=" .. idBinding .. ", output=" .. output .. ", command_name=" .. command_name)
    
    local command
    if (idBinding == TUNER_PROXY_BINDINGID) then 
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[idBinding][command_name]
		  end		  
	   else
		  command = tTunerTuningCommandMap[output] .. CMDS[idBinding][command_name]
	   end	
    else
	   if (gControlMethod == "IR") then			
		  --command = CMDS_IR_ZONES[output][command_name]
		  if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
			 command = CMDS_IR_ZONES[output][command_name]
		  else	--"Pioneer"
			 command = CMDS_IR[idBinding][command_name]
		  end		  
	   else
		  command = tNetTuneCommandMap[output] .. CMDS[idBinding][command_name]
	   end
    end
	
    if (command == nil) then
	   LOG:Warn('GetTuningCommand(): Command not defined - '.. command_name)
    end	
    return command
end

function GetPresetTuningCommand(idBinding, output, command_name)
    --PAGE_UP PAGE_DOWN mapping for SR Remotes...
    if (command_name == "PAGE_UP") then command_name = "PRESET_UP" end
    if (command_name == "PAGE_DOWN") then command_name = "PRESET_DOWN" end
    
    if (output == nil) then
	   output = findOutputFromBindingID(idBinding)
    end
    LogTrace("GetPresetTuningCommand(), idBinding=" .. idBinding .. ", output=" .. output .. ", command_name=" .. command_name)
    
    local command
    if (gControlMethod == "IR") then			
	   --command = CMDS_IR_ZONES[output][command_name]
	   if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
		  command = CMDS_IR_ZONES[output][command_name]
	   else	--"Pioneer"
		  command = CMDS_IR[idBinding][command_name]
	   end	   
    else
	   command = tTunerPresetCommandMap[output] .. CMDS[idBinding][command_name]
    end	

    if (command == nil) then
	   LOG:Warn('GetPresetTuningCommand(): Command not defined - '.. command_name)
    end	
    return command
end

function printIOMap()
	for j,k in pairs(gCurrentlySelectedInput) do
		if (tInputConnMapByID[k] ~= nil) then
			print (tOutputConnMap[j],tInputConnMapByID[k].Name)
		end	
	end
end

--[[
	Helper Function: GET_DEVICE_STATUS
--]]
function GetDeviceVolumeStatus(output)
    LogTrace("GetDeviceVolumeStatus(), output = " .. output)
	local command = tVolumeQueryMap[output] 
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Volume", command)
	
	command = tMuteCommandMap[output] .. "QSTN"
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Mute", command)

end

 end)
package.preload['properties'] = (function (...)
--[[=============================================================================
    Properties Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

function ON_PROPERTY_CHANGED.NetworkKeepAliveIntervalSeconds(propertyValue)
	gNetworkKeepAliveInterval = tonumber(propertyValue)

end

function ON_PROPERTY_CHANGED.TunerFrequencyStep(propertyValue)
	--Tuner Frequency Steps - to support Global AM/FM Tuning functions
	gFreqStep = propertyValue
end

function ON_PROPERTY_CHANGED.RampingMethod(propertyValue)
    if (gInitializingDriver == true) then return end
    
    if (propertyValue == "Pulse Commands") then 
	   gReceiverProxy._UsePulseCommandsForVolumeRamping = true 
    else
	   gReceiverProxy._UsePulseCommandsForVolumeRamping = false
    end
    
end

function ON_PROPERTY_CHANGED.VolumeRampSteps(propertyValue)
    if (gInitializingDriver == true) then return end
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    
end

function ON_PROPERTY_CHANGED.VolumeRampSlope(propertyValue)
    if (gInitializingDriver == true) then return end
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)

	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end end)
package.preload['receiver_init_conditional'] = (function (...)
-----------------------------------------------------------------------------------------------------
------------------------------------------------Integra----------------------------------------------
-----------------------------------------------------------------------------------------------------

function Integra_CommandsTableInit_IR()
	LOG:Trace("CommandsTableInit_IR()")

	CMDS_IR = {}		
	
	--IR ZONE Commands
	CMDS_IR_ZONES = {}
	
	CMDS_IR_ZONES[0] = {
		--index:  Proxy Command Name
		--value:  IR Code ID
		["CONNECT_OUTPUT"] = "21001",
		["DISCONNECT_OUTPUT"] = "21000",	
		["VOLUME_UP"] = "21005",
		["VOLUME_DOWN"] = "21006",
		["MUTE_ON"] = "21003",
		["MUTE_OFF"] = "21002",
		["MUTE_TOGGLE"] = "21004",
		
		["STAR"] = "21007",			--Direct Access
		["NUMBER_0"] = "21008", 	
		["NUMBER_1"] = "21009",	
		["NUMBER_2"] = "21010",	
		["NUMBER_3"] = "21011",	
		["NUMBER_4"] = "21012",	
		["NUMBER_5"] = "21013",	
		["NUMBER_6"] = "21014",	
		["NUMBER_7"] = "21015",	
		["NUMBER_8"] = "21016",	
		["NUMBER_9"] = "21017",
		
		["PULSE_CH_UP"] = "21018",
		["START_CH_UP"] = "21018",		--used for PULSE_CH_UP in some Navigators
		["SKIP_FWD"] = "21018",		--used for PULSE_CH_UP in some Navigators	
		["PULSE_CH_DOWN"] = "21019",
		["START_CH_DOWN"] = "21019",	--used for PULSE_CH_DOWN in some Navigators
		["SKIP_REV"] = "21019",		--used for PULSE_CH_DOWN in some Navigators
		["PRESET_UP"] = "21020",	
		["PRESET_DOWN"] = "21021",
		["INPUT_TOGGLE"] = "",	  --BAND Toggle NOT MAPPED?		
		
		["PLAY"] = "21022",		
		["STOP"] = "21023",
		["PAUSE"] = "21024",	
		["SKIP_FWD"] = "21025",
		["SKIP_REV"] = "21026",
		["SCAN_FWD"] = "21027",
		["SCAN_REV"] = "21028",
		
		["INFO"] = "21029",  	--DISPLAY
		
		["RIGHT"] = "21030",
		["LEFT"] = "21031",
		["UP"] = "21032",
		["DOWN"] = "21033",
		["ENTER"] = "21034",
		
		["CANCEL"] = "21035",	--RETURN
		["MENU"] = "21036",
		["GUIDE"] = "21037",	--TOP MENU

	}	
	
	CMDS_IR_ZONES[1] = {
		["CONNECT_OUTPUT"] = "21063",
		["DISCONNECT_OUTPUT"] = "21062",	
		["VOLUME_UP"] = "21064",
		["VOLUME_DOWN"] = "21065",
		["MUTE_ON"] = "21066",
		["MUTE_OFF"] = "21067",
		["MUTE_TOGGLE"] = "21068",
		
		["STAR"] = "21069",			--Direct Access
		["NUMBER_0"] = "21070", 	
		["NUMBER_1"] = "21071",	
		["NUMBER_2"] = "21072",	
		["NUMBER_3"] = "21073",	
		["NUMBER_4"] = "21074",	
		["NUMBER_5"] = "21075",	
		["NUMBER_6"] = "21076",	
		["NUMBER_7"] = "21077",	
		["NUMBER_8"] = "21078",	
		["NUMBER_9"] = "21079",	
		
		["PULSE_CH_UP"] = "21080",
		["START_CH_UP"] = "21080",		--used for PULSE_CH_UP in some Navigators
		--["SKIP_FWD"] = "21080",		--used for PULSE_CH_UP in some Navigators	
		["PULSE_CH_DOWN"] = "21081",
		["START_CH_DOWN"] = "21081",	--used for PULSE_CH_DOWN in some Navigators
		--["SKIP_REV"] = "21081",		--used for PULSE_CH_DOWN in some Navigators
		["PRESET_UP"] = "21082",	
		["PRESET_DOWN"] = "21083",
		--["INPUT_TOGGLE"] = "",	  --BAND Toggle NOT MAPPED?
		
		["PLAY"] = "21084",		
		["STOP"] = "21085",
		["PAUSE"] = "21086",	
		["SKIP_FWD"] = "21087",
		["SKIP_REV"] = "21088",
		["SCAN_FWD"] = "21089",
		["SCAN_REV"] = "21090",		
		
		["INFO"] = "21091",  	--DISPLAY
		["RIGHT"] = "21092",
		["LEFT"] = "21093",
		["UP"] = "21094",
		["DOWN"] = "21095",
		["ENTER"] = "21096",	
		
		["CANCEL"] = "21097",	--RETURN
		--["MENU"] = "21098",
		--["GUIDE"] = "21099",	--TOP MENU		
	}		
		

	CMDS_IR_ZONES[2] = {
		["CONNECT_OUTPUT"] = "21125",
		["DISCONNECT_OUTPUT"] = "21124",	
		["VOLUME_UP"] = "21126",
		["VOLUME_DOWN"] = "21127",
		["MUTE_ON"] = "21128",
		["MUTE_OFF"] = "21129",
		["MUTE_TOGGLE"] = "21130",
		
		["STAR"] = "21131",			--Direct Access
		["NUMBER_0"] = "21132", 	
		["NUMBER_1"] = "21133",	
		["NUMBER_2"] = "21134",	
		["NUMBER_3"] = "21135",	
		["NUMBER_4"] = "21136",	
		["NUMBER_5"] = "21137",	
		["NUMBER_6"] = "21138",	
		["NUMBER_7"] = "21139",	
		["NUMBER_8"] = "21140",	
		["NUMBER_9"] = "21141",	
		
		["PULSE_CH_UP"] = "21142",
		["START_CH_UP"] = "21142",		--used for PULSE_CH_UP in some Navigators
		["SKIP_FWD"] = "21142",		--used for PULSE_CH_UP in some Navigators	
		["PULSE_CH_DOWN"] = "21143",
		["START_CH_DOWN"] = "21143",	--used for PULSE_CH_DOWN in some Navigators
		["SKIP_REV"] = "21143",		--used for PULSE_CH_DOWN in some Navigators
		["PRESET_UP"] = "21144",	
		["PRESET_DOWN"] = "21145",
		--["INPUT_TOGGLE"] = "",	  --BAND Toggle NOT MAPPED?
		
		["PLAY"] = "21146",		
		["STOP"] = "21147",
		["PAUSE"] = "21148",	
		["SKIP_FWD"] = "21149",
		["SKIP_REV"] = "21150",
		["SCAN_FWD"] = "21151",
		["SCAN_REV"] = "21152",		
		
		["INFO"] = "21153",  	--DISPLAY
		["RIGHT"] = "21154",
		["LEFT"] = "21155",
		["UP"] = "21156",
		["DOWN"] = "21157",
		["ENTER"] = "21158",	
		
		["CANCEL"] = "21159",	--RETURN
		["MENU"] = "21160",
		["GUIDE"] = "21161",	--TOP MENU		
	}	
	
	CMDS_IR_ZONES[3] = {
		["CONNECT_OUTPUT"] = "21186",
		["DISCONNECT_OUTPUT"] = "21187",	
		["VOLUME_UP"] = "21188",
		["VOLUME_DOWN"] = "21189",
		["MUTE_ON"] = "21190",
		["MUTE_OFF"] = "21191",
		["MUTE_TOGGLE"] = "21192",
		
		["STAR"] = "21193",			--Direct Access
		["NUMBER_0"] = "21194", 	
		["NUMBER_1"] = "21195",	
		["NUMBER_2"] = "21196",	
		["NUMBER_3"] = "21197",	
		["NUMBER_4"] = "21198",	
		["NUMBER_5"] = "21199",	
		["NUMBER_6"] = "21200",	
		["NUMBER_7"] = "21201",	
		["NUMBER_8"] = "21202",	
		["NUMBER_9"] = "21203",	
		
		["PULSE_CH_UP"] = "21204",
		["START_CH_UP"] = "21204",		--used for PULSE_CH_UP in some Navigators
		["SKIP_FWD"] = "21204",		--used for PULSE_CH_UP in some Navigators	
		["PULSE_CH_DOWN"] = "21205",
		["START_CH_DOWN"] = "21205",	--used for PULSE_CH_DOWN in some Navigators
		["SKIP_REV"] = "21205",		--used for PULSE_CH_DOWN in some Navigators
		["PRESET_UP"] = "21206",	
		["PRESET_DOWN"] = "21207",
		--["INPUT_TOGGLE"] = "",	  --BAND Toggle NOT MAPPED?
		
		["PLAY"] = "21208",		
		["STOP"] = "21209",
		["PAUSE"] = "21210",	
		["SKIP_FWD"] = "21211",
		["SKIP_REV"] = "21212",
		["SCAN_FWD"] = "21213",
		["SCAN_REV"] = "21214",		
		
		["INFO"] = "21215",  	--DISPLAY
		["RIGHT"] = "21216",
		["LEFT"] = "21217",
		["UP"] = "21218",
		["DOWN"] = "21219",
		["ENTER"] = "21220",	
		
		["CANCEL"] = "21221",	--RETURN
		["MENU"] = "21222",
		["GUIDE"] = "21223",	--TOP MENU			
	}
	
	CMDS_IR[RECEIVER_PROXY_BINDINGID] = {}	
	
	CMDS_IR[TUNER_PROXY_BINDINGID] = {
		----values are IRCode ID as defined in IRCode XML declarations
		--["PULSE_CH_UP"]="",
		--["START_CH_UP"]="",		--used for PULSE_CH_UP in some Navigators
		--["SKIP_FWD"]="",		--used for PULSE_CH_UP in some Navigators	
		--["PULSE_CH_DOWN"]="",
		--["START_CH_DOWN"]="",	--used for PULSE_CH_DOWN in some Navigators
		--["SKIP_REV"]="",		--used for PULSE_CH_DOWN in some Navigators
		--["PRESET_UP"]="",	
		--["PRESET_DOWN"]="",
		--["INPUT_TOGGLE"]="",	  --BAND Toggle NOT MAPPED?
		
		----these entries are only required if 
		----specific command codes are required for the Tuner
		--["STAR"]="TAC",
		--["NUMBER_0"]="", 	
		--["NUMBER_1"]="",	
		--["NUMBER_2"]="",	
		--["NUMBER_3"]="",	
		--["NUMBER_4"]="",	
		--["NUMBER_5"]="",	
		--["NUMBER_6"]="",	
		--["NUMBER_7"]="",	
		--["NUMBER_8"]="",	
		--["NUMBER_9"]="",
	}	
	
	EX_CMD_IR = {
		["HDMI OUPUT: OFF"] = "21248",
		["HDMI OUPUT: MAIN"] = "21249",
		["HDMI OUPUT: SUB"] = "21250",
		["HDMI OUPUT: BOTH"] = "21251",
		["HDMI OUPUT: BOTH (MAIN)"] = "21252",
		["HDMI OUPUT: BOTH (SUB)"] = "21253",	
	}
end

function Integra_InputOutputTableInit()
	tInputCommandMap_IR = { 
	}
	--Main Zone Input Selection Commands	
	tInputCommandMap_IR[0] = { 
		--index:  Connection Name
		--value:  IR Code ID
		["INPUT AUX"] = "21043",
		["INPUT BD/DVD"] = "21049",
		["INPUT CBL/SAT"] = "21041",
		["INPUT GAME"] = "21042",
		["INPUT GAME1"] = "21042",
		["INPUT GAME2"] = "21044",
		["INPUT EXTRA1"] = "21046",
		["INPUT EXTRA2"] = "21047",
		["INPUT EXTRA3"] = "21048",
		["INPUT PC"] = "21045",
		["INPUT PHONO"] = "21051",
		["INPUT STB/DVR"] = "21040",
		["INPUT TV/CD"] = "21052",
		["INPUT CD"] = "21052",
		["INPUT STRM BOX"] = "21254",
		["INPUT TV"] = "21258",	
		["INPUT TV/TAPE"] = "21050",
		["INPUT DAB"] = "21055",
		["FM Antenna"] = "21053",
		["AM Antenna"] = "21054",
		["AM/FM TUNER"] = "21038",
		["NET"] = "21039",
		["TUNEIN"] = "21039",		--NET code, discrete code not supported via IR
		["SIRIUSXM"] = "21039",		--NET code, discrete code not supported via IR
		["SPOTIFY"] = "21039",		--NET code, discrete code not supported via IR
		["MP3TUNES"] = "21039",		--NET code, discrete code not supported via IR
		["DLNA"] = "21039",			--NET code, discrete code not supported via IR
		["PANDORA"] = "21039",		--NET code, discrete code not supported via IR
		["LAST.FM"] = "21039",		--NET code, discrete code not supported via IR
		["AUPEO!"] = "21039",		--NET code, discrete code not supported via IR
		["MY FAVORITES"] = "21039",	--NET code, discrete code not supported via IR
		["HOME MEDIA"] = "21039",	--NET code, discrete code not supported via IR
		["RHAPSODY"] = "21039",		--NET code, discrete code not supported via IR
		["SLACKER"] = "21039",		--NET code, discrete code not supported via IR
		["USB (FRONT)"] = "21058",	--"21060",
		["USB (REAR)"] = "21059",
		["BLUETOOTH"] = "21061",
		["DEEZER"] = "21039",		--NET code, discrete code not supported via IR
		["AIRPLAY"] = "21039",		--NET code, discrete code not supported via IR
	}
	--Zone2 Input Selection Commands
	tInputCommandMap_IR[1] = { 	
		["INPUT AUX"] = "21105",
		["INPUT BD/DVD"] = "21111",
		["INPUT CBL/SAT"] = "21103",
		["INPUT GAME"] = "21104",
		["INPUT GAME1"] = "21104",
		["INPUT GAME2"] = "21106",
		["INPUT EXTRA1"] = "21108",
		["INPUT EXTRA2"] = "21109",
		["INPUT EXTRA3"] = "21110",
		["INPUT PC"] = "21107",
		["INPUT PHONO"] = "21113",
		["INPUT STB/DVR"] = "21102",
		["INPUT TV/CD"] = "21114",
		["INPUT CD"] = "21114",
		["INPUT STRM BOX"] = "21255",
		["INPUT TV"] = "21259",	
		["INPUT TV/TAPE"] = "21112",
		["INPUT DAB"] = "21117",
		["FM Antenna"] = "21115",
		["AM Antenna"] = "21116",
		["AM/FM TUNER"] = "21100",
		["NET"] = "21101",
		["TUNEIN"] = "21101",		--NET code, discrete code not supported via IR
		["SIRIUSXM"] = "21101",		--NET code, discrete code not supported via IR
		["SPOTIFY"] = "21101",		--NET code, discrete code not supported via IR
		["MP3TUNES"] = "21101",		--NET code, discrete code not supported via IR
		["DLNA"] = "21101",			--NET code, discrete code not supported via IR
		["PANDORA"] = "21101",		--NET code, discrete code not supported via IR
		["LAST.FM"] = "21101",		--NET code, discrete code not supported via IR
		["AUPEO!"] = "21101",		--NET code, discrete code not supported via IR
		["MY FAVORITES"] = "21101",	--NET code, discrete code not supported via IR
		["HOME MEDIA"] = "21101",	--NET code, discrete code not supported via IR
		["RHAPSODY"] = "21101",		--NET code, discrete code not supported via IR
		["SLACKER"] = "21101",		--NET code, discrete code not supported via IR
		["USB (FRONT)"] = "21120",	--"21122",
		["USB (REAR)"] = "21121",
		["BLUETOOTH"] = "21123",
		["DEEZER"] = "21101",		--NET code, discrete code not supported via IR
		["AIRPLAY"] = "21101",		--NET code, discrete code not supported via IR
	}
	--Zone3 Input Selection Commands
	tInputCommandMap_IR[2] = { 
		["INPUT AUX"] = "21167",
		["INPUT BD/DVD"] = "21173",
		["INPUT CBL/SAT"] = "21165",
		["INPUT GAME"] = "21166",
		["INPUT GAME1"] = "21166",
		["INPUT GAME2"] = "21168",
		["INPUT EXTRA1"] = "21170",
		["INPUT EXTRA2"] = "21171",
		["INPUT EXTRA3"] = "21172",
		["INPUT PC"] = "21169",
		["INPUT PHONO"] = "21175",
		["INPUT STB/DVR"] = "21164",
		["INPUT TV/CD"] = "21176",
		["INPUT CD"] = "21176",
		["INPUT STRM BOX"] = "21256",
		["INPUT TV"] = "21260",	
		["INPUT TV/TAPE"] = "21174",
		["INPUT DAB"] = "21179",
		["FM Antenna"] = "21177",
		["AM Antenna"] = "21178",
		["AM/FM TUNER"] = "21162",
		["NET"] = "21163",
		["TUNEIN"] = "21163",		--NET code, discrete code not supported via IR
		["SIRIUSXM"] = "21163",		--NET code, discrete code not supported via IR
		["SPOTIFY"] = "21163",		--NET code, discrete code not supported via IR
		["MP3TUNES"] = "21163",		--NET code, discrete code not supported via IR
		["DLNA"] = "21163",			--NET code, discrete code not supported via IR
		["PANDORA"] = "21163",		--NET code, discrete code not supported via IR
		["LAST.FM"] = "21163",		--NET code, discrete code not supported via IR
		["AUPEO!"] = "21163",		--NET code, discrete code not supported via IR
		["MY FAVORITES"] = "21163",	--NET code, discrete code not supported via IR
		["HOME MEDIA"] = "21163",	--NET code, discrete code not supported via IR
		["RHAPSODY"] = "21163",		--NET code, discrete code not supported via IR
		["SLACKER"] = "21163",		--NET code, discrete code not supported via IR
		["USB (FRONT)"] = "21182",	--21184
		["USB (REAR)"] = "21183",
		["BLUETOOTH"] = "21185",
		["DEEZER"] = "21163",		--NET code, discrete code not supported via IR
		["AIRPLAY"] = "21163",		--NET code, discrete code not supported via IR
	}
	--HD Zone Input Selection Commands	
	tInputCommandMap_IR[3] = { 
		["INPUT AUX"] = "21229",
		["INPUT BD/DVD"] = "21235",
		["INPUT CBL/SAT"] = "21227",
		["INPUT GAME"] = "21228",
		["INPUT GAME1"] = "21228",
		["INPUT GAME2"] = "21230",
		["INPUT EXTRA1"] = "21232",
		["INPUT EXTRA2"] = "21233",
		["INPUT EXTRA3"] = "21234",
		["INPUT PC"] = "21231",
		["INPUT PHONO"] = "21237",
		["INPUT STB/DVR"] = "21226",
		["INPUT TV/CD"] = "21238",
		["INPUT CD"] = "21238",
		["INPUT STRM BOX"] = "21257",
		["INPUT TV"] = "21261",		
		["INPUT TV/TAPE"] = "21236",
		["INPUT DAB"] = "21241",
		["FM Antenna"] = "21239",
		["AM Antenna"] = "21240",
		["AM/FM TUNER"] = "21224",
		["NET"] = "21225",
		["TUNEIN"] = "21225",		--NET code, discrete code not supported via IR
		["SIRIUSXM"] = "21225",		--NET code, discrete code not supported via IR
		["SPOTIFY"] = "21225",		--NET code, discrete code not supported via IR
		["MP3TUNES"] = "21225",		--NET code, discrete code not supported via IR
		["DLNA"] = "21225",			--NET code, discrete code not supported via IR
		["PANDORA"] = "21225",		--NET code, discrete code not supported via IR
		["LAST.FM"] = "21225",		--NET code, discrete code not supported via IR
		["AUPEO!"] = "21225",		--NET code, discrete code not supported via IR
		["MY FAVORITES"] = "21225",	--NET code, discrete code not supported via IR
		["HOME MEDIA"] = "21225",	--NET code, discrete code not supported via IR
		["RHAPSODY"] = "21225",		--NET code, discrete code not supported via IR
		["SLACKER"] = "21225",		--NET code, discrete code not supported via IR
		["USB (FRONT)"] = "21244",	--21246
		["USB (REAR)"] = "21245",
		["BLUETOOTH"] = "21247",
		["DEEZER"] = "21225",		--NET code, discrete code not supported via IR
		["AIRPLAY"] = "21225",		--NET code, discrete code not supported via IR
	}	
	
	tInputCommandMap_IR[TUNER_PROXY_BINDINGID] = {
		--values are IRCode ID as defined in IRCode XML declarations
		["AM Antenna"] = "",
		["FM Antenna"] = "", 
	}
end



-----------------------------------------------------------------------------------------------------
------------------------------------------------Pioneer----------------------------------------------
-----------------------------------------------------------------------------------------------------

function Pioneer_CommandsTableInit_IR()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--['POWER_ON'] = '51071',	--handled by CONNECT_OUTPUT
		--['POWER_OFF'] = '51072',	--handled by DISCONNECT_OUTPUT
		['POWER_TOGGLE'] = '51070',		
		['VIRTUAL HEIGHT'] = '51413',
		['VIRTUAL HEIGHT OFF'] = '51415',
		['VIRTUAL HEIGHT ON'] = '51414',
		['VIRTUAL SB'] = '51410',
		['VIRTUAL SB OFF'] = '51412',
		['VIRTUAL SB ON'] = '51411',
		['MENU'] = '51401',
		['CANCEL'] = '51402', 		--Pioneer code "RETURN"
		['ENTER'] = '51403',
		['UP'] = '51079',
		['DOWN'] = '51080',
		['LEFT'] = '51082',
		['RIGHT'] = '51081',
		['LOUDNESS_TOGGLE'] = '51095',	
		['HDMI OUT 1+2 ON']='51404',
		['HDMI OUT 1 ON']='51405',
		['HDMI OUT 2 ON']='51406',
	}		
	
	CMDS_IR[RECEIVER_PROXY_BINDINGID] = {}	
	
	CMDS_IR[TUNER_PROXY_BINDINGID] = {
		['CLASS'] = '51121',
		['DIRECT_ACCESS'] = '51124',
		['GUIDE'] = '51123',		--Pioneer code "DISPLAY"
		['MPX/NOISE CUT'] = '51119',
		['PTY SEARCH'] = '51366',
		['TUNER EDIT'] = '51122',
		['NUMBER_0'] = '51104',
		['NUMBER_1'] = '51105',
		['NUMBER_2'] = '51106',
		['NUMBER_3'] = '51107',
		['NUMBER_4'] = '51108',
		['NUMBER_5'] = '51109',
		['NUMBER_6'] = '51110',
		['NUMBER_7'] = '51111',
		['NUMBER_8'] = '51112',
		['NUMBER_9'] = '51113',
		['CANCEL'] = '51365', 		--Pioneer code "RETURN"
		['ENTER'] = '51118',
		--['TUNE_UP'] = '51116',
		--['TUNE_DOWN'] = '51117',
		['PULSE_CH_UP'] = '51116',
		['START_CH_UP'] = '51116',
		['SKIP_FWD']='51116',	
		['PULSE_CH_DOWN'] = '51117',
		['START_CH_DOWN'] = '51117',
		['SKIP_REV']='51117',
		['PRESET_UP'] = '51114',
		['PRESET_DOWN'] = '51115',
		['AM Antenna'] = '51136',
		['FM Antenna'] = '51137',
		['TOGGLE_AM_FM'] = '51120',
		['PAGE_UP'] = '51114',
		['PAGE_DOWN'] = '51115',		
	}
	
	CMDS_IR[BLUETOOTH_PROXY_BINDINGID] = {
		['PLAY']='51183',	
		['PAUSE']='51184',
		['STOP']='51185',	
		['SKIP_REV']='51186',
		['SKIP_FWD']='51187',	 
		['SCAN_REV']='51188',
		['SCAN_FWD']='51189',
		['MENU']='51190',	--DISPLAY
		['UP']='51191',
		['DOWN']='51192',
		['LEFT']='51194',
		['RIGHT']='51193',
		['ENTER'] = '51195',
		['CANCEL'] = '51196',  --RETURN
		['REPEAT'] = '51197',
		['RANDOM'] = '51198',
		['GUIDE'] = '51199',  --TOP MENU
	}

	--[[
	CMDS_IR[IPODUSB_PROXY_BINDINGID] = {
		['iPod_Control'] = '51367',
		['iPod_Repeat'] = '51211',
		['iPod_Shuffle'] = '51212',
		['PLAY'] = '51204',
		['STOP'] = '51206',
		['PAUSE'] = '51205',
		['SKIP_FWD'] = '51208',
		['SKIP_REV'] = '51207',
		['SCAN_FWD'] = '51210',
		['SCAN_REV'] = '51209',
		['MENU'] = '51220',
		['RETURN'] = '51219',
		['ENTER'] = '51218',
		['UP'] = '51214',
		['DOWN'] = '51215',
		['LEFT'] = '51217',
		['RIGHT'] = '51216',
		['DISPLAY'] = '51213',
	}
	

	CMDS_IR[MHL_PROXY_BINDINGID] = {
		['SELECT']='51200',
		['UP']='51201',
		['DOWN']='51202',
		['LEFT']='51203',
		['RIGHT']='51222',
		['GUIDE']='51223', --ROOT MENU
		['CANCEL']='51254',  --EXIT
		['NUMBER_0']='51255', 	
		['NUMBER_1']='51256',	
		['NUMBER_2']='51257',	
		['NUMBER_3']='51258',	
		['NUMBER_4']='51259',	
		['NUMBER_5']='51260',	
		['NUMBER_6']='51261',	
		['NUMBER_7']='51262',	
		['NUMBER_8']='51263',	
		['NUMBER_9']='51264',
		['ENTER']='51265',
		['CLEAR']='51266',
		['CHANNEL_UP']='51267',
		['CHANNEL_DOWN']='51268',
		['SOUND_SELECT']='51269',
		['PLAY']='51270',	
		['STOP']='51271',
		['PAUSE']='51272',
		['SCAN_FWD']='51274',	--FAST FORWARD
		['SCAN_REV']='51273',	--REWIND
		['SKIP_REV']='51276',	--BACKWARD
		['SKIP_FWD']='51275',	--FORWARD
	}	
	--]]
	
	CMDS_IR[NET_PROXY_BINDINGID] = {
		['CLASS'] = '51400',
		['CLEAR'] = '51395',
		['EDIT'] = '51399',
		['PROGRAM'] = '51394',
		['RANDOM'] = '51397',
		['REPEAT'] = '51396',
		['PLAY'] = '51385',
		['STOP'] = '51387',
		['SKIP_FWD'] = '51324',
		['SKIP_REV'] = '51325',
		['NUMBER_0'] = '51375',
		['NUMBER_1'] = '51376',
		['NUMBER_2'] = '51377',
		['NUMBER_3'] = '51378',
		['NUMBER_4'] = '51379',
		['NUMBER_5'] = '51380',
		['NUMBER_6'] = '51381',
		['NUMBER_7'] = '51382',
		['NUMBER_8'] = '51383',
		['NUMBER_9'] = '51384',
		['MENU'] = '51398',
		['CANCEL'] = '51393',	--Pioneer code "RETURN"
		['ENTER'] = '51392',
		['UP'] = '51388',
		['DOWN'] = '51389',
		['LEFT'] = '51391',
		['RIGHT'] = '51390',
		['INFO'] = '51386',
	}
	
	CMDS_IR_ZONES = {}
	
	CMDS_IR_ZONES[0] = {
		['CONNECT_OUTPUT'] = '51071',
		['DISCONNECT_OUTPUT'] = '51072',	
		['VOLUME_UP'] = '51077',				
		['VOLUME_DOWN'] = '51078',			
		['MUTE_ON'] = '51074',
		['MUTE_OFF'] = '51075',
		['MUTE_TOGGLE'] = '51073',
	}	
	
	CMDS_IR_ZONES[1] = {
		['CONNECT_OUTPUT'] = '51484',			--51140
		['DISCONNECT_OUTPUT'] = '51485',		--51141
		['VOLUME_UP'] = '51487',				--51154
		['VOLUME_DOWN'] = '51488',			--51155
		['MUTE_ON'] = '51065',
		['MUTE_OFF'] = '51066',
		['MUTE_TOGGLE'] = '51064',
	}		
	
	CMDS_IR_ZONES[2] = {
		['CONNECT_OUTPUT'] = '51174',
		['DISCONNECT_OUTPUT'] = '51173',	
		['VOLUME_UP'] = '51084',
		['VOLUME_DOWN'] = '51085',
		['MUTE_ON'] = '51076',
		['MUTE_OFF'] = '51083',
		['MUTE_TOGGLE'] = '51067'		
	}	
	
	CMDS_IR_ZONES[3] = {
		['CONNECT_OUTPUT'] = '51089',
		['DISCONNECT_OUTPUT'] = '51090',	
		['VOLUME_UP'] = '51457',
		['VOLUME_DOWN'] = '51458',
		['MUTE_ON'] = '51455',
		['MUTE_OFF'] = '51456',
		['MUTE_TOGGLE'] = '51454'		
	}		
		
end

function Pioneer_InputOutputTableInit()

	tInputCommandMap_IR = {}
	
	tInputCommandMap_IR[0] = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		["INPUT AUX"] = "51478",
		["INPUT BD/DVD"] = "51314",	
		["INPUT CBL/SAT"] = "51448",
		["INPUT GAME"] = "51476",
		["INPUT PHONO"] = "51059",
		["INPUT CD"] = "51060",
		["INPUT STRM BOX"] = "51478",
		["INPUT TV/TAPE"] = "51063",
		["FM Antenna"] = "51137",
		["AM Antenna"] = "51136",
		["AM/FM TUNER"] = "51061",
		["NET"] = "51127",
		["USB (FRONT)"] = "51157",
		["BLUETOOTH"] = "51362",
		["INPUT HDMI 5"] = "51354",
		["INPUT HDMI 6"] = "51086",
		["INPUT HDMI 7"] = "51087",
		["INPUT TV"] = "51063",
	}
	
	tInputCommandMap_IR[1] = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		["INPUT AUX"] = "51479",
		["INPUT BD/DVD"] = "51480",		--51145
		["INPUT CBL/SAT"] = "51449",
		["INPUT GAME"] = "51477",
		["INPUT PHONO"] = "51482",		--51152
		["INPUT CD"] = "51481",			--51149
		["INPUT STRM BOX"] = "51479",
		["INPUT TV/TAPE"] = "51486",		--51146
		["AM/FM TUNER"] = "51483",		--51151
		["NET"] = "51134",	
		["USB (FRONT)"] = "51158",
		["BLUETOOTH"] = "51363",	--51133
		["INPUT HDMI 5"] = "51489",
		["INPUT TV"] = "51486",		--51146
	}	

end end)
package.preload['receiver_init_shared'] = (function (...)
--[[=============================================================================
    Receiver Protocol Initialization Functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2014.10.23"
end

PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_EARLY_INIT.receiver_init()
end

function ON_DRIVER_INIT.receiver_init()
end

function ON_DRIVER_LATEINIT.receiver_init()
    DRIVER_MANUFACTURER = C4:GetDriverConfigInfo("manufacturer") or "Integra"	--"Onkyo" "Pioneer"
    
	--LogTrace("ON_DRIVER_INIT.ProtocolDeclarations()")
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
end

----------------------------------------- Protocol Helper Functions -----------------------------------------

function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
    if (DRIVER_MANUFACTURER == "Integra" or DRIVER_MANUFACTURER == "Onkyo") then
	   Integra_CommandsTableInit_IR()
	   Integra_InputOutputTableInit()
    else	--"Pioneer"
	   Pioneer_CommandsTableInit_IR()
	   Pioneer_InputOutputTableInit()
    end
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
	LOG:Trace("CommandsTableInit_Serial()")

	CMDS = {}
	
	CMDS[RECEIVER_PROXY_BINDINGID] = {
		["GUIDE"] = "OSDHOME",
		["MENU"] = "OSDMENU",
		["CANCEL"] = "OSDEXIT",
		["UP"] = "OSDUP",
		["DOWN"] = "OSDDOWN",
		["LEFT"] = "OSDLEFT",
		["RIGHT"] = "OSDRIGHT",
		["ENTER"] = "OSDENTER",	
	}
	
	CMDS[TUNER_PROXY_BINDINGID] = {
		["PULSE_CH_UP"]="UP",
		["START_CH_UP"]="UP",		--used for PULSE_CH_UP in some Navigators
		["SKIP_FWD"]="UP",		--used for PULSE_CH_UP in some Navigators	
		["PULSE_CH_DOWN"]="DOWN",
		["START_CH_DOWN"]="DOWN",	--used for PULSE_CH_DOWN in some Navigators
		["SKIP_REV"]="DOWN",		--used for PULSE_CH_DOWN in some Navigators
		["PRESET_UP"]="UP",	
		["PRESET_DOWN"]="DOWN",
		["INPUT_TOGGLE"]="",	  --BAND Toggle NOT MAPPED?
		
		--these entries are only required if 
		--specific command strings are required for the Tuner
		["STAR"]	   = "DIRECT",
		["NUMBER_0"] = "0",
		["NUMBER_1"] = "1",
		["NUMBER_2"] = "2",
		["NUMBER_3"] = "3",
		["NUMBER_4"] = "4",
		["NUMBER_5"] = "5",
		["NUMBER_6"] = "6",
		["NUMBER_7"] = "7",
		["NUMBER_8"] = "8",
		["NUMBER_9"] = "9",
	}	
	
	CMDS[NET_PROXY_BINDINGID] = {
		--Menu
		["INFO"] = "DISPLAY",	--Display
		["GUIDE"] = "TOP",
		["MENU"] = "MENU",
		["CANCEL"] = "RETURN",
		["UP"] = "UP",
		["DOWN"] = "DOWN",
		["LEFT"] = "LEFT",
		["RIGHT"] = "RIGHT",
		["ENTER"] = "SELECT",	
		--["RECALL"]         = "",
		--["PREV"]           = "",
		
		--Digits
		["NUMBER_0"] = "0", 	
		["NUMBER_1"] = "1",		
		["NUMBER_2"] = "2",	
		["NUMBER_3"] = "3",	
		["NUMBER_4"] = "4",	
		["NUMBER_5"] = "5",	
		["NUMBER_6"] = "6",	
		["NUMBER_7"] = "7",	
		["NUMBER_8"] = "8",	
		["NUMBER_9"] = "9",
		--["STAR"]   = "",
		--["DOT"]      = "",
		
		--iRadio Tuning
		["PULSE_CH_UP"]="CHUP",
		["PULSE_CH_DOWN"]="CHDN",	
		
		--TRANSPORT
		["PLAY"]="PLAY",		
		["STOP"]="STOP",
		["PAUSE"]="PAUSE",	
		["SKIP_FWD"]="TRUP",
		["SKIP_REV"]="TRDN",
		["SCAN_FWD"]="FF",
		["SCAN_REV"]="REW",
		
	}	
	
	CMDS[USBFRONT_PROXY_BINDINGID] = CMDS[NET_PROXY_BINDINGID]
	CMDS[USBREAR_PROXY_BINDINGID] = CMDS[NET_PROXY_BINDINGID]
	CMDS[BLUETOOTH_PROXY_BINDINGID] = CMDS[NET_PROXY_BINDINGID]
	
	EX_CMD["HDMI OUPUT: OFF"] = "HDO00"
	EX_CMD["HDMI OUPUT: MAIN"] = "HDO01"
	EX_CMD["HDMI OUPUT: SUB"] = "HDO02"
	EX_CMD["HDMI OUPUT: BOTH"] = "HDO03"
	EX_CMD["HDMI OUPUT: BOTH (MAIN)"] = "HDO04"
	EX_CMD["HDMI OUPUT: BOTH (SUB)"] = "HDO05"
	
end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	LogTrace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	
	tOutputCommandMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data
		[0] = "SLI",
		[1] = "SLZ",
		[2] = "SL3",
		[3] = "SL4",
	}
	
	tInputCommandMap = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		["INPUT AUX"] = "03",
		["INPUT BD/DVD"] = "10",
		["INPUT CBL/SAT"] = "01",
		["INPUT GAME"] = "02",
		--["INPUT GAME1"] = "02",
		["INPUT GAME2"] = "04",
		["INPUT EXTRA1"] = "07",
		["INPUT EXTRA2"] = "08",
		["INPUT EXTRA3"] = "09",
		["INPUT PC"] = "05",
		["INPUT PHONO"] = "22",
		["INPUT STB/DVR"] = "00",
		["INPUT CD"] = "23",
		["INPUT STRM BOX"] = "11",
		["INPUT TV"] = "12",
		["INPUT TV/TAPE"] = "20",
		["FM Antenna"] = "24",
		["AM Antenna"] = "25",
		["AM/FM TUNER"] = "26",
		["NET"] = "2B",
		--["TUNEIN"] = "2B",	
		--["SIRIUSXM"] = "2B",	
		--["SPOTIFY"] = "2B",
		--["MP3TUNES"] = "2B",
		--["DLNA"] = "2B",		--2B	--27
		--["PANDORA"] = "2B",
		--["LAST.FM"] = "2B",
		--["AUPEO!"] = "2B",
		--["MY FAVORITES"] = "2B",	--2B	--28
		--["HOME MEDIA"] = "2B",
		--["RHAPSODY"] = "2B",
		--["SLACKER"] = "2B",
		--["DEEZER"] = "2B",
		["USB (FRONT)"] = "29",
		["USB (REAR)"] = "2A",
		["BLUETOOTH"] = "2E",
		--["AIRPLAY"] = "2D",		--listed in protocol doc but not avialble on .7 models...
		--["AIRPLAY"] = "2B",
		--["RADIKO"] = "2B",
		--["E-ONKYO"] = "2B",
		--["SIMFY"] = "2B",
		--["TIDAL"] = "2B",
		--["ONKYO MUSIC"] = "2B",	
		["DAB Antenna"] = "33",
		["INPUT HDMI 5"] = "55",
		["INPUT HDMI 6"] = "56",
		["INPUT HDMI 7"] = "57",
		--["FIRECONNECT"] = "2B",	
		--["CHROMECAST"] = "2B",	
	}
	
	tInputResponseMap = ReverseTable(tInputCommandMap)	-- Reverses the tInputCommandMap table	
	
	tInputCommandMap[TUNER_PROXY_BINDINGID] = {
		["AM Antenna"] = "01",
		["FM Antenna"] = "00", 
	}	

	-- The first 2 values are the NET code, 
	-- The last value is a 0 (zero) indicating no account information for the service.
	tNetInputCommandMap = {
		["DLNA"] = "000", 			-- DLNA, Media Server
		["MY FAVORITES"] = "010",	--My Favorites
		["vTuner"] = "020",
		["SIRIUSXM"] = "030", 		-- SiriusXM
		["PANDORA"] = "040",		--PANDORA
		["RHAPSODY"] = "050",		--Rhapsody
		["LAST.FM"] = "060",		--Last.fm
		["Napster"] = "070",
		["SLACKER"] = "080",		--Slacker
		["Mediafly"] = "090",
		["SPOTIFY"] = "0A0",		--Spotify
		["AUPEO!"] = "0B0",			--AUPEO!
		["RADIKO"] = "0C0",
		["E-ONKYO"] = "0D0",
		["TUNEIN"] = "0E0",			--TuneIn
		["MP3TUNES"] = "0F0",		--MP3tunes
		["SIMFY"] = "100",
		["HOME MEDIA"] = "110",		--Home Media
		["DEEZER"] = "120",
		["iHeartRadio"] = "130",
		["AIRPLAY"] = "180",
		["TIDAL"] = "1B0",
		["ONKYO MUSIC"] = "1A0",
		["FIRECONNECT"] = "410",	
		["CHROMECAST"] = "400"
	}	
	
	----------------------------------------- [*I/O HELPER TABLES*] -----------------------------------------
	tOutputConnMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Output Connection Name
		[0] = "Main Output",
		[1] = "Zone2 Output",
		[2] = "Zone3 Output",
		[3] = "Zone4 Output",
	}
	
	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		--[0] = {Name = "INPUT HDMI 1",BindingID = RECEIVER_PROXY_BINDINGID,},
		[0] = {Name = "INPUT STB/DVR", BindingID = RECEIVER_PROXY_BINDINGID},
		[1] = {Name = "INPUT CBL/SAT", BindingID = RECEIVER_PROXY_BINDINGID},
		[2] = {Name = "INPUT GAME", BindingID = RECEIVER_PROXY_BINDINGID},
		[3] = {Name = "INPUT AUX", BindingID = RECEIVER_PROXY_BINDINGID},
		[4] = {Name = "INPUT GAME2", BindingID = RECEIVER_PROXY_BINDINGID},
		[5] = {Name = "INPUT PC", BindingID = RECEIVER_PROXY_BINDINGID},
		[6] = {Name = "INPUT EXTRA1", BindingID = RECEIVER_PROXY_BINDINGID},
		[7] = {Name = "INPUT EXTRA2", BindingID = RECEIVER_PROXY_BINDINGID},
		[8] = {Name = "INPUT EXTRA3", BindingID = RECEIVER_PROXY_BINDINGID},
		[9] = {Name = "INPUT BD/DVD", BindingID = RECEIVER_PROXY_BINDINGID},
		[10] = {Name = "INPUT PHONO", BindingID = RECEIVER_PROXY_BINDINGID},
		[11] = {Name = "INPUT CD", BindingID = RECEIVER_PROXY_BINDINGID},
		[12] = {Name = "AM/FM TUNER", BindingID = RECEIVER_PROXY_BINDINGID},
		[13] = {Name = "AM Antenna", BindingID = TUNER_PROXY_BINDINGID},
		[14] = {Name = "FM Antenna", BindingID = TUNER_PROXY_BINDINGID},
		[15] = {Name = "NET", BindingID = RECEIVER_PROXY_BINDINGID},
		[16] = {Name = "TUNEIN", BindingID = RECEIVER_PROXY_BINDINGID},
		[17] = {Name = "SIRIUSXM", BindingID = RECEIVER_PROXY_BINDINGID},
		[18] = {Name = "SPOTIFY", BindingID = RECEIVER_PROXY_BINDINGID},
		[19] = {Name = "MP3TUNES", BindingID = RECEIVER_PROXY_BINDINGID},
		[20] = {Name = "DLNA", BindingID = RECEIVER_PROXY_BINDINGID},
		[21] = {Name = "PANDORA", BindingID = RECEIVER_PROXY_BINDINGID},
		[22] = {Name = "LAST.FM", BindingID = RECEIVER_PROXY_BINDINGID},
		[23] = {Name = "AUPEO!", BindingID = RECEIVER_PROXY_BINDINGID},
		[24] = {Name = "MY FAVORITES", BindingID = RECEIVER_PROXY_BINDINGID},
		[25] = {Name = "HOME MEDIA", BindingID = RECEIVER_PROXY_BINDINGID},
		[26] = {Name = "RHAPSODY", BindingID = RECEIVER_PROXY_BINDINGID},
		[27] = {Name = "SLACKER", BindingID = RECEIVER_PROXY_BINDINGID},
		[28] = {Name = "USB (FRONT)", BindingID = RECEIVER_PROXY_BINDINGID},
		[29] = {Name = "USB (REAR)", BindingID = RECEIVER_PROXY_BINDINGID},
		[30] = {Name = "BLUETOOTH", BindingID = RECEIVER_PROXY_BINDINGID},
		[31] = {Name = "DEEZER", BindingID = RECEIVER_PROXY_BINDINGID},
		[32] = {Name = "INPUT STRM BOX", BindingID = RECEIVER_PROXY_BINDINGID},
		[33] = {Name = "INPUT TV", BindingID = RECEIVER_PROXY_BINDINGID},
		[34] = {Name = "AIRPLAY", BindingID = RECEIVER_PROXY_BINDINGID},
		[35] = {Name = "RADIKO", BindingID = RECEIVER_PROXY_BINDINGID},
		[36] = {Name = "E-ONKYO", BindingID = RECEIVER_PROXY_BINDINGID},
		[37] = {Name = "SIMFY", BindingID = RECEIVER_PROXY_BINDINGID},
		[38] = {Name = "TIDAL", BindingID = RECEIVER_PROXY_BINDINGID},
		[39] = {Name = "ONKYO MUSIC", BindingID = RECEIVER_PROXY_BINDINGID},
		[40] = {Name = "INPUT TV/TAPE", BindingID = RECEIVER_PROXY_BINDINGID},
		[41] = {Name = "DAB Antenna", BindingID = TUNER_PROXY_BINDINGID},
		[42] = {Name = "INPUT HDMI 5", BindingID = RECEIVER_PROXY_BINDINGID},
		[43] = {Name = "INPUT HDMI 6", BindingID = RECEIVER_PROXY_BINDINGID},
		[44] = {Name = "INPUT HDMI 7", BindingID = RECEIVER_PROXY_BINDINGID},
		[45] = {Name = "FIRECONNECT", BindingID = RECEIVER_PROXY_BINDINGID},
		[46] = {Name = "CHROMECAST", BindingID = RECEIVER_PROXY_BINDINGID},
	}
	
	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--['INPUT HDMI 1'] = {ID = 0,BindingID = RECEIVER_PROXY_BINDINGID,},
		["INPUT AUX"] = {ID = 3, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT BD/DVD"] = {ID = 9, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT CBL/SAT"] = {ID = 1, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT GAME"] = {ID = 2, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT GAME2"] = {ID = 4, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT EXTRA1"] = {ID = 6, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT EXTRA2"] = {ID = 7, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT EXTRA3"] = {ID = 8, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT PC"] = {ID = 5, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT PHONO"] = {ID = 10, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT STB/DVR"] = {ID = 0, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT CD"] = {ID = 11, BindingID = RECEIVER_PROXY_BINDINGID},
		["AM/FM TUNER"] = {ID = 12, BindingID = RECEIVER_PROXY_BINDINGID},
		["AM Antenna"] = {ID = 13, BindingID = TUNER_PROXY_BINDINGID},
		["FM Antenna"] = {ID = 14, BindingID = TUNER_PROXY_BINDINGID},
		["NET"] = {ID = 15, BindingID = RECEIVER_PROXY_BINDINGID},
		["TUNEIN"] = {ID = 16, BindingID = RECEIVER_PROXY_BINDINGID},
		["SIRIUSXM"] = {ID = 17, BindingID = RECEIVER_PROXY_BINDINGID},
		["SPOTIFY"] = {ID = 18, BindingID = RECEIVER_PROXY_BINDINGID},
		["MP3TUNES"] = {ID = 19, BindingID = RECEIVER_PROXY_BINDINGID},
		["DLNA"] = {ID = 20, BindingID = RECEIVER_PROXY_BINDINGID},
		["PANDORA"] = {ID = 21, BindingID = RECEIVER_PROXY_BINDINGID},
		["LAST.FM"] = {ID = 22, BindingID = RECEIVER_PROXY_BINDINGID},
		["AUPEO!"] = {ID = 23, BindingID = RECEIVER_PROXY_BINDINGID},
		["MY FAVORITES"] = {ID = 24, BindingID = RECEIVER_PROXY_BINDINGID},
		["HOME MEDIA"] = {ID = 25, BindingID = RECEIVER_PROXY_BINDINGID},
		["RHAPSODY"] = {ID = 26, BindingID = RECEIVER_PROXY_BINDINGID},
		["SLACKER"] = {ID = 27, BindingID = RECEIVER_PROXY_BINDINGID},
		["USB (FRONT)"] = {ID = 28, BindingID = RECEIVER_PROXY_BINDINGID},
		["USB (REAR)"] = {ID = 29, BindingID = RECEIVER_PROXY_BINDINGID},
		["BLUETOOTH"] = {ID = 30, BindingID = RECEIVER_PROXY_BINDINGID},
		["DEEZER"] = {ID = 31, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT STRM BOX"] = {ID = 32, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT TV"] = {ID = 33, BindingID = RECEIVER_PROXY_BINDINGID},
		["AIRPLAY"] = {ID = 34, BindingID = RECEIVER_PROXY_BINDINGID},
		["RADIKO"] = {ID = 35, BindingID = RECEIVER_PROXY_BINDINGID},
		["E-ONKYO"] = {ID = 36, BindingID = RECEIVER_PROXY_BINDINGID},
		["SIMFY"] = {ID = 37, BindingID = RECEIVER_PROXY_BINDINGID},
		["TIDAL"] = {ID = 38, BindingID = RECEIVER_PROXY_BINDINGID},
		["ONKYO MUSIC"] = {ID = 39, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT TV/TAPE"] = {ID = 40, BindingID = RECEIVER_PROXY_BINDINGID},
		["DAB Antenna"] = {ID = 41, BindingID = TUNER_PROXY_BINDINGID},
		["INPUT HDMI 5"] = {ID = 42, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT HDMI 6"] = {ID = 43, BindingID = RECEIVER_PROXY_BINDINGID},
		["INPUT HDMI 7"] = {ID = 44, BindingID = RECEIVER_PROXY_BINDINGID},
		["FIRECONNECT"] = {ID = 45, BindingID = RECEIVER_PROXY_BINDINGID},
		["CHROMECAST"] = {ID = 46, BindingID = RECEIVER_PROXY_BINDINGID},
	}
	
	tProxyInputNameMapByProxyID = {
		[TUNER_PROXY_BINDINGID] = "AM/FM TUNER", 
	}	
end


function PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()")
	tPowerCommandMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data (Power)
		[0] = "PWR",
		[1] = "ZPW",
		[2] = "PW3",
		[3] = "PW4",
	}
	
end


function PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()")
	tVolumeCommandMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data (Pulse Volume - command prefix/suffix)
		[0] = "MVL",
		[1] = "ZVL",
		[2] = "VL3",
		[3] = "VL4",
	}

	tVolumeSetCommandMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data (Discreet Volume - command prefix/suffix)	
		[0] = "MVL",
		[1] = "ZVL",
		[2] = "VL3",
		[3] = "VL4",
	}
	
	tVolumeQueryMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data (Volume Query)		
		[0] = "MVLQSTN",
		[1] = "ZVLQSTN",
		[2] = "VL3QSTN",
		[3] = "VL4QSTN",	
	}	

	tMuteCommandMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Protocol Command Data (Mute - command prefix/suffix)		
		[0] = "AMT",
		[1] = "ZMT",
		[2] = "MT3",
		[3] = "MT4",
	}	
	
end

function PROTOCOL_DECLARATIONS.SurroundModes()
	--index:  value of Surround Mode id
	--value:  Protocol Command Data

	tSurroundModes = {
		[22637] = "LMD40", -- 5.1ch Surround
		[22625] = "LMD05", -- Game-Action
		[22632] = "LMD0C", -- ALL CH STEREO
		[22714] = "LMD16", -- Audyssey DSX
		[22621] = "LMD01", -- DIRECT 
		[22713] = "LMD15", -- DTS Surround Sensation
		[22641] = "LMD41", -- Dolby EX
		[22743] = "LMDA7", -- Dolby EX + Audyssey DSX
		[22639] = "LMD41", -- Dolby EX/DTS ES
		[22640] = "LMD14", -- Dolby VIRTUAL	--Surround Enhancer
		[22634] = "LMD0E", -- Game-Sports
		[22623] = "LMD03", -- Game-RPG
		[22638] = "LMD13", -- FULL MONO
		[22635] = "LMD0F", -- MONO 
		[22627] = "LMD07", -- MONO MOVIE
		[22646] = "LMD12", -- MULTIPLEX
		[22626] = "LMD06", -- Game-Rock
		[22645] = "LMD26", -- MusicMode
		[22739] = "LMDA3", -- Neo:6 Cinema + Audyssey DSX
		[22727] = "LMD91", -- Neo:6 Cinema DTS Surround Sensation
		[22649] = "LMD82", -- Neo:6 Cinema/Neo:X Cinema     DTS:X/Neural:X
		[22740] = "LMDA4", -- Neo:6 Music + Audyssey DSX
		[22728] = "LMD92", -- Neo:6 Music DTS Surround Sensation
		[22650] = "LMD83", -- Neo:6 Music/Neo:X Music
		[22652] = "LMD85", -- Neo:6/Neo:X THX Cinema  DTS Neural:X THX Cinema
		[22720] = "LMD8A", -- Neo:6/Neo:X THX Games  DTS Neural:X THX Games
		[22722] = "LMD8C", -- Neo:6/Neo:X THX Music  DTS Neural:X THX Music
		[22745] = "LMD9A", -- Neo:X Game
		[22729] = "LMD93", -- Neural Digital Music
		[22654] = "LMDA6", -- Neural Digital Music + Audyssey DSX
		[22655] = "LMD87", -- Neural Surr
		[22656] = "LMDA5", -- Neural Surround + Audussey DSX
		[22723] = "LMD8D", -- Neural THX Cinema
		[22725] = "LMD8F", -- Neural THX Games
		[22724] = "LMD8E", -- Neural THX Music
		[22718] = "LMD88", -- Neural THX/Neural Surround
		[22628] = "LMD08", -- ORCHESTRA
		[22653] = "LMD86", -- PLII/PLIIx Game PLII Game 
		[22647] = "LMD80", -- PLII/PLIIx Movie, Dolby Atmos/Dolby Surround
		[22648] = "LMD81", -- PLII/PLIIx Music PLII Music
		[22651] = "LMD84", -- PLII/PLIIx THX Cinema Dolby Surround THX Cinema
		[22719] = "LMD89", -- PLII/PLIIx THX Games  Dolby Surround THX Games
		[22721] = "LMD8B", -- PLII/PLIIx THX Music  Dolby Surround THX Music
		[22738] = "LMDA2", -- PLIIx/PLII Game + Audyssey DSX
		[22733] = "LMDA0", -- PLIIx/PLII Movie + Audyssey DSX
		[22737] = "LMDA1", -- PLIIx/PLII Music + Audyssey DSX
		[22726] = "LMD90", -- PLIIz Height
		[22730] = "LMD94", -- PLIIz Height + THX Cinema
		[22732] = "LMD96", -- PLIIz Height + THX Games
		[22731] = "LMD95", -- PLIIz Height + THX Music
		[22657] = "LMD97", -- PLIIz Height + THX U2/S2 Cinema
		[22735] = "LMD99", -- PLIIz Height + THX U2/S2 Games
		[22734] = "LMD98", -- PLIIz Height + THX U2/S2 Music
		[22636] = "LMD11", -- PURE AUDIO 
		[22620] = "LMD00", -- STEREO
		[22630] = "LMD0A", -- STUDIO-MIX
		[22622] = "LMD02", -- SURROUND
		[22658] = "LMD2E", -- Sports
		[22659] = "LMD23", -- Stage
		[22660] = "LMD40", -- Straight Decode
		[22633] = "LMD0D", -- THEATER-DIMENSIONAL 
		[22624] = "LMD04", -- THX
		[22642] = "LMD42", -- THX Cinema
		[22716] = "LMD45", -- THX Games
		[22661] = "LMD52", -- THX Games Mode, THX U2/S2/I/S Games
		[22715] = "LMD44", -- THX Music
		[22662] = "LMD51", -- THX Music Mode, THX U2/S2/I/S Music
		[22643] = "LMD43", -- THX Surround EX
		[22644] = "LMD50", -- THX U2/S2/I/S Cinema/Cinema2
		[22631] = "LMD0B", -- TV LOGIC
		[22629] = "LMD09", -- UNPLUGGED
		[22744] = "LMD1F", -- Whole House Mode
		[22746] = "LMDMOVIE", -- Movie Toggle
		[22747] = "LMDMUSIC", -- Music Toggle
		[22748] = "LMDGAME", -- Game Toggle
		[22749] = "LMDAUTO", -- Auto Toggle		--Pioneer models
		[22750] = "LMDSURR", -- Surround Toggle		--Pioneer models
		[22751] = "LMDSTEREO", -- Stereo Toggle		--Pioneer models
		[22752] = "LMDFF", 	-- Auto Surround		--Pioneer models
		[22753] = "LMD17", -- DTS Virｔual:X
	}	

	tSurroundResponses = {
		--index:  Protocol Command/Response Data 
		--value:  {ID = value of Surround Mode id, Name = "Name of Surround Mode"
		--["0001"] = { ID = 23303, Name = "STEREO (cyclic)"},
		["40"] = {ID = 22637, Name = "5.1ch Surround"}, 
		["05"] = {ID = 22625, Name = "Game-Action"}, 
		["0C"] = {ID = 22632, Name = "ALL CH STEREO"}, 
		["16"] = {ID = 22714, Name = "Audyssey DSX"}, 
		["01"] = {ID = 22621, Name = "DIRECT"},  
		["15"] = {ID = 22713, Name = "DTS Surround Sensation"}, 
		["41"] = {ID = 22641, Name = "STEREO (cyclic)"}, 
		["A7"] = {ID = 22743, Name = "Dolby EX"},
		--["41"] = {ID = 23303, Name = "Dolby EX/DTS ES"}, 
		["14"] = {ID = 22640, Name = "Dolby VIRTUAL"}, 	--Surround Enhancer
		["0E"] = {ID = 22634, Name = "Game-Sports"}, 
		["03"] = {ID = 22623, Name = "Game-RPG"}, 
		["13"] = {ID = 22638, Name = "FULL MONO"}, 
		["0F"] = {ID = 22635, Name = "Mono"}, 
		["07"] = {ID = 22627, Name = "Mono Movie"}, 
		["12"] = {ID = 22646, Name = "MULTIPLEX"}, 
		["06"] = {ID = 22626, Name = "Game-Rock"}, 
		["26"] = {ID = 22645, Name = "MusicMode"}, 
		["A3"] = {ID = 22739, Name = "Neo:6 Cinema + Audyssey DSX"}, 
		["91"] = {ID = 22727, Name = "Neo:6 Cinema DTS Surround Sensation"}, 
		["82"] = {ID = 22649, Name = "Neo:6 Cinema DTS:X/Neural:X"}, 
		["A4"] = {ID = 22740, Name = "Neo:6 Music + Audyssey DSX"}, 
		["92"] = {ID = 22728, Name = "Neo:6 Music DTS Surround Sensation"}, 
		["83"] = {ID = 22650, Name = "Neo:6 Music/Neo:X Music"}, 
		["85"] = {ID = 22652, Name = "Neo:6/DTS Neural:X THX Cinema"}, 
		["8A"] = {ID = 22720, Name = "Neo:6/DTS Neural:X THX Games"}, 
		["8C"] = {ID = 22722, Name = "Neo:6/DTS Neural:X THX Music"}, 
		["9A"] = {ID = 22745, Name = "Neo:X Game"}, 
		["93"] = {ID = 22729, Name = "Neural Digital Music"}, 
		["A6"] = {ID = 22654, Name = "Neural Digital Music + Audyssey DSX"}, 
		["87"] = {ID = 22655, Name = "Neural Surr"}, 
		["A5"] = {ID = 22656, Name = "Neural Surround + Audussey DSX"}, 
		["8D"] = {ID = 22723, Name = "Neural THX Cinema"}, 
		["8F"] = {ID = 22725, Name = "Neural THX Games"}, 
		["8E"] = {ID = 22724, Name = "Neural THX Music"}, 
		["88"] = {ID = 22718, Name = "Neural THX/Neural Surround"}, 
		["08"] = {ID = 22628, Name = "ORCHESTRA"}, 
		["86"] = {ID = 22653, Name = "PLII Game"}, 
		["80"] = {ID = 22647, Name = "Dolby Atmos/Dolby Surround"},
		["81"] = {ID = 22648, Name = "PLII Music"}, 
		["84"] = {ID = 22651, Name = "Dolby Surround THX Cinema"}, 
		["89"] = {ID = 22719, Name = "Dolby Surround THX Games"}, 
		["8B"] = {ID = 22721, Name = "Dolby Surround THX Music"}, 
		["A2"] = {ID = 22738, Name = "PLIIx/PLII Game + Audyssey DSX"}, 
		["A0"] = {ID = 22733, Name = "PLIIx/PLII Movie + Audyssey DSX"}, 
		["A1"] = {ID = 22737, Name = "PLIIx/PLII Music + Audyssey DSX"}, 
		["90"] = {ID = 22726, Name = "PLIIz Height"}, 
		["94"] = {ID = 22730, Name = "PLIIz Height + THX Cinema"}, 
		["96"] = {ID = 22732, Name = "PLIIz Height + THX Games"}, 
		["95"] = {ID = 22731, Name = "PLIIz Height + THX Music"}, 
		["97"] = {ID = 22657, Name = "PLIIz Height + THX U2/S2 Cinema"}, 
		["99"] = {ID = 22735, Name = "PLIIz Height + THX U2/S2 Games"}, 
		["98"] = {ID = 22734, Name = "PLIIz Height + THX U2/S2 Music"}, 
		["11"] = {ID = 22636, Name = "Pure Audio"},  
		["00"] = {ID = 22620, Name = "Stereo"}, 
		["0A"] = {ID = 22630, Name = "Studio-Mix"}, 
		["02"] = {ID = 22622, Name = "Surround"}, 
		["2E"] = {ID = 22658, Name = "Sports"}, 
		["23"] = {ID = 22659, Name = "Stage"}, 
		["40"] = {ID = 22660, Name = "Straight Decode"}, 
		["0D"] = {ID = 22633, Name = "Theater-Dimensional"},  
		["04"] = {ID = 22624, Name = "THX"}, 
		["42"] = {ID = 22642, Name = "THX Cinema"}, 
		["45"] = {ID = 22716, Name = "THX Games"}, 
		["52"] = {ID = 22661, Name = "THX S2 Games"},
		["44"] = {ID = 22715, Name = "STEREO (cyclic)"}, 
		["51"] = {ID = 22662, Name = "THX S2 Music"}, 
		["43"] = {ID = 22643, Name = "THX Surround EX"}, 
		["50"] = {ID = 22644, Name = "THX S2 Cinema"}, 
		["0B"] = {ID = 22631, Name = "TV Logic"}, 
		["09"] = {ID = 22629, Name = "UnPlugged"}, 
		["1F"] = {ID = 22744, Name = "Whole House Mode"}, 
		["MOVIE"] = {ID = 22746, Name = "Movie Toggle"},
		["MUSIC"] = {ID = 22747, Name = "Music Toggle"},
		["GAME"] = {ID = 22748, Name = "Game Toggle"},
		["AUTO"] = {ID = 22749, Name = "Auto Toggle"},
		["SURR"] = {ID = 22750, Name = "Surround Toggle"},
		["STEREO"] = {ID = 22751, Name = "Stereo Toggle"},	
		["FF"] = {ID = 22752, Name = "Auto Surround"},
		["17"] = {ID = 22753, Name = "DTS Virｔual:X"},
	}
	
	tSurroundModes_IR = {
		--index:  Pvalue of Surround Mode id
		--value:  IR Code ID
		--[22637] = "LMD40", -- 5.1ch Surround
		--[22625] = "LMD05", -- Game-Action
		--[22632] = "LMD0C", -- ALL CH STEREO
		--[22714] = "LMD16", -- Audyssey DSX
		--[22621] = "LMD01", -- DIRECT 
		--[22713] = "LMD15", -- DTS Surround Sensation
		--[22641] = "LMD41", -- Dolby EX
		--[22743] = "LMDA7", -- Dolby EX + Audyssey DSX
		--[22639] = "LMD41", -- Dolby EX/DTS ES
		--[22640] = "LMD14", -- Dolby VIRTUAL
		--[22634] = "LMD0E", -- Game-Sports
		--[22623] = "LMD03", -- Game-RPG
		--[22638] = "LMD13", -- FULL MONO
		--[22635] = "LMD0F", -- MONO 
		--[22627] = "LMD07", -- MONO MOVIE
		--[22646] = "LMD12", -- MULTIPLEX
		--[22626] = "LMD06", -- Game-Rock
		--[22645] = "LMD26", -- MusicMode
		--[22739] = "LMDA3", -- Neo:6 Cinema + Audyssey DSX
		--[22727] = "LMD91", -- Neo:6 Cinema DTS Surround Sensation
		--[22649] = "LMD82", -- Neo:6 Cinema/Neo:X Cinema
		--[22740] = "LMDA4", -- Neo:6 Music + Audyssey DSX
		--[22728] = "LMD92", -- Neo:6 Music DTS Surround Sensation
		--[22650] = "LMD83", -- Neo:6 Music/Neo:X Music
		--[22652] = "LMD85", -- Neo:6 THX Cinema
		--[22720] = "LMD8A", -- Neo:6 THX Games
		--[22722] = "LMD8C", -- Neo:6 THX Music
		--[22745] = "LMD9A", -- Neo:X Game
		--[22729] = "LMD93", -- Neural Digital Music
		--[22654] = "LMDA6", -- Neural Digital Music + Audyssey DSX
		--[22655] = "LMD87", -- Neural Surr
		--[22656] = "LMDA5", -- Neural Surround + Audussey DSX
		--[22723] = "LMD8D", -- Neural THX Cinema
		--[22725] = "LMD8F", -- Neural THX Games
		--[22724] = "LMD8E", -- Neural THX Music
		--[22718] = "LMD88", -- Neural THX/Neural Surround
		--[22628] = "LMD08", -- ORCHESTRA
		--[22653] = "LMD86", -- PLII/PLIIx Game
		--[22647] = "LMD80", -- PLII/PLIIx Movie, Dolby Atmos/Dolby Surround
		--[22648] = "LMD81", -- PLII/PLIIx Music
		--[22651] = "LMD84", -- PLII/PLIIx THX Cinema
		--[22719] = "LMD89", -- PLII/PLIIx THX Games
		--[22721] = "LMD8B", -- PLII/PLIIx THX Music
		--[22738] = "LMDA2", -- PLIIx/PLII Game + Audyssey DSX
		--[22733] = "LMDA0", -- PLIIx/PLII Movie + Audyssey DSX
		--[22737] = "LMDA1", -- PLIIx/PLII Music + Audyssey DSX
		--[22726] = "LMD90", -- PLIIz Height
		--[22730] = "LMD94", -- PLIIz Height + THX Cinema
		--[22732] = "LMD96", -- PLIIz Height + THX Games
		--[22731] = "LMD95", -- PLIIz Height + THX Music
		--[22657] = "LMD97", -- PLIIz Height + THX U2/S2 Cinema
		--[22735] = "LMD99", -- PLIIz Height + THX U2/S2 Games
		--[22734] = "LMD98", -- PLIIz Height + THX U2/S2 Music
		--[22636] = "LMD11", -- PURE AUDIO 
		--[22620] = "LMD00", -- STEREO
		--[22630] = "LMD0A", -- STUDIO-MIX
		--[22622] = "LMD02", -- SURROUND
		--[22658] = "LMD2E", -- Sports
		--[22659] = "LMD23", -- Stage
		--[22660] = "LMD40", -- Straight Decode
		--[22633] = "LMD0D", -- THEATER-DIMENSIONAL 
		--[22624] = "LMD04", -- THX
		--[22642] = "LMD42", -- THX Cinema
		--[22716] = "LMD45", -- THX Games
		--[22661] = "LMD52", -- THX Games Mode, THX U2/S2/I/S Games
		--[22715] = "LMD44", -- THX Music
		--[22662] = "LMD51", -- THX Music Mode, THX U2/S2/I/S Music
		--[22643] = "LMD43", -- THX Surround EX
		--[22644] = "LMD50", -- THX U2/S2/I/S Cinema/Cinema2
		--[22631] = "LMD0B", -- TV LOGIC
		--[22629] = "LMD09", -- UNPLUGGED
		--[22744] = "LMD1F", -- Whole House Mode
	}	
end

function PROTOCOL_DECLARATIONS.Tuner()
	
	--INT"L TUNING:Global Table
	tFreqStepInfo = {
		["FM50/AM9"] =   { FmMin = 8750, FmMax = 10800, FmSpacing = 5,  AmMin = 531, AmMax = 1611, AmSpacing = 9,  },
		["FM100/AM9"] =  { FmMin = 7600, FmMax = 9000,  FmSpacing = 10, AmMin = 531, AmMax = 1611, AmSpacing = 9,  },
		["FM100/AM10"] = { FmMin = 8750, FmMax = 10800, FmSpacing = 10, AmMin = 530, AmMax = 1710, AmSpacing = 10, },
		["FM200/AM10"] = { FmMin = 8750, FmMax = 10790, FmSpacing = 20, AmMin = 530, AmMax = 1710, AmSpacing = 10, },
	}
	
	tTunerPresetCommandMap = {
		[0] = "PRS",
		[1] = "PRZ",
		[2] = "PR3",
		[3] = "PR4",
	}

	tTunerTuningCommandMap = {
		[0] = "TUN",
		[1] = "TUZ",
		[2] = "TU3",
		[3] = "TU4",
	}	
	
end

function PROTOCOL_DECLARATIONS.NETTUNE()
	tNetTuneCommandMap = {
		[0] = 'NTC',
		[1] = 'NTZ',
		[2] = 'NT3',
		[3] = 'NT4',
	}
	
	tNetTunePresetCommandMap = {
		--[0] = '',	--N/A for main zone
		[1] = 'NPZ',
		[2] = 'NP3',
		[3] = 'NP4',
	}	
	
end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end end)
package.preload['device_messages'] = (function (...)
--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2014 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
TEMPLATE_VERSION.driver = "2014.08.14"

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
	local message, pos
	local etx = string.pack('b',0x1A)
	local pattern = "!1(.-)" .. etx .. "()"
	
	if (gReceiveBuffer:len() > 0) then
		message, pos = string.match(gReceiveBuffer, pattern)
		if (message == nil) then
						--LOG:Info("Do not have a complete message")
						return ""
		end
		gReceiveBuffer = gReceiveBuffer:sub(pos)
	end	
	
	return message
end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)
    LogTrace(string.format("HandleMessage. Message is ==>%s<==", message))

    if (message:sub(message:len()-2) == "N/A") then 
	   LOG:Info("HandleMessage(): N/A response recieved with " .. message:sub(1,3))
	   return
    end 
	
    local name, value, pos
    if (message:sub(1,3) == "TPD") then 
	   name, value, pos = string.match(message, "(TPD)(.+)()")
    else
	   name, value, pos = string.match(message, "^(...)(%w+)()")
    end
	
    name = name or message
    value = value or ""
	
    DispatchMessage(name, value)
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LOG:Info("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		DEV_MSG[MsgKey](MsgData)
	else
		LogTrace("HandleMessage: Unhandled command = " .. MsgKey)
	end
end

-- TODO: Create DEV_MSG functions for all messages to call Notifies

function DEV_MSG.TPD(value)
    LogTrace("DEV_MSG.TPD(), value = " .. value)
    local fahrenheit, celsius = string.match(value, ".-(%d+).-(%d+)")
    print("Current Temperature Data: " .. fahrenheit .. " degrees fahrenheit, " .. celsius .. " degrees celsius")
end


--Input Status Functions
function DEV_MSG.SLI(value)
	LogTrace("DEV_MSG.SLI(), value = " .. value)
	local output = ZONE_MAIN 								--mod 1000 value of Output Connection ID
	local input
	if (value == "2B") then 
	   input = gLastSelectedNetService[output] 
	else 
	   input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	end
	--change AM & FM Antenna to TUNER input
	if (input == tInputConnMapByName["AM Antenna"].ID) or (input == tInputConnMapByName["FM Antenna"].ID) then 
	   input = tInputConnMapByName["AM/FM TUNER"].ID
     end
	local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection
	LOG:Info("INPUT_OUTPUT_CHANGED, input = " .. tInputConnMapByID[input].Name .. ", output = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end

function DEV_MSG.SLZ(value)
	LogTrace("DEV_MSG.SLZ(), value = " .. value)
	local output = ZONE_2 									--mod 1000 value of Output Connection ID
	local input
	if (value == "2B") then 
	   input = gLastSelectedNetService[output] 
	else 
	   input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	end
	--change AM & FM Antenna to TUNER input
	if (input == tInputConnMapByName["AM Antenna"].ID) or (input == tInputConnMapByName["FM Antenna"].ID) then 
	   input = tInputConnMapByName["AM/FM TUNER"].ID
     end
	local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection
	LOG:Info("INPUT_OUTPUT_CHANGED, input = " .. tInputConnMapByID[input].Name .. ", output = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end

function DEV_MSG.SL3(value)
	LogTrace("DEV_MSG.SL3(), value = " .. value)
	local output = ZONE_3									--mod 1000 value of Output Connection ID
	local input
	if (value == "2B") then 
	   input = gLastSelectedNetService[output] 
	else 
	   input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	end
	--change AM & FM Antenna to TUNER input
	if (input == tInputConnMapByName["AM Antenna"].ID) or (input == tInputConnMapByName["FM Antenna"].ID) then 
	   input = tInputConnMapByName["AM/FM TUNER"].ID
     end
	local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection
	LOG:Info("INPUT_OUTPUT_CHANGED, input = " .. tInputConnMapByID[input].Name .. ", output = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end	

function DEV_MSG.SL4(value)
	LogTrace("DEV_MSG.SL4(), value = " .. value)
	local output = ZONE_4									--mod 1000 value of Output Connection ID
	local input
	if (value == "2B") then 
	   input = gLastSelectedNetService[output] 
	else 
	   input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	end
	--change AM & FM Antenna to TUNER input
	if (input == tInputConnMapByName["AM Antenna"].ID) or (input == tInputConnMapByName["FM Antenna"].ID) then 
	   input = tInputConnMapByName["AM/FM TUNER"].ID
     end
	local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection
	LOG:Info("INPUT_OUTPUT_CHANGED, input = " .. tInputConnMapByID[input].Name .. ", ouput = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end	

--Power Status Functions
function DEV_MSG.PWR(value)
	LogTrace("DEV_MSG.PWR(), value = " .. value)
	local output = ZONE_MAIN
	if (value == '01') then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== '00') then
		gReceiverProxy:dev_PowerOff(output)
	end		
end
function DEV_MSG.ZPW(value)
	LogTrace("DEV_MSG.ZPW(), value = " .. value)
	local output = ZONE_2
	if (value == '01') then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== '00') then
		gReceiverProxy:dev_PowerOff(output)
	end	
end
function DEV_MSG.PW3(value)
	LogTrace("DEV_MSG.PW3(), value = " .. value)
	local output = ZONE_3
	if (value == '01') then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== '00') then
		gReceiverProxy:dev_PowerOff(output)
	end	
end
function DEV_MSG.PW4(value)
	LogTrace("DEV_MSG.PW4(), value = " .. value)
	local output = ZONE_4
	if (value == '01') then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== '00') then
		gReceiverProxy:dev_PowerOff(output)
	end	
end

--Volume Status Functions	
function DEV_MSG.MVL(value)
    if (tonumber(value, 16) == nil) then
	   LogTrace("DEV_MSG.MVL(), value not valid, exiting...")
	   return
    end
    LogTrace("DEV_MSG.MVL(), value = " .. value)
    local output = ZONE_MAIN
			
    --device message is in  hex, convert to decimal here
    local deviceLevel = tonumber(value, 16)  
	
    --c4 volume level represented as percentage value
    local c4Level
    if (MAX_DEVICE_LEVEL ~= 100) then
	   --level = ProcessVolumeLevel(level,0,MAX_DEVICE_LEVEL,0,100)
	   c4Level = ConvertVolumeToC4(deviceLevel, 0, MAX_DEVICE_LEVEL)
    else
	   c4Level = deviceLevel
    end	
    gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
end
function DEV_MSG.ZVL(value)
	if (tonumber(value, 16) == nil) then
		LogTrace("DEV_MSG.ZVL(), value not valid, exiting...")
		return
	end
	LogTrace("DEV_MSG.ZVL(), value = " .. value)
	local output = ZONE_2
			
    --device message is in  hex, convert to decimal here
    local deviceLevel = tonumber(value, 16)  
	
    --c4 volume level represented as percentage value
    local c4Level
    if (MAX_DEVICE_LEVEL ~= 100) then
	   --level = ProcessVolumeLevel(level,0,MAX_DEVICE_LEVEL,0,100)
	   c4Level = ConvertVolumeToC4(deviceLevel, 0, MAX_DEVICE_LEVEL)
    else
	   c4Level = deviceLevel
    end	
    gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
end	
function DEV_MSG.VL3(value)
	if (tonumber(value, 16) == nil) then
		LogTrace("DEV_MSG.VL3(), value not valid, exiting...")
		return
	end
	LogTrace("DEV_MSG.VL3(), value = " .. value)
	local output = ZONE_3
			
    --device message is in  hex, convert to decimal here
    local deviceLevel = tonumber(value, 16)  
	
    --c4 volume level represented as percentage value
    local c4Level
    if (MAX_DEVICE_LEVEL ~= 100) then
	   --level = ProcessVolumeLevel(level,0,MAX_DEVICE_LEVEL,0,100)
	   c4Level = ConvertVolumeToC4(deviceLevel, 0, MAX_DEVICE_LEVEL)
    else
	   c4Level = deviceLevel
    end	
    gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
end	
function DEV_MSG.VL4(value)
    if (tonumber(value, 16) == nil) then
	   LogTrace("DEV_MSG.VL4(), value not valid, exiting...")
	   return
    end
    LogTrace("DEV_MSG.VL4(), value = " .. value)
    local output = ZONE_4
			
    --device message is in  hex, convert to decimal here
    local deviceLevel = tonumber(value, 16)  
	
    --c4 volume level represented as percentage value
    local c4Level
    if (MAX_DEVICE_LEVEL ~= 100) then
	   --level = ProcessVolumeLevel(level,0,MAX_DEVICE_LEVEL,0,100)
	   c4Level = ConvertVolumeToC4(deviceLevel, 0, MAX_DEVICE_LEVEL)
    else
	   c4Level = deviceLevel
    end	
    gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
end	

--Mute Status Functions
function DEV_MSG.AMT(value)
    LogTrace("DEV_MSG.AMT(), value = " .. value)
    local output = ZONE_MAIN
    local state   						--Mute state represented as "True" or "False"
    if (value == "01") then 
	   state = "True"
    elseif (value == "00") then
	   state = "False"
    else
	   LOG:Info("DEV_MSG.AMT(), value not valid, exiting...")
	   return
    end
    gReceiverProxy:dev_MuteChanged(output, state)
end	
function DEV_MSG.ZMT(value)
    LogTrace("DEV_MSG.ZMT(), value = " .. value)
    local output = ZONE_2
    local state   						--Mute state represented as "True" or "False"
    if (value == "01") then 
	   state = "True"
    elseif (value == "00") then
	   state = "False"
    else
	   LOG:Info("DEV_MSG.ZMT(), value not valid, exiting...")
	   return
    end
    gReceiverProxy:dev_MuteChanged(output, state)
end	
function DEV_MSG.MT3(value)
    LogTrace("DEV_MSG.MT3(), value = " .. value)
    local output = ZONE_3
    local state   						--Mute state represented as "True" or "False"
    if (value == "01") then 
	   state = "True"
    elseif (value == "00") then
	   state = "False"
    else
	   LOG:Info("DEV_MSG.MT3(), value not valid, exiting...")
	   return
    end
    gReceiverProxy:dev_MuteChanged(output, state)
end		
function DEV_MSG.MT4(value)
    LogTrace("DEV_MSG.MT4(), value = " .. value)
    local output = ZONE_4
    local state   						--Mute state represented as "True" or "False"
    if (value == "01") then 
	   state = "True"
    elseif (value == "00") then
	   state = "False"
    else
	   LOG:Info("DEV_MSG.MT4(), value not valid, exiting...")
	   return
    end
    gReceiverProxy:dev_MuteChanged(output, state)
end	

--AM/FM Tuner Status
function DEV_MSG.TUN(value)
    LogTrace("DEV_MSG.TUN(), value = " .. value)
    send_tuner_status(value)
end
function DEV_MSG.TUZ(value)
	LogTrace("DEV_MSG.TUZ(), value = " .. value)
    send_tuner_status(value)
end	
function DEV_MSG.TU3(value)
    LogTrace("DEV_MSG.TU3(), value = " .. value)
    send_tuner_status(value)
end		
function DEV_MSG.TU4(value)
    LogTrace("DEV_MSG.TU4(), value = " .. value)
    send_tuner_status(value)
end		

function send_tuner_status(value)
    local channel = value
    local numChannel = tonumber(value)
    local input = get_input_from_channel(numChannel)
    gReceiverProxy:dev_TunerStatus(input, channel)
end

function get_input_from_channel(numChannel)
    local input
    if  numChannel > 1760 then
	   input = tInputConnMapByName["FM Antenna"].ID + 3000 --3000 Audio, 1000 Video
    else
	   input = tInputConnMapByName["AM Antenna"].ID + 3000 --3000 Audio, 1000 Video
    end	
    return input
end

--Integra 12 Volt Trigger Status
function DEV_MSG.TGA(value)
	local state   						
	if (value == "01") then 
		state = "ON"
	elseif (value == "00") then
		state = "OFF"
	else
		return
	end
	LOG:Info("DEV_MSG.TGA(), Trigger A is " .. state)
end	
function DEV_MSG.TGB(value)
	local state   						
	if (value == "01") then 
		state = "ON"
	elseif (value == "00") then
		state = "OFF"
	else
		return
	end
	LOG:Info("DEV_MSG.TGB(), Trigger B is " .. state)
end		
function DEV_MSG.TGC(value)
	local state   						
	if (value == "01") then 
		state = "ON"
	elseif (value == "00") then
		state = "OFF"
	else
		return
	end
	LOG:Info("DEV_MSG.TGC(), Trigger C is " .. state)
end	

--Surround Mode 
function DEV_MSG.LMD(value)
    LogTrace("DEV_MSG.LMD(), value = " .. value .. " ID = " .. tSurroundResponses[value].ID .. ", name = " .. tSurroundResponses[value].Name)  
    local mode_id = tSurroundResponses[value].ID
    local output = 0
    gReceiverProxy:dev_SurroundModeChanged(output, mode_id)
end		
 end)
package.preload['device_specific_commands'] = (function (...)
-- This macro is utilized to identify the version string of the driver template version used.
TEMPLATE_VERSION.driver = "2014.08.14"

--[[
	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
	ExecuteCommand Code
	
	Define any functions for device specific commands (EX_CMD.<command>) received from ExecuteCommand that need to be handled by the driver.
	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--]]

--function EX_CMD.NEW_COMMAND(tParams)
--	LogTrace("EX_CMD.NEW_COMMAND")
--	LogTrace(tParams)
--end

--[[
	Device Command: ZonePower
	Parameters:
		Power: Power state: values are "ON" and "OFF"
		Zone: zone id (1 based, must be corrected for zero based connection ids)
--]]
function EX_CMD.ZonePower(tParams)
    LogTrace("EX_CMD.ZonePower()")
    LogTrace(tParams)
    if(tParams["Power"] == "ON") then
	   local output = tonumber(tParams["Zone"]) - 1
	   --we should clear gPowerState for this zone since it may be out of sync, 
	   --(i.e. the user may be firing this command to correct a problem where the the unit is off when it should be on)
	   gReceiverProxy._PowerState[output] = ""		
	   PRX_CMD.CONNECT_OUTPUT(RECEIVER_BINDING_ID, {OUTPUT = output})
    else
	   local output = tonumber(tParams["Zone"]) - 1	
	   PRX_CMD.DISCONNECT_OUTPUT(RECEIVER_BINDING_ID, {OUTPUT = output})
    end
end

function EX_CMD.HDMIOutputSelector(tParams)
    LogTrace("EX_CMD.HDMIOutputSelector()")
    LogTrace(tParams)
    local command
    if(tParams["Mode"] == "OFF") then 
	   command = "HDMI OUPUT: OFF"
    elseif(tParams["Mode"] == "MAIN") then 
	   command = "HDMI OUPUT: MAIN"
    elseif(tParams["Mode"] == "HDBaseT/SUB") then 
	   command = "HDMI OUPUT: SUB"
    elseif(tParams["Mode"] == "BOTH") then 
	   command = "HDMI OUPUT: BOTH"
    elseif(tParams["Mode"] == "BOTH (MAIN)") then 
	   command = "HDMI OUPUT: BOTH (MAIN)"
    elseif(tParams["Mode"] == "BOTH (SUB)") then 
	   command = "HDMI OUPUT: BOTH (SUB)"
    end
    if (gControlMethod == "IR") then
	   if (EX_CMD_IR[command] ~= nil) then
		  PackAndQueueCommand("HDMIOutputSelector", EX_CMD_IR[command])
	   end	
    else
	   PackAndQueueCommand("HDMIOutputSelector", EX_CMD[command])
    end		
end

function EX_CMD.SpeakerLayoutCommand(tParams)
    LogTrace("EX_CMD.SpeakerLayoutCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Surround Back") then 
		  command = "SPLSB"
	   elseif(tParams["Mode"] == "Front High") then
		  command = "SPLFH"
	   elseif(tParams["Mode"] == "Front Wide") then
		  command = "SPLFW"
	   elseif(tParams["Mode"] == "Front High/Front Wide") then
		  command = "SPLHW"	
	   elseif(tParams["Mode"] == "Height1") then
		  command = "SPLH1"	
	   elseif(tParams["Mode"] == "Height2") then
		  command = "SPLH2"			  
	   end
	   PackAndQueueCommand("SpeakerLayoutCommand", command)
    end		
end

function EX_CMD.AudioSelectorCommand(tParams)
    LogTrace("EX_CMD.AudioSelectorCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Analog") then 
			command = "SLA02"
	   elseif(tParams["Mode"] == "HDMI") then
			command = "SLA04"
	   elseif(tParams["Mode"] == "COAX/OPT") then
			command = "SLA05"
	   elseif(tParams["Mode"] == "BALANCE") then
			command = "SLA06"
	   elseif(tParams["Mode"] == "ARC") then
			command = "SLA07"		
	   end
	   PackAndQueueCommand("AudioSelectorCommand", command)
	end		
end

function EX_CMD.HDMIAudioOutMAIN(tParams)
    LogTrace("EX_CMD.HDMIAudioOutMAIN()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "HAO00"
	   elseif(tParams["Mode"] == "On") then
		  command = "HAO01"
	   elseif(tParams["Mode"] == "Auto") then
		  command = "HAO02"		
	   end
	   PackAndQueueCommand("HDMIAudioOutMAIN", command)
    end	
end

function EX_CMD.HDMIAudioOutSUB(tParams)
    LogTrace("EX_CMD.HDMIAudioOutSUB()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "HAS00"
	   elseif(tParams["Mode"] == "On") then
		  command = "HAS01"	
	   end
	   PackAndQueueCommand("HDMIAudioOutSUB", command)
    end	
end

function EX_CMD.HDMICEC(tParams)
    LogTrace("EX_CMD.HDMICEC()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "CEC00"
	   elseif(tParams["Mode"] == "On") then
		  command = "CEC01"
	   end
	   PackAndQueueCommand("HDMICEC", command)
    end	
end

function EX_CMD.VideoPictureMode(tParams)
    LogTrace("EX_CMD.VideoPictureMode()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "ISF Day") then 
		  command = "VPM05"
	   elseif(tParams["Mode"] == "ISF Night") then
		  command = "VPM06"	
	   end
	   PackAndQueueCommand("VideoPictureMode", command)
    end	
end

function EX_CMD.Audyssey(tParams)
    LogTrace("EX_CMD.Audyssey()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "ADY00"
	   elseif(tParams["Mode"] == "On/Movie") then
		  command = "ADY01"
	   elseif(tParams["Mode"] == "Music") then
		  command = "ADY02"		
	   end
	   PackAndQueueCommand("Audyssey", command)
    end	
end

function EX_CMD.AudysseyDynamicEQ(tParams)
    LogTrace("EX_CMD.AudysseyDynamicEQ()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "ADQ00"
	   elseif(tParams["Mode"] == "On") then
		  command = "ADQ01"	
	   end
	   PackAndQueueCommand("AudysseyDynamicEQ", command)
    end	
end

function EX_CMD.AudysseyDynamicVolume(tParams)
    LogTrace("EX_CMD.AudysseyDynamicVolume()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "ADV00"
	   elseif(tParams["Mode"] == "On") then
		  command = "ADV01"	
	   end
	   PackAndQueueCommand("AudysseyDynamicVolume", command)
    end	
end

function EX_CMD.AccuEQ(tParams)
    LogTrace("EX_CMD.AccuEQ()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "AEQ00"
	   elseif(tParams["Mode"] == "On") then
		  command = "AEQ01"	
	   end
	   PackAndQueueCommand("AccuEQ", command)
    end	
end

function EX_CMD.MonitorOutResolution(tParams)
    LogTrace("EX_CMD.MonitorOutResolution()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Through") then 
		  command = "RES00"
	   elseif(tParams["Mode"] == "Auto(HDMI Output Only)") then
		  command = "RES01"	
	   elseif(tParams["Mode"] == "480p") then
		  command = "RES02"
	   elseif(tParams["Mode"] == "720p") then
		  command = "RES03"	
	   elseif(tParams["Mode"] == "1080i") then
		  command = "RES04"	
	   elseif(tParams["Mode"] == "1080p(HDMI Output Only)") then
		  command = "RES05"	
	   elseif(tParams["Mode"] == "4K Upcaling(HDMI Output Only)") then
		  command = "RES08"	
	   elseif(tParams["Mode"] == "Source") then
		  command = "RES06"				
	   end
	   PackAndQueueCommand("MonitorOutResolution", command)
    end	
end

function EX_CMD.VideoWideMode(tParams)
    LogTrace("EX_CMD.VideoWideMode()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Auto") then 
		  command = "VWM00"
	   elseif(tParams["Mode"] == "4:3") then
		  command = "VWM01"	
	   elseif(tParams["Mode"] == "Full") then
		  command = "VWM02"
	   elseif(tParams["Mode"] == "Zoom") then
		  command = "VWM03"	
	   elseif(tParams["Mode"] == "Wide Zoom") then
		  command = "VWM04"	
	   end
	   PackAndQueueCommand("VideoWideMode", command)
    end	
end

function EX_CMD.TriggerA(tParams)
    LogTrace("EX_CMD.TriggerA()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "TGA00"
	   elseif(tParams["Mode"] == "On") then
		  command = "TGA01"	
	   end
	   PackAndQueueCommand("TriggerA", command)
    end	
end

function EX_CMD.TriggerB(tParams)
    LogTrace("EX_CMD.TriggerB()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "TGB00"
	   elseif(tParams["Mode"] == "On") then
		  command = "TGB01"	
	   end
	   PackAndQueueCommand("TriggerB", command)
    end	
end

function EX_CMD.TriggerC(tParams)
    LogTrace("EX_CMD.TriggerC()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "TGC00"
	   elseif(tParams["Mode"] == "On") then
		  command = "TGC01"	
	   end
	   PackAndQueueCommand("TriggerC", command)
    end	
end

function EX_CMD.WholeHouseMode(tParams)
    LogTrace("EX_CMD.WholeHouseMode()")
    if (gControlMethod ~= "IR") then
	   local command = "LMD1F"
	   PackAndQueueCommand("WholeHouseMode", command)
    end	
end

function EX_CMD.SpeakerACommand(tParams)
    LogTrace("EX_CMD.SpeakerACommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "SPA00"
	   elseif(tParams["Mode"] == "On") then
		  command = "SPA01"	
	   elseif(tParams["Mode"] == "Toggle") then
		  command = "SPAUP"	
	   end
	   PackAndQueueCommand("SpeakerACommand", command)
    end	
end

function EX_CMD.SpeakerBCommand(tParams)
    LogTrace("EX_CMD.SpeakerBCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "SPB00"
	   elseif(tParams["Mode"] == "On") then
		  command = "SPB01"	
	   elseif(tParams["Mode"] == "Toggle") then
		  command = "SPBUP"	
	   end
	   PackAndQueueCommand("SpeakerBCommand", command)
    end	
end

function EX_CMD.Zone2SpeakerACommand(tParams)
    LogTrace("EX_CMD.Zone2SpeakerACommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "ZPA00"
	   elseif(tParams["Mode"] == "On") then
		  command = "ZPA01"	
	   end
	   PackAndQueueCommand("Zone2SpeakerACommand", command)
    end	
end

function EX_CMD.Zone2SpeakerBCommand(tParams)
    LogTrace("EX_CMD.Zone2SpeakerBCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "ZPB00"
	   elseif(tParams["Mode"] == "On") then
		  command = "ZPB01"	
	   end
	   PackAndQueueCommand("Zone2SpeakerBCommand", command)
    end	
end

function EX_CMD.BassToneFrontCommand(tParams)
    LogTrace("EX_CMD.BassToneFrontCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tonumber(tParams.Level)
	   if(level == -10) then 
		  level = "-A"
	   elseif(level == 10) then 
		  level = "+A"
	   elseif(level == 0) then 
		  level = "00"
	   elseif(level < 0) then 
		  level = tostring(level)
	   else
		   level = "+" .. tostring(level)
	   end
	   local command = "TFRB" .. level
	   PackAndQueueCommand("BassToneFrontCommand", command)
    end	
end

function EX_CMD.TrebleToneFrontCommand(tParams)
    LogTrace("EX_CMD.TrebleToneFrontCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tonumber(tParams.Level)
	   if(level == -10) then 
		  level = "-A"
	   elseif(level == 10) then 
		  level = "+A"
	   elseif(level == 0) then 
		  level = "00"
	   elseif(level < 0) then 
		  level = tostring(level)
	   else
		   level = "+" .. tostring(level)
	   end
	   local command = "TFRT" .. level
	   PackAndQueueCommand("TrebleToneFrontCommand", command)
    end	
end

function EX_CMD.SleepTimerSetCommand(tParams)
    LogTrace("EX_CMD.SleepTimerSetCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = string.format("%02X", tParams.Minutes)
	   local command = "SLP" .. level
	   PackAndQueueCommand("SleepTimerSetCommand", command)
    end	
end

function EX_CMD.SleepTimerOffCommand()
    LogTrace("EX_CMD.SleepTimerOffCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command = "SLPOFF"
	   PackAndQueueCommand("SleepTimerOffCommand", command)
    end	
end

function EX_CMD.SubwooferLevelCommand(tParams)
    LogTrace("EX_CMD.SubwooferLevelCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Level
	   if(level == "-15.0dB") then 
		  level = "-0F"
	   elseif(level == "-0.0dB") then 
		  level = "000"
	   elseif(level == "+12.0dB") then 
		  level = "+0C"
	   else
		   level = "UP"	--Wrap Around Up
	   end
	   local command = "SWL" .. level
	   PackAndQueueCommand("SubwooferLevelCommand", command)
    end	
end

function EX_CMD.DimmerLevelCommand(tParams)
    LogTrace("EX_CMD.DimmerLevelCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Level
	   if(level == "Bright") then 
		  level = "00"
	   elseif(level == "Dim") then 
		  level = "01"
	   elseif(level == "Dark") then 
		  level = "02"
	   else
		   level = "DIM"	--Wrap Around Up
	   end
	   local command = "DIM" .. level
	   PackAndQueueCommand("DimmerLevelCommand", command)
    end	
end

function EX_CMD.LateNightCommand(tParams)
    LogTrace("EX_CMD.LateNightCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Level
	   if(level == "Off") then 
		  level = "00"
	   elseif(level == "Low@DolbyDigital,On@Dolby TrueHD") then 
		  level = "01"
	   elseif(level == "Auto@Dolby TrueHD") then 
		  level = "03"
	   else
		   level = "UP"	--Wrap Around Up
	   end
	   local command = "LTN" .. level
	   PackAndQueueCommand("LateNightCommand", command)
    end	
end

function EX_CMD.EqualizerSelect(tParams)
    LogTrace("EX_CMD.EqualizerSelect()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Preset
	   if(level == "Off") then 
		  level = "00"
	   elseif(level == "Preset 1") then 
		  level = "01"
	   elseif(level == "Preset 2") then 
		  level = "02"
	   elseif(level == "Preset 3") then 
		  level = "03"
	   elseif(level == "Wrap-Around Up") then 
		  level = "UP"
	   else
		   level = "DOWN"	--Wrap Around Down
	   end
	   local command = "EQS" .. level
	   PackAndQueueCommand("EqualizerSelect", command)
    end	
end

function EX_CMD.EQforStandingWave(tParams)
    LogTrace("EX_CMD.EQforStandingWave()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Mode
	   if(level == "Off") then 
		  level = "00"
	   elseif(level == "On") then 
		  level = "01"
	   else
		   level = "UP"	--Wrap Around Up
	   end
	   local command = "STW" .. level
	   PackAndQueueCommand("EQforStandingWave", command)
    end	
end

function EX_CMD.MusicOptimizerSoundRetriever(tParams)
    LogTrace("EX_CMD.MusicOptimizerSoundRetriever()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Mode
	   if(level == "Off") then 
		  level = "00"
	   elseif(level == "On") then 
		  level = "01"
	   else
		   level = "UP"	--Wrap Around Up
	   end
	   local command = "MOT" .. level
	   PackAndQueueCommand("MusicOptimizerSoundRetriever", command)
    end	
end

function EX_CMD.ChangeAVSync(tParams)
    LogTrace("EX_CMD.ChangeAVSync()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = tParams.Mode
	   if (level == "Increase") then 
		  level = "UP"
	   elseif (level == "Decrease") then 
		  level = "DOWN"
	   else
		   level = "UP"	--Wrap Around Up
	   end
	   local command = "AVS" .. level
	   PackAndQueueCommand("ChangeAVSync", command)
    end	
end

function EX_CMD.IMAX(tParams)
    LogTrace("EX_CMD.IMAX()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local command
	   if(tParams["Mode"] == "Off") then 
		  command = "IMX00"
	   elseif(tParams["Mode"] == "On") then
		  command = "IMX01"	
	   elseif(tParams["Mode"] == "Auto") then
		  command = "IMX02"	
	   end
	   PackAndQueueCommand("IMAX", command)
    end	
end

function EX_CMD.VocalDialogCommand(tParams)
    LogTrace("EX_CMD.VocalDialogCommand()")
    LogTrace(tParams)
    if (gControlMethod ~= "IR") then
	   local level = string.format("<%02d>", tParams.Level)
	   local command = "VOC" .. level
	   PackAndQueueCommand("VocalDialogCommand", command)
    end	
end end)
package.preload['actions'] = (function (...)
-- This macro is utilized to identify the version string of the driver template version used.
TEMPLATE_VERSION.driver = "2014.08.14"

--[[
	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
	Lua Action Code
	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--]]

-- TODO: Create a function for each action defined in the driver

function LUA_ACTION.TemplateVersion()
	TemplateVersion()
end

function LUA_ACTION.DisplayGlobals()
	print ("Global Variables")
	print ("----------------------------")

	for k,v in pairs(_G) do	-- globals`
		if not (type(v) == "function") then
			--print(k .. ":  " .. tostring(v))
			if (string.find(k, "^g%L")  == 1) then
				print(k .. ":  " .. tostring(v))
				if (type(v) == "table") then
					PrintTable(v, "   ")
				end
			end
		end
	end

	print ("")
end

function LUA_ACTION.PrintTemperatureData()
    LogTrace("Received ACTION Print Temperature Data")
    PackAndQueueCommand("Print Temperature Data", "TPDQSTN")
end

function LUA_ACTION.MemorySetup(tParams)
    local command = tParams.Command
    LogTrace("Received ACTION Memory Setup Command: " .. command)
    
    local cmd = ""
    if (command == "Recall Memory") then
	   cmd = "RCL"
    elseif (command == "Lock Memory") then
	   cmd = "LOCK"
    elseif (command == "Unlock Memory") then
	   cmd = "UNLK"
    end
    
    PackAndQueueCommand("Memory Setup: " .. command, "MEM" .. cmd)
end

function LUA_ACTION.PrintVolumeCurve()
    print("===== Volume Curve =====")
    for j,k in pairs(tVolumeCurve) do
	   print(j,k)
    end
end

function LUA_ACTION.ACTION_SetDeviceVisibility(tParams)
    LogTrace("Received ACTION Set Device Visibility")
    LogTrace(tParams)

    if (gProxyNameToBindingIdMap[tParams.Proxy] ~= nil) then
      local proxyBindingID = gProxyNameToBindingIdMap[tParams.Proxy]
      set_proxy_visbility(proxyBindingID, tParams.Action)
    else
      LogWarn("ACTION_SetDeviceVisibility(), the " .. tParams.Proxy .. " proxy is not included in this driver!")
    end
end

function PrintTable(tValue, sIndent)
	sIndent = sIndent or "   "
	for k,v in pairs(tValue) do
		print(sIndent .. tostring(k) .. ":  " .. tostring(v))
		if (type(v) == "table") then
			PrintTable(v, sIndent .. "   ")
		end
	end
end
 end)
package.preload['common.c4_url_connection'] = (function (...)
--[[=============================================================================
    Base for a url connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_url_connection = "2016.01.08"
end

UrlConnectionBase = inheritsFrom(DeviceConnectionBase)

function UrlConnectionBase:construct(Url)
	self.superClass():construct()
	self._Url = Url
end

function UrlConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "URL"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	OnURLConnectionChanged()
end

function UrlConnectionBase:ControlMethod()
	return "URL"
end

function UrlConnectionBase:SetUrl(Url)
	self._Url = Url
end

function UrlConnectionBase:SendCommand(sCommand, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(self._Url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(self._Url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:SendCommandUrl(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlPost(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlGet(url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlGet(url, sHeader)
		else
			ticketId = C4:urlGet(url)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:ReceivedAsync(ticketId, sData, responseCode, tHeaders)
	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(sData))
	LogTrace(tHeaders)
	local tMessage = {
		["ticketId"] = ticketId,
		["sData"] = sData,
		["responseCode"] = responseCode,
		["tHeaders"] = tHeaders
	}
	
	status, err = pcall(HandleMessage, tMessage)
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

function ConnectURL()
	gIsUrlConnected = true
	SetControlMethod()
end

function DisconnectURL()
	gIsUrlConnected = false
	SetControlMethod()
end
 end)
package.preload['connections'] = (function (...)
--[[=============================================================================
    Functions for managing the status of the drivers bindings and connection state
 
    Copyright 2014 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_url_connection"

-- constants
COM_USE_ACK = false
COM_COMMAND_DELAY_MILLISECONDS = 250
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4

NETWORK_PORT = 60128
IR_BINDING_ID = 2
SERIAL_BINDING_ID = 1
NETWORK_BINDING_ID = 6000

function ON_DRIVER_LATEINIT.connections()
    local xml = C4:GetDriverConfigInfo("connections")
    for c in string.gmatch(xml, '<connection>(.-)</connection>') do
	   local id, port = string.match(c, '<id>(%d+)</id>.-<classname>TCP</classname>.-<number>(%d+)</number>')
	   if id ~= nil and port ~= nil then
		  NETWORK_BINDING_ID = id
		  NETWORK_PORT	= port
		  print(NETWORK_BINDING_ID, NETWORK_PORT)
		  --return
	   end
    end
  
    GetMACAddress()
    GetIPAddress()

end

function GetMACAddress()
    if (LuaC4Object.GetDeviceMAC) then
	   local mac = C4:GetDeviceMAC(NETWORK_BINDING_ID) or ""
	   UpdateProperty("MAC Address", mac)
    end
end

function GetIPAddress()
    if (LuaC4Object.GetBindingAddress) then
	   local addr = C4:GetBindingAddress(NETWORK_BINDING_ID) or ""
	   UpdateProperty("IP Address", addr)
    end
end

--[[=============================================================================
    OnSerialConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when a serial binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnSerialConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnIRConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when an IR binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnIRConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnNetworkConnectionChanged(idBinding, bIsBound)
  
    Description:
    Function called when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkConnectionChanged(idBinding, bIsBound)
    if (bIsBound) then
	   GetMACAddress()
	   GetIPAddress()
    end
end

--[[=============================================================================
    OnNetworkStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Called when the network connection status changes. Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnNetworkStatusChanged(idBinding, nPort, sStatus)
	
end

--[[=============================================================================
    OnURLConnectionChanged(url)
  
    Description:
    Function called when the c4_url_connection is created.
  
    Parameters:
    url - url used by the url connection.
  
    Returns:
    None
===============================================================================]]
function OnURLConnectionChanged(url)
	
end

--[[=============================================================================
    DoEvents()
  
    Description:
    sda
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function DoEvents()
	if (gReceiverProxy._VolumeIsRamping == true) then
		for j,k in pairs(gReceiverProxy._VolumeRamping) do
			if (k.state == true) then gReceiverProxy:ContinueVolumeRamping(j) end
		end
	end
end

--[[=============================================================================
    SendKeepAlivePollingCommand()
  
    Description:
    Sends a driver specific polling command to the connected system
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function SendKeepAlivePollingCommand()
    LogTrace("SendKeepAlivePollingCommand()")
	
    command = "DIMQSTN"
    LogTrace("command = " .. command)
    PackAndQueueCommand("SendKeepAlivePollingCommand", command)		
end end)
package.preload['common.c4_utils'] = (function (...)
--[[=============================================================================
    Helper functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_utils = "2016.01.08"
end

--[[=============================================================================
    AsciiToBCD(InString)

    Description
    Convert an ascii string to a binary coded decimal. Each decimal digit is
    stored in one byte, with the lower four bits encoding the digit in BCD form.

    Parameters
    InString(string) - Ascii string that is to be converted into bcd

    Returns
    The binary coded decimal
===============================================================================]]
function AsciiToBCD(InString)
	local WorkVal = 0
	local RetValStr = ""
	local DoingHighNybble = false
	local WorkStr = ((#InString % 2) == 0) and (InString) or ("0" .. InString)	-- make sure length is an even number

	for CharCount = 1, #WorkStr do
		local NumVal = tonumber(WorkStr:sub(CharCount, CharCount))

		WorkVal = bit.lshift(WorkVal, 4) + NumVal
		if (DoingHighNybble) then
			RetValStr = RetValStr .. string.char(WorkVal)
			WorkVal = 0
		end

		DoingHighNybble = (not DoingHighNybble)
	end

	return RetValStr
end

--[[=============================================================================
    BCDToAscii(InByte)

    Description
    Convert an BCD string to an ascii string.

    Parameters
    InByte(string) - Binary coded decimal that is to be converted into ascii

    Returns
    The ascii string
===============================================================================]]
function BCDToAscii(InByte)
	return tostring(bit.rshift(InByte, 4)) .. tostring(bit.band(InByte, 0x0F))
end

--[[=============================================================================
    MakeXMLNode(Tag, Value)

    Description
    Create an Xml element

    Parameters
    Tag(string)   - The Xml elements name
    Value(string) - The Xml elements value

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLNode(Tag, Value)
	return "<" .. Tag .. ">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)

    Description
    Create an Xml element with an attribute

    Parameters
    Tag(string)       - The Xml elements name
    Value(string)     - The Xml elements value
    Attribute(string) - The attribute to be added to the Xml element
    AttrValue(string) - The value of the attribute to be added

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)
    return "<" .. Tag .. " " .. Attribute .. "=\"" .. AttrValue .. "\">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    StringFromUnicode(UnicodeString)

    Description
    Convert a unicode string

    Parameters
    UnicodeString(string) - The unicode string to be converted to ascii

    Returns
    The ascii representation of the unicode string
===============================================================================]]
function StringFromUnicode(UnicodeString)
	local RetVal = ""

	-- extract every other byte from the unicode string
	for Index = 2, #UnicodeString, 2 do
		RetVal = RetVal .. string.sub(UnicodeString, Index, Index)
	end

	return RetVal
end

--[[=============================================================================
    StringSplit(s)

    Description
    Splits a string into multiple strings at every space character

    Parameters
    s(string) - The string that is to be split into several strings at the
                space character

    Returns
    A table of strings containing all the seperate values in the given string
===============================================================================]]
function StringSplit(s)
	local delim = " "
	local result = {}

	if s == nil or s == "" then
		return result
	end

	for match in (s..delim):gmatch("(.-)"..delim) do
		table.insert(result, match)
	end

	return result
end

--[[=============================================================================
    toboolean(s)

    Description
    Returns a boolean representation of the given string

    Parameters
    s(string) - The string to evaluate the boolean representation from

    Returns
    The value true or false based on the given string
===============================================================================]]
function toboolean(val)
	rval = false;

	if type(val) == "string" and (string.lower(val) == "true" or val == "1") then
		rval = true
	elseif type(val) == "number" and val ~= 0 then
		rval =  true
	elseif type(val) == "boolean" and val == true then
		rval = true
	end

	return rval
end

--[[=============================================================================
    Go(to, err, ...)

    Description
    Call a function with the given arguments if it exists or report the error

    Parameters
    to(string)  - The string to evaluate the boolean representation from
    err(string) - The error to report if the function does not exist
    ...         - Additional optional parameters for the function specified by
                  the "to" parameter

    Returns
    Nothing
===============================================================================]]
function Go(to, err, ...)
	if (type(to) == "function") then
		return to(...)
	else
		LogTrace(err)
	end
end

--[[=============================================================================
    IsEmpty(str)

    Description
    Identifies if the string given is nil or empty

    Parameters
    str(string) - The string to evaluate for the empty condition

    Returns
    True if the given value is empty, false otherwise
===============================================================================]]
function IsEmpty(str)
	return str == nil or str == ""
end

--[[=============================================================================
    ReverseTable(a)

    Description
    Reverse table entries (key=value, value=key)

    Parameters
    a(table) - The table to reverse

    Returns
    new reversed table
===============================================================================]]
function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function tonumber_loc(str, base)
  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

--[[=============================================================================
    HexToString(InString)

    Description
    Converts a string of Hex characters to a readable string of ASCII characters

    Parameters
    InString(string) - The string to be converted

    Returns
    A string showing the hex bytes of the InString
===============================================================================]]
function HexToString(InString)
	local RetVal = ""

	for Index = 1, #InString do
		RetVal = RetVal .. string.format("%02X ", InString:byte(Index))
	end
	return RetVal
end


--[[=============================================================================
    StringToHex(InString)

    Description
    Converts a string of ASCII characters to as string with the actual Hex bytes in them.
	Basically an array of hex bytes.

    Parameters
    InString(string) - The string to be converted

    Returns
    A string of hex bytes (really an array of hex values) 
===============================================================================]]
function StringToHex(InString)
	local RetVal = ""

	for HexByteString in string.gfind(InString, "%x%x") do
		RetVal = RetVal .. string.char(tonumber(HexByteString, 16))
	end
	return RetVal
end

function RecordHistory(severity, eventType, category, subcategory, description)
	C4:RecordHistory(severity, eventType, category, subcategory, description)
end

function RecordCriticalHistory(eventType, category, subcategory, description)
	RecordHistory("Critical", eventType, category, subcategory, description)
end

function RecordWarningHistory(eventType, category, subcategory, description)
	RecordHistory("Warning", eventType, category, subcategory, description)
end

function RecordInfoHistory(eventType, category, subcategory, description)
	RecordHistory("Info", eventType, category, subcategory, description)
end


 end)
package.preload['common.c4_ir_connection'] = (function (...)
--[[=============================================================================
    Base for an IR connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_ir_connection = "2016.01.08"
end

IRConnectionBase = inheritsFrom(DeviceConnectionBase)

function IRConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function IRConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("tSerConBase:Initialize")
	gControlMethod = "IR"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function IRConnectionBase:ControlMethod()
	return "IR"
end

function IRConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendIR(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("IR connection is not bound. Command not sent.")
	end
end
 end)
package.preload['common.c4_serial_connection'] = (function (...)
--[[=============================================================================
    Base for a serial connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_serial_connection = "2016.01.08"
end

SerialConnectionBase = inheritsFrom(DeviceConnectionBase)

function SerialConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function SerialConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "Serial"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function SerialConnectionBase:ControlMethod()
	return "Serial"
end

function SerialConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendToSerial(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("Not connected to serial. Command not sent.")
	end
end

function SerialConnectionBase:SendRaw(sData)
	C4:SendToSerial(self._BindingID, sData)
end


function SerialConnectionBase:ReceivedFromSerial(idBinding, sData)
	self:ReceivedFromCom(sData)
end
 end)
package.preload['lib.c4_queue'] = (function (...)
--[[=============================================================================
    c4_queue Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_queue = "2016.01.08"
end

c4_queue = inheritsFrom(nil)

function c4_queue:construct()
	-- entry table
	self._et = {first = 0, last = -1}
	self._maxSize = 0	-- no size limit
	self._name = ""

	local mt = getmetatable(self)
	if (mt ~= nil) then
		mt.__tostring = self.__tostring
	end
end

function c4_queue:__tostring()
	local tOutputString = {}
	table.insert(tOutputString, "--- Queue ---")
	if (not IsEmpty(self._name)) then
		table.insert(tOutputString, "  name = " .. tostring(self._name))
	end
	table.insert(tOutputString, "  first = " .. tostring(self._et.first))
	table.insert(tOutputString, "  last = " .. tostring(self._et.last))
	table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
	table.insert(tOutputString, "  maximum size = " .. self._maxSize)
	table.insert(tOutputString, "  next value = " .. tostring(self:value()))
	return table.concat(tOutputString, "\n")
end


-- push a value on the queue
function c4_queue:push(value, ...)
	local numItems = self._et.last - self._et.first + 1

	if ((self._maxSize <= 0) or (numItems < self._maxSize) ) then
		local last = self._et.last + 1
		self._et.last = last
		
		local interval = select(1, ...)
		local units = select(2, ...)
		local command_name = select(3, ...)
		self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}		
		--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
	else
		-- if addToQueue == true then push value to queue
		if (self:OnMaxSizeReached()) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}
			--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		end
	end
end

function c4_queue:OnMaxSizeReached()
	--LogTrace ("Max Size Reached - clear queue and push value to the queue (default).")
	local addToQueue = true

	self:clear()
	return (addToQueue)
end

-- pop a value from the queue
function c4_queue:pop()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace("Queue:pop(), queue is empty")
		return ""
	end

	local value = self._et[first]
	self._et[first] = nil        -- to allow garbage collection
	self._et.first = first + 1
	--LogTrace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)

	return value
end

-- clear queue
function c4_queue:clear()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace ("Queue:clear(), queue is empty")
		return ""
	end

	self._et = {first = 0, last = -1}
	--LogTrace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
	--LogTrace (self._et)
end

-- return value of first item
function c4_queue:value()
	local first = self._et.first

	if (first > self._et.last) then
		return ""
	else
		return self._et[first]
	end
end

-- return queue's maximum size
function c4_queue:MaxSize()
	return self._maxSize
end

-- return queue's maximum size
function c4_queue:SetMaxSize(size)
	self._maxSize = size
end

function c4_queue:SetName(name)
	self._name = name
end

-- return the queue's current size
function c4_queue:Size()
	return self._et.last - self._et.first + 1
end

-- is queue empty?
function c4_queue:empty()
	-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
	if (self._et.first > self._et.last) then
		return true
	else
		return false
	end
end

--[[
    c4_queue unit tests
--]]
function __test_c4_queue()
	require "test.C4Virtual"
	require "lib.c4_log"

	local LOG = c4_log:new("test_c4_queue")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	-- create an instance of the queue
	local c4Queue = c4_queue:new()

	c4Queue:SetMaxSize(3)
	assert(c4Queue:MaxSize() == 3, "_maxSize is not equal to '3' it is: " .. c4Queue:MaxSize())

	c4Queue:push("Item #1 in Queue")
	c4Queue:push("Item #2 in Queue")
	c4Queue:push("Item #3 in Queue")
	c4Queue:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4Queue:Size() == 1, "queue size is not equal to '1' it is: " .. c4Queue:Size())

	print (c4Queue)

	-- Test inheritance overriding OnMaxSizeReached
	-- Create a new class
	c4_queue_new = inheritsFrom(c4_queue)

	-- override construct()
	function c4_queue_new:construct()
		self.superClass():construct()  -- call base class
		self._maxSizeOption = 1

		local mt = getmetatable(self)
		if (mt ~= nil) then
			mt.__tostring = self.__tostring
		end
	end

	-- override OnMaxSizeReached()
	function c4_queue_new:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,

		-- Option 1: Do Nothing, new item is not added to queue
		if (self._maxSizeOption == 1) then
			
			LogInfo("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)
		-- Option 2: pop value, and push new value on queue
		elseif(self._maxSizeOption == 2) then
			LogInfo("Max Size Reached - pop value, and push new value on queue (option 2)")
			self:pop()
			return (true)
		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(self._maxSizeOption == 3) then
			LogInfo("Max Size Reached - clear queue and DO NOT push new value onto queue")
			self:clear()
			return (false)
		end
	end

	-- create an instance of the new queue
	local c4QueueNew = c4_queue_new:new()
	c4QueueNew:SetMaxSize(3)
	c4QueueNew:push("Item #1 in Queue")
	c4QueueNew:push("Item #2 in Queue")
	c4QueueNew:push("Item #3 in Queue")
	c4QueueNew:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4QueueNew:Size() == 3, "queue size is not equal to '3' it is: " .. c4QueueNew:Size())

	print(c4QueueNew)
	print ("done...")
end end)
package.preload['common.c4_device_connection_base'] = (function (...)
--[[=============================================================================
    DeviceConnectionBase Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_common"
require "lib.c4_object"
require "lib.c4_log"
require "lib.c4_timer"
require "lib.c4_queue"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_device_connection_base = "2016.01.08"
end

COMMAND_QUEUE_SIZE = 100
DEFAULT_COMMAND_DELAY_INTERVAL = 100            -- Don't send consecutive commands faster than this many milliseconds
DEFAULT_COMMAND_RESPONSE_INTERVAL = 3           -- If we haven't received and ACK after this many seconds, try again
DEFAULT_RETRY_COUNT_MAX = 3

function ON_DRIVER_EARLY_INIT.c4_device_connection_base()
	gReceiveBuffer = ""
	gIsUrlConnected = false
end

DeviceConnectionBase = inheritsFrom(nil)

function DeviceConnectionBase:construct()

	self._IsConnected = false
	self._SendTimer = nil
	self._WaitResponseTimer = nil
	self._CommandQueue = nil
	self._Priority1CommandQueue = nil
	self._Priority2CommandQueue = nil
	self._LastCommand = nil
	self._ExpectAck = false
	self._CommandRetryCount = 0
	self._RetryCountMax = DEFAULT_RETRY_COUNT_MAX

	self._SendCommandDelayMS = DEFAULT_COMMAND_DELAY_INTERVAL
	self._CommandResponseWaitS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	-- Polling
	self._PollingInterval = 0
	self._PollingUnits = "SECONDS"
	self._PollingTimer = nil
end

function DeviceConnectionBase:Initialize(ExpectAck, CommandDelayInterval, CommandResponseInterval, CallbackParam)

	if (ExpectAck ~= nil) then
		self._ExpectAck = ExpectAck
	end

	if (CommandDelayInterval ~= nil) then
		self._SendCommandDelayMS = CommandDelayInterval
	end

	if (CommandResponseInterval ~= nil) then
		self._CommandResponseWaitS = CommandResponseInterval
	end

	self._CommandQueue = c4_queue:new()
	self._CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._CommandQueue:SetName("Command Queue")

	self._Priority1CommandQueue = c4_queue:new()
	self._Priority1CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority1CommandQueue:SetName("P1 Queue")

	self._Priority2CommandQueue = c4_queue:new()
	self._Priority2CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority2CommandQueue:SetName("P2 Queue")

	-- usually only one of these timers will be used, but it's pretty low overhead to instantiate both of them
	self._SendTimer = c4_timer:new("SendCommand", self._SendCommandDelayMS, "MILLISECONDS", DeviceConnectionBase.OnSendTimeExpired, false, CallbackParam)
	self._WaitResponseTimer = c4_timer:new("WaitResponse", self._CommandResponseWaitS, "SECONDS", DeviceConnectionBase.OnWaitTimeExpired, false, CallbackParam)
end

function DeviceConnectionBase:InitPolling(PollingInterval, PollingUnits, CallbackParam)
	LogFatal("DeviceConnectionBase:InitPolling()")
	if (PollingInterval ~= nil) then
		self._PollingInterval = PollingInterval
	end
	
	self._PollingUnits = PollingUnits or self._PollingUnits

	LogFatal("self._PollingInterval: %s, self._PollingUnits: %s", tostring(self._PollingInterval), tostring(self._PollingUnits))
	
	-- create polling timer
	self._PollingTimer = c4_timer:new("Polling", self._PollingInterval, self._PollingUnits, DeviceConnectionBase.OnPollingTimerExpired, false, CallbackParam)
end

function DeviceConnectionBase:StartPolling(interval, units)
	LogFatal("DeviceConnectionBase:StartPolling()")
	LogFatal("self._PollingTimer: %s", tostring(self._PollingTimer))
	
	if (self._PollingTimer ~= nil) then
		self._PollingTimer:KillTimer()
		
		local timer_units = units or self._PollingTimer._units
		local timer_interval = interval or self._PollingInterval

		self._PollingTimer:StartTimer(timer_interval, timer_units)
	end
end

function DeviceConnectionBase:StopPolling()
	LogFatal("DeviceConnectionBase:StopPolling()")
	self._PollingTimer:KillTimer()
end

function DeviceConnectionBase:SetExpectACK(ExpectACK)
	self._ExpectAck = ExpectACK
end

function DeviceConnectionBase:SetCommandDelayInterval(DelayInterval)
	self._SendCommandDelayMS = DelayInterval
end

function DeviceConnectionBase:SetResponseWaitInterval(WaitInterval)
	self._CommandResponseWaitS = WaitInterval
end

function DeviceConnectionBase:ReceivedFromCom(sData)

	gReceiveBuffer = gReceiveBuffer .. sData
	LogTrace("ReceivedFromCom  ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)

	message = self:GetMessage()
	while (message ~= nil and message ~= "") do
		status, err = pcall(HandleMessage, message)
		if (status) then
			message = self:GetMessage()
		else
			LogError("LUA_ERROR: " .. err)
			message = ""
			gReceiveBuffer = ""
		end
	end
end

function DeviceConnectionBase:SetConnection(IsConnected, method)
	self._IsConnected = IsConnected
	gControlMethod = method
end

function DeviceConnectionBase:ControlMethod()
	-- Override in derived class
	print("WARNING: Need to override ControlMethod - should never be called")
	
	return ""
end

function DeviceConnectionBase:StartCommandTimer(...)
	local value = select(1, ...)
	local units = select(2, ...)
	local command_name = select(3, ...) or ""

	self._WaitResponseTimer:KillTimer()
	self._SendTimer:KillTimer()

	if (self._ExpectAck) then
		-- expecting an ACK set the Response Wait timer
		local timer_units = units or self._WaitResponseTimer._units
		local timer_interval = value or self._CommandResponseWaitS

		self._WaitResponseTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting wait Timer:  %d", self._WaitResponseTimer._timerID) .. " for " .. command_name)
	else
		-- no ACK expected, just wait the designated amount of time and send another command
		local timer_units = units or self._SendTimer._units
		local timer_interval = value or self._SendCommandDelayMS

		self._SendTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting Send Timer:  %d for %s (timer_interval = %d, timer_units = %s)", self._SendTimer._timerID, command_name, timer_interval, timer_units))
	end
end

-- Note the '.' instead of the ':'
function DeviceConnectionBase.OnSendTimeExpired(Instance)
	LogTrace("Send Timer expired")
	Instance._SendTimer:KillTimer()

	local tCommand = Instance._LastCommand
	if (tCommand.command_name ~= nil) then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand.command_name))
	elseif (type(tCommand) == "string") then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand))
	end
	
	Instance._LastCommand = nil
	Instance:SendNextCommand()

	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase.OnWaitTimeExpired(Instance)
	LogTrace("Wait Timer expired")
	Instance._WaitResponseTimer:KillTimer()
	Instance._CommandRetryCount = Instance._CommandRetryCount + 1

	if (Instance._CommandRetryCount >= Instance._RetryCountMax) then
		-- To many retries, pop the current command and try the next one
		Instance._CommandRetryCount = 0
		Instance:SendNextCommand()
	else
		Instance:SendLastCommand()
	end
end

function DeviceConnectionBase.OnPollingTimerExpired(Instance)
	LogTrace("Polling Timer expired")
	Instance._PollingTimer:KillTimer()

	OnPollingTimerExpired()
	
	Instance._PollingTimer:StartTimer(Instance._PollingInterval)
end

function DeviceConnectionBase:HandleACK()
	self._LastCommand = nil
	
	self._WaitResponseTimer:KillTimer()
	self._CommandRetryCount = 0
	self:SendNextCommand()
end

function DeviceConnectionBase:QueueEmpty()
	return (self._CommandQueue:empty() and self._Priority1CommandQueue:empty() and self._Priority2CommandQueue:empty())
end

function DeviceConnectionBase:QueueCommand(sCommand, ...)
--	LogTrace("QueueCommand(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority1Command(sCommand, ...)
	LogTrace("QueuePriority1Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority1CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority2Command(sCommand, ...)
	LogTrace("QueuePriority2Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority2CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:SendNextCommand()
	LogTrace("DeviceConnectionBase:SendNextCommand")

	local tCommand = nil
	if (not self._Priority1CommandQueue:empty()) then
		tCommand = self._Priority1CommandQueue:pop()
		LogTrace(tostring(gCon._Priority1CommandQueue))
	elseif (not self._Priority2CommandQueue:empty()) then
		tCommand = self._Priority2CommandQueue:pop()
		LogTrace(tostring(gCon._Priority2CommandQueue))
	elseif (not self._CommandQueue:empty()) then
		tCommand = self._CommandQueue:pop()
		LogTrace(tostring(gCon._CommandQueue))
	end
	
	if (tCommand ~= nil) then
		self._LastCommand = tCommand
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendLastCommand()
--	LogTrace("DeviceConnectionBase:SendLastCommand")

	local tCommand = self._LastCommand
	if (tCommand ~= nil) then
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendCommand()
	-- Dummy routine.  Override in derived class
	print("Need to override SendCommand - should never be called")
end


function DeviceConnectionBase:GetMessage()
	
	-- Brain dead version of this routine. Just return the current receive buffer.
	-- It's very likely that a GetMessage() function will need to be created
	if (GetMessage ~= nil and type(GetMessage) == "function") then
		return GetMessage()
	else
		local ComMessage = gReceiveBuffer
		gReceiveBuffer = ""

		return ComMessage
	end
end

--[[=============================================================================
    Other Connection Functions
===============================================================================]]

function ReceivedFromSerial(idBinding, sData)
	if (gCon.ReceivedFromSerial == nil) then return end --serial is bound but not the current control method
	gCon:ReceivedFromSerial(idBinding, sData)
end

function ReceivedFromNetwork(idBinding, nPort, sData)
	gCon:ReceivedFromNetwork(idBinding, nPort, sData)
end

function OnServerDataIn(nHandle, strData)
--	LogTrace("Received Data on Handle: " .. nHandle .. ": " .. strData)
--	LogTrace("Data Is: %s", HexToString(strData))
	gCon:ReceivedFromNetworkServer(nHandle, strData)
end


--[[=============================================================================
    The ReceivedAsync function is called in response to 'url_get_request'. 
    The ticketId is the number returned from the request.
===============================================================================]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

--	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
--	LogTrace(tHeaders)

	gCon:ReceivedAsync(ticketId, strData, responseCode, tHeaders)
end
	
--[[=============================================================================
    OnBindingChanged(idBinding, class, bIsBound)
  
    Description:
    Function called by Director when a binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnBindingChanged(idBinding, class, bIsBound)
	
	LogTrace("OnBindingChanged(): idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	if (idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
		SetControlMethod()
		OnSerialConnectionChanged(idBinding, class, bIsBound)
	elseif (idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
		SetControlMethod()
		OnIRConnectionChanged(idBinding, class, bIsBound)
	end
end

--[[=============================================================================
    OnNetworkBindingChanged(idBinding, bIsBound)
  
    Description:
    Function called by Director when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	LogTrace('OnNetworkBindingChanged(): idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))

	gIsNetworkConnected = bIsBound
	SetControlMethod()
	OnNetworkConnectionChanged(idBinding, bIsBound)
	if (bIsBound) then
		-- Start a special instance of reconnect timer to eventually do NetConnect if not done automatically
		gCon._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval) 	
	end	
end

--[[=============================================================================
    OnConnectionStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. tostring(nPort) .. ")]: " .. sStatus)

	local isOnline = false

	gNetworkStatus = sStatus	
	if (sStatus == "ONLINE") then
		isOnline = true
	end

	gCon:SetOnlineStatus(isOnline)
	OnNetworkStatusChanged(idBinding, nPort, sStatus)
end

--[[=============================================================================
    SetControlMethod()
  
    Description:
    Sets the control method type for the drivers internal infrastructure
  
    Parameters:
    None
  
    Returns:
    The type of control method for the drivers connection(i.e. Network, Serial,
    IR, or (none))
===============================================================================]]
function SetControlMethod()
	if (gIsNetworkConnected == false) and (gCon._NetworkReconnectTimer ~= nil) then
		--housekeeping when changing from network control to serial or IR control
		gCon._NetworkReconnectTimer:KillTimer() 
	end


	if( gIsNetworkServerConnected) then
		-- connect to NetworkServer communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "NetworkServer") then
			gCon = NetworkServerConnectionBase:new()
			gCon:Initialize()
		end
		gCon:SetConnection(true, "NetworkServer")
	elseif (gIsNetworkConnected) then
		-- connect to Network communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Network") then
			gCon = NetworkConnectionBase:new(NETWORK_BINDING_ID, NETWORK_PORT)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "Network")
	elseif (gIsUrlConnected) then
		-- connect to URL communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "URL") then
			gCon = UrlConnectionBase:new()
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "URL")
	elseif (gIsSerialConnected) then
		-- connect to Serial communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Serial") then
			gCon = SerialConnectionBase:new(SERIAL_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
			gCon:InitPolling(tonumber(gPollingTimerInterval), "MINUTES", gCon)
		end
		gCon:SetConnection(true, "Serial")
	elseif (gIsIRConnected) then
		-- connect to IR communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "IR") then
			gCon = IRConnectionBase:new(IR_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "IR")
	else
		if (gCon ~= nil) then
			gCon:SetConnection(false, "(none)")
		end
		-- gCon = nil
	end

	gCon._CommandQueue:clear()
	gCon._Priority1CommandQueue:clear()
	gCon._Priority2CommandQueue:clear()
end

--[[=============================================================================
    ValidateControlMethod(controlMethod)
  
    Description:
    Identifies whether the specified control method has a valid connection
  
    Parameters:
    controlMethod(string) - The communication we are validating against
                            Valid types are (Network, Serial, and IR)
  
    Returns:
    true if the controlMethod specified has been connected, false otherwise.
===============================================================================]]
function ValidateControlMethod(controlMethod)
	local isValid = false

	if (controlMethod == "Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (controlMethod == "URL") and (gIsUrlConnected) then
		isValid = true
	elseif (controlMethod == "Serial") and (gIsSerialConnected) then
		isValid = true
	elseif (controlMethod == "IR") and (gIsIRConnected) then
		isValid = true
	end

	return isValid
end
 end)
package.preload['common.c4_network_connection'] = (function (...)
--[[=============================================================================
    Base for a network connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"
require "lib.c4_log"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_network_connection = "2016.01.08"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30
DEFAULT_RECONNECT_INTERVAL_SECONDS = 5

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS
gNetworkReconnectInterval = DEFAULT_RECONNECT_INTERVAL_SECONDS

NetworkConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkConnectionBase:construct(BindingID, Port)
	self.superClass():construct()

	self._BindingID = BindingID
	self._Port = Port
	self._LastCheckin = 0
	self._IsOnline = false
	self._KeepAliveTimer = nil
end

function NetworkConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetConBase:Initialize")
	gControlMethod = "Network"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	self._KeepAliveTimer = c4_timer:new("PollingTimer", gNetworkKeepAliveInterval, "SECONDS", NetworkConnectionBase.OnKeepAliveTimerExpired, false, self)
	self._NetworkReconnectTimer = c4_timer:new("NetworkReconnectTimer", gNetworkReconnectInterval, "SECONDS", NetworkConnectionBase.OnNetworkReconnectTimerExpired, false, self)
end

function NetworkConnectionBase:ControlMethod()
	return "Network"
end

function NetworkConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)

			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end

function NetworkConnectionBase:ReceivedFromNetwork(idBinding, nPort, sData)
	self._LastCheckin = 0
	self:ReceivedFromCom(sData)
end

function NetworkConnectionBase:CheckNetworkConnectionStatus()
	if (self._IsConnected and (not self._IsOnline)) then
		LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(self._BindingID, self._Port)
		--C4:NetConnect(self._BindingID, self._Port)
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	end
end

function NetworkConnectionBase.OnKeepAliveTimerExpired(Instance)
	Instance._LastCheckin = Instance._LastCheckin + 1

	if(Instance._LastCheckin == 3) then
		LogWarn("Failed to receive poll responses... initiating network recovery mode...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		return
	elseif(Instance._LastCheckin > 4) then	
		Instance._LastCheckin = 4
	end

	if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function" and Instance._IsOnline) then
		SendKeepAlivePollingCommand()
	end

	Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
end

function NetworkConnectionBase.OnNetworkReconnectTimerExpired(Instance)
	if (Instance._IsConnected) then
		LogWarn("OnNetworkReconnectTimerExpired: Attempting to reactivate network connection...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		C4:NetConnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	else
		LogWarn("Cannot attempt to reactivate, the network connection is not bound")
	end
end

function NetworkConnectionBase:SetOnlineStatus(IsOnline)
	self._IsOnline = IsOnline

	if(IsOnline) then
		self._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
		self._NetworkReconnectTimer:KillTimer()
		self._LastCheckin = 0
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "true")
		end

		self:SendNextCommand()
	else
		self._KeepAliveTimer:KillTimer()
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "false")
		end
	end
end

function ON_DRIVER_LATEINIT.c4_network_connection()
	-- Ensure existing connection is taken into consideration (useful on Driver Update)
	if (gCon ~= nil) then
		local tmp = C4:GetBindingAddress(gCon._BindingID)
		if (tmp ~= nil and string.len(tmp) > 0) then OnNetworkBindingChanged(gCon._BindingID, true) end
	end
end
 end)
package.preload['common.c4_notify'] = (function (...)
--[[=============================================================================
    Notification Functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_notify = "2016.01.08"
end

--[[=============================================================================
    SendNotify(notifyText, tParams, bindingID)

    Description
    Forwards a notification to the proxy with a list of parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    tParams(table)     - Table of key value pairs that hold the the parameters
                         and their values used in the proxy function
    bindingID(int)     - The requests binding id

    Returns
    Nothing
===============================================================================]]
function SendNotify(notifyText, tParams, bindingID)
	C4:SendToProxy(bindingID, notifyText, tParams, "NOTIFY")
end

--[[=============================================================================
    SendSimpleNotify(notifyText, ...)

    Description
    Forwards a notification to the proxy with no parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    bindingID(int)     - Optional parameter containing the requests binding id,
                         if not specified then the DEFAULT_PROXY_ID is given.

    Returns
    Nothing
===============================================================================]]
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end end)
package.preload['common.c4_command'] = (function (...)
--[[=============================================================================
    Functions for handling and executing commands and actions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_command = "2016.01.08"
end

--[[=============================================================================
    ExecuteCommand(sCommand, tParams)

    Description
    Function called by Director when a command is received for this DriverWorks
    driver. This includes commands created in Composer programming.

    Parameters
    sCommand(string) - Command to be sent
    tParams(table)   - Lua table of parameters for the sent command

    Returns
    Nothing
===============================================================================]]
function ExecuteCommand(sCommand, tParams)
	LogTrace("ExecuteCommand(" .. sCommand .. ")")
	LogInfo(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")
	local status, ret

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		status, ret = pcall(EX_CMD[sCommand], tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		status, ret = pcall(EX_CMD[trimmedCommand], tParams)
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
		status = true
	else
		LogInfo("ExecuteCommand: Unhandled command = " .. sCommand)
		status = true
	end
	
	if (not status) then
		LogError("LUA_ERROR: " .. ret)
	end
	
	return ret -- Return whatever the function returns because it might be xml, a return code, and so on
end

--[[=============================================================================
    EX_CMD.LUA_ACTION(tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    tParams(table) - Lua table of parameters for the command option

    Returns
    Nothing
===============================================================================]]
function EX_CMD.LUA_ACTION(tParams)
	if (tParams ~= nil) then
		for cmd, cmdv in pairs(tParams) do
			if (cmd == "ACTION" and cmdv ~= nil) then
				local status, err = pcall(LUA_ACTION[cmdv], tParams)
				if (not status) then
					LogError("LUA_ERROR: " .. err)
				end
				break
			end
		end
	end
end

--[[=============================================================================
    ReceivedFromProxy(idBinding, sCommand, tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    idBinding(int)   - Binding ID of the proxy that sent a BindMessage to the
                       DriverWorks driver.
    sCommand(string) - Command that was sent
    tParams(table)   - Lua table of received command parameters

    Returns
    Nothing
===============================================================================]]
function ReceivedFromProxy(idBinding, sCommand, tParams)

	if (sCommand ~= nil) then

		-- initial table variable if nil
		if (tParams == nil) then
			tParams = {}
		end
		
		LogTrace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function PRX_CMD." .. sCommand .. "()")
		LogInfo(tParams)
	
	    --support for universal miniapps
	     if (idBinding == SWITCHER_PROXY and sCommand == 'PASSTHROUGH') then
		     idBinding = PASSTHROUGH_PROXY
		     sCommand = tParams.PASSTHROUGH_COMMAND
	     end		

		if ((PRX_CMD[sCommand]) ~= nil) then
			local status, err = pcall(PRX_CMD[sCommand], idBinding, tParams)
			if (not status) then
				LogError("LUA_ERROR: " .. err)
			end
		else
			LogInfo("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

--[[
	This function is called when a UI (Navigator) requests data, and
	calls the function requested.
--]]
function UIRequest(sRequest, tParams)
	local ret = ""

	if (sRequest ~= nil) then
		tParams = tParams or {}   -- initial table variable if nil
		LogTrace("UIRequest(): " .. sRequest .. "; Call Function UI_REQ." .. sRequest .. "()")
		LogInfo(tParams)

		if (UI_REQ[sRequest]) ~= nil then
			ret = UI_REQ[sRequest](tParams)
		else
			LogWarn("UIRequest: Unhandled request = " .. sRequest)
		end
	end

	return ret
end
 end)
package.preload['common.c4_property'] = (function (...)
--[[=============================================================================
    Function for changing properties

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_property = "2016.01.08"
end

--[[=============================================================================
    OnPropertyChanged(sProperty)

    Description
    Function called by Director when a property changes value. The value of the
    property that has changed can be found with: Properties[sName]. Note that
    OnPropertyChanged is not called when the Property has been changed by the
    driver calling the UpdateProperty command, only when the Property is changed
    by the user from the Properties Page. This function is called by Director
    when a property changes value.

    Parameters
    sProperty(string) - Name of property that has changed.

    Returns
    Nothing
===============================================================================]]
function OnPropertyChanged(sProperty)
	local propertyValue = Properties[sProperty]

	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	end

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	local status = true
	local err = ""

	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[sProperty], propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[trimmedProperty], propertyValue)
	end

	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)
	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end
 end)
package.preload['common.c4_init'] = (function (...)
--[[=============================================================================
    Initial driver initialization and destruction functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_property"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_init = "2016.01.08"
end

--[[=============================================================================
    OnDriverInit()

    Description
    Invoked by director when a driver is loaded. This API is provided for the
    driver developer to contain all of the driver objects that will require
    initialization.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverInit()
	gInitializingDriver = true
	C4:ErrorLog("INIT_CODE: OnDriverInit()")

	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_EARLY_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property
	-- global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		C4:ErrorLog("INIT_CODE: Calling OnPropertyChanged - " .. k .. ": " .. v)
		local status, err = pcall(OnPropertyChanged, k)
		if (not status) then
			C4:ErrorLog("LUA_ERROR: " .. err)
		end
	end

	gInitializingDriver = false
end

--[[=============================================================================
    OnDriverLateInit()

    Description
    Invoked by director after all drivers in the project have been loaded. This
    API is provided for the driver developer to contain all of the driver
    objects that will require initialization after all drivers in the project
    have been loaded.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[=============================================================================
    OnDriverDestroyed()
    Function called by Director when a driver is removed. Release things this
    driver has allocated such as timers.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	
	-- Call all ON_DRIVER_DESTROYED functions
	for k, v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end end)
package.preload['lib.c4_timer'] = (function (...)
--[[=============================================================================
    c4_timer Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_timer = "2016.01.08"
end

c4_timer = inheritsFrom(nil)

function c4_timer:construct(name, interval, units, Callback, repeating, CallbackParam)
	self._name = name
	self._timerID = TimerLibGetNextTimerID()
	self._interval = interval
	self._units = units
	self._repeating = repeating or false
	self._Callback = Callback
	self._CallbackParam = CallbackParam or ""
	self._id = 0

	gTimerLibTimers[self._timerID] = self
	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("Created timer " .. self._name)
	end
end

function c4_timer:StartTimer(...)
	c4_timer:KillTimer()

	-- optional parameters (interval, units, repeating)
	if ... then
		local interval = select(1, ...)
		local units = select(2, ...)
		local repeating = select(3, ...)

		self._interval = interval or self._interval
		self._units = units or self._units
		self._repeating = repeating or self._repeating
	end

	if (tonumber(self._interval) > 0) then
		if (LOG ~= nil and type(LOG) == "table") then
			LogTrace("Starting Timer: " .. self._name)
		end

		self._id = C4:AddTimer(self._interval, self._units, self._repeating)
	end
end

function c4_timer:KillTimer()
	if (self._id) then
		self._id = C4:KillTimer(self._id)
	end
end

function c4_timer:TimerStarted()
	return (self._id ~= 0)
end

function c4_timer:TimerStopped()
	return (self._id == 0)
end

function c4_timer:GetTimerInterval()
	return (self._interval)
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.c4_timer()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.c4_timer()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[=============================================================================
    OnTimerExpired(idTimer)

    Description:
    Function called by Director when the specified Control4 timer expires.

    Parameters:
    idTimer(string) - Timer ID of expired timer.
===============================================================================]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._CallbackParam)
			end
		end
	end
end

--[[=============================================================================
    CreateTimer(name, interval, units, callback, repeating, callbackParam)

    Description: 
    Creates a named timer with the given attributes

    Parameters:
    name(string)       - The name of the timer being created
    interval(int)      - The amount of the given time between calls to the
                         timers callback function
    units(string)      - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    callback(string)   - The function to call when the timer expires
    repeating(bool)    - Parameter indicating whether the timer should be
                         called repeatedly until cancelled
    callbackParam(...) - Parameters to be passed to the callback function

    Returns:
    A handle to the timer
===============================================================================]]
function CreateTimer(name, interval, units, callback, repeating, callbackParam)
	timer = c4_timer:new(name, interval, units, callback, repeating, callbackParam)
	return timer
end

--[[=============================================================================
    StartTimer(handle, ...)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer)   - Handle to a created timer object
    interval(int)   - The amount of the given time between calls to the
                      timers callback function
    units(string)   - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    repeating(bool) - Parameter indicating whether the timer should be
                      called repeatedly until cancelled

    Returns:
    None
===============================================================================]]
function StartTimer(handle, ...)
	handle:StartTimer(...)
end

--[[=============================================================================
    KillTimer(handle)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    None
===============================================================================]]
function KillTimer(handle)
	handle:KillTimer()
end

--[[=============================================================================
    TimerStarted(handle)

    Description: 
    Identifies whether a timer has been started or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been started, or false otherwise
===============================================================================]]
function TimerStarted(handle)
	return handle:TimerStarted()
end

--[[=============================================================================
    TimerStopped(handle)

    Description: 
    Identifies whether a timer has been stopped or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been stopped, or false otherwise
===============================================================================]]
function TimerStopped(handle)
	return handle:TimerStopped()
end

--[[=============================================================================
    GetTimerInterval(handle)

    Description: 
    Gets the interval setting of the given timer

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns the interval setting of the given timer
===============================================================================]]
function GetTimerInterval(handle)
	return handle:GetTimerInterval()
end

--[[=============================================================================
    c4_timer Unit Tests
===============================================================================]]
function __test_c4_timer()
	require "test.C4Virtual"
	require "lib.c4_log"
	require "common.c4_init"

	OnDriverInit()

	local LOG = c4_log:new("test_c4_timer")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	function OnTestTimerExpired()
		c4Timer:KillTimer()
	end

	-- create an instance of the timer
	c4Timer = c4_timer:new("Test", 45, "MINUTES", OnTestTimerExpired)

	assert(c4Timer._id == 0, "_id is not equal to '0' it is: " .. c4Timer._id)
	c4Timer:StartTimer()
	assert(c4Timer._id == 10001, "_id is not equal to '10001' it is: " .. c4Timer._id)
	assert(c4Timer:TimerStarted() == true, "TimerStarted is not equal to true it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == false, "TimerStopped is not equal to false it is: " .. tostring(c4Timer:TimerStopped()))
	OnTimerExpired(c4Timer._id)
	assert(c4Timer:TimerStarted() == false, "TimerStarted is not equal to false it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == true, "TimerStopped is not equal to true it is: " .. tostring(c4Timer:TimerStopped()))
end end)
package.preload['lib.c4_object'] = (function (...)
--[[=============================================================================
    c4_object Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_object = "2016.01.08"
end

function inheritsFrom( baseClass )
	local new_class = {}
	local class_mt = { __index = new_class }

	function new_class:create(...)
		local newinst = {}

		setmetatable( newinst, class_mt )

		-- Call the constructor when we create this class
		if newinst.construct then
			-- Allow returning a different obj than self. This allows for readonly tables etc...
			newinst = newinst:construct(...) or newinst
		end

		return newinst
	end

	if nil ~= baseClass then
		setmetatable( new_class, { __index = baseClass } )
	end

	--[[=============================================================================
	    Implementation of additional OO properties starts here 
	===============================================================================]]
	
	-- Return the class object of the instance
	function new_class:class()
		return new_class
	end

	--[[=============================================================================
	    Return the super class object of the instance. 

	    Note Calling methods on the base class itself will modify
	    the base table's static properties. In order to have call
	    the base class' methods and have them modify the current object
	    use super() or superAsSelf(). 
	===============================================================================]]
	function new_class:superClass()
		return baseClass
	end

	--[[=============================================================================
	    Returns a table that allows calling of the base class's method
	    while maintaining the objects state as the modified state of the base
	    class' methods. For example consider the following statements (order matters):

	    -- The child sees the parents property if the child hasn't overriden the property
	       obj:superClass().id = "parent"
	       obj.id == "parent" -- true

	    -- Setting the property on the child overrides (hides) the parents property
	       obj.id = "child"
	       obj.id == "child" -- true

	    -- The super() method pass
	       obj:super().id == "parent" -- true
	       obj:super().id = "child"
	       obj:super().id == "parent" -- still true
	       obj.id == "child" -- still true
	===============================================================================]]
	function new_class:super()
		local holder = {}

		holder.child = self
		holder.parent = baseClass

		local mt = {}
		mt.__index = function(table, index)
			if table.parent[index] then
				return table.parent[index]
			else 
				return table.child[index]
			end
		end

		-- Only set the new values to the child.
		mt.__newindex = function(table, key, value)
			table.child[key] = value
		end

		mt.__tostring = function(table)
			return tostring(table.child)
		end

		setmetatable(holder, mt)
		return holder
	end

	new_class.new = new_class.create

	--[[=============================================================================
	    Return true if the caller is an instance of theClass
	===============================================================================]]
	function new_class:isa( theClass )
		local b_isa = false
		local cur_class = new_class

		while ( nil ~= cur_class ) and ( false == b_isa ) do
			if cur_class == theClass then
				b_isa = true
			else
				cur_class = cur_class:superClass()
			end
		end

		return b_isa
	end

	return new_class
end

--[[=============================================================================
    Inheritance unit tests
===============================================================================]]
function __test_inheritance()
	local b = inheritsFrom(nil)

	b.construct = function(self, msg)
		self._msg = msg
	end

	local t = inheritsFrom(b)
	t.construct = function(self, msg)
		self:super():construct(msg)
	end

	t1 = t:new("t1")
	t2 = t:new("t2")
	assert(t1._msg == "t1", "t1 message is not equal to 't1' it''s: " .. t1._msg)
	assert(t2._msg == "t2", "t2 message is not equal to 't2' it''s: " .. t2._msg)    
	assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");
	assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

	t1:superClass().id = "parent"
	assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

	-- Setting the property on the child overrides (hides) the parents property
	t1.id = "child"
	assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

	-- The super() method maintains the self pointer to the child and not to the base
	assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
	t1:super().id = "child1"
	assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
	assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")
end end)
package.preload['lib.c4_log'] = (function (...)
--[[=============================================================================
    c4_log Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_log = "2016.01.08"
end

c4_log = inheritsFrom(nil)

function c4_log:construct(logName)
	self._logLevel = tonumber(string.sub(Properties['Log Level'] or "", 1, 1)) or 5
	self._outputPrint = Properties['Log Mode']:find("Print") ~= nil
	self._outputC4Log = Properties['Log Mode']:find("Log") ~= nil
	self._logName = logName or ""

	-- make sure Property is up to date (no harm if absent)
	C4:UpdateProperty("Log Level", Properties['Log Level'])
end

function c4_log:SetLogLevel(level)
	self._logLevel = tonumber(string.sub(level or "", 1, 1)) or self._logLevel
end

function c4_log:LogLevel()
	return self._logLevel
end

function c4_log:OutputPrint(value)
	self._outputPrint = value
end

function c4_log:OutputC4Log(value)
	self._outputC4Log = value
end

function c4_log:SetLogName(logName)

	if (logName == nil or logName == "") then
		logName = ""
	else
		logName = logName .. ": "
	end

	self._logName = logName
end

function c4_log:LogName()
	return self._logName
end

function c4_log:Enabled()
	return (self._outputPrint or self._outputC4Log)
end

function c4_log:PrintEnabled()
	return self._outputPrint
end

function c4_log:C4LogEnabled()
	return self._outputC4Log
end

function c4_log:CreateTableText(tValue, tableText)
	tableText = tableText or ""

	if (type(tValue) == "table") then

		tableText = tableText .. "{"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ","
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableText(v, tableText)
				tableText = tableText .. ","
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ","
			end
		end

		tableText = tableText .. "}"
	end

	return tableText
end

function InsertIndent(indentLevel)
	local indentStr = ""

	for i=1, indentLevel do
		indentStr = indentStr .. "\t"
	end

	return indentStr
end

function c4_log:CreateTableTextFormatted(tValue, tableText, indentLevel)
	tableText = tableText or ""
	indentLevel = indentLevel or 0

	if (type(tValue) == "table") then

		indentLevel = indentLevel + 1
		tableText = tableText .. "{\n"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. InsertIndent(indentLevel) .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. InsertIndent(indentLevel) .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ",\n"
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',\n"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableTextFormatted(v, tableText, indentLevel)
				tableText = tableText .. ",\n"
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ",\n"
			end
		end

		indentLevel = indentLevel - 1
		tableText = tableText .. InsertIndent(indentLevel) .. "}"
	end

	return tableText
end

MAX_TABLE_LEVELS = 10
function c4_log:PrintTable(tValue, tableText, sIndent, level)
	tableText = tableText or ""
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				if (tableText == "") then
					tableText = sIndent .. tostring(k) .. ":  " .. tostring(v)
					if (sIndent == ".   ") then sIndent = "    " end
				else
					tableText = tableText .. "\n" .. sIndent .. tostring(k) .. ":  " .. tostring(v)
				end
				if (type(v) == "table") then
					tableText = self:PrintTable(v, tableText, sIndent .. "   ", level)
				end
			end
		else
			tableText = tableText .. "\n" .. sIndent .. tostring(tValue)
		end
	end
	
	return tableText
end

function c4_log:LogTable(tValue, sIndent, level)
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				C4:ErrorLog(self._logName .. sIndent .. tostring(k) .. ":  " .. tostring(v))
				if (type(v) == "table") then
					self:LogTable(v, sIndent .. "   ", level)
				end
			end
		else
			C4:ErrorLog(self._logName .. sIndent .. tValue)
		end
	end
end

function c4_log:Print(logLevel, sLogText)

	if (self._logLevel >= logLevel) then
		if (type(sLogText) == "table") then
			if (self._outputPrint) then
				print (self:PrintTable(sLogText, tableText, ".   ", 0))
			end

			if (self._outputC4Log) then
				self:LogTable(sLogText, "   ", 0)
			end

			return
		end

		if (self._outputPrint) then
			print (sLogText)
		end

		if (self._outputC4Log) then
			C4:ErrorLog(self._logName .. tostring(sLogText))
		end
	end
end

function c4_log:Fatal(sLogText, ...)
	self:LogOutput(0, sLogText, ...)
end

function c4_log:Error(sLogText, ...)
	self:LogOutput(1, sLogText, ...)
end

function c4_log:Warn(sLogText, ...)
	self:LogOutput(2, sLogText, ...)
end

function c4_log:Info(sLogText, ...)
	self:LogOutput(3, sLogText, ...)
end

function c4_log:Debug(sLogText, ...)
	self:LogOutput(4, sLogText, ...)
end

function c4_log:Trace(sLogText, ...)
	self:LogOutput(5, sLogText, ...)
end

function c4_log:LogOutput(level, sLogText, ...)
	if (LogEnabled()) then
		if (type(sLogText) == "string") then
			sLogText = string.format(sLogText, ...)
		end

		self:Print(level, sLogText)
	end
end

--[[=============================================================================
    c4_log wrapper functions
===============================================================================]]
function TryLog(level, sLogText, ...)
	LOG:LogOutput(level, sLogText, ...)
end

--[[=============================================================================
    SetLogLevel(level)

    Description: 
    Sets the desired log level to view

    Parameters:
    level(int) - The logging level to set the message to
                 0 = Fatal
                 1 = Error
                 2 = Warn
                 3 = Info
                 4 = Debug
                 5 = Trace

    Returns:
    None
===============================================================================]]
function SetLogLevel(level)
	LOG:SetLogLevel(level)
end

--[[=============================================================================
    LogLevel()

    Description: 
    Returns the currently set log level

    Parameters:
    None

    Returns:
    The current log level
        0 = Fatal
        1 = Error
        2 = Warn
        3 = Info
        4 = Debug
        5 = Trace
===============================================================================]]
function LogLevel()
	return LOG:LogLevel()
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputPrint(value)
	LOG:OutputPrint(value)
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages to file or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputC4Log(value)
	LOG:OutputC4Log(value)
end

--[[=============================================================================
    SetLogName(logName)

    Description: 
    Sets the name of the log file where the messages will be written to

    Parameters:
    logName(string) - Sets the name of the log to write messages to

    Returns:
    None
===============================================================================]]
function SetLogName(logName)
	LOG:SetLogName(logName)
end

--[[=============================================================================
    LogName(logName)

    Description: 
    Gets the name of the log file where the messages will be written to

    Parameters:
    None

    Returns:
    The value of the log file that has been set
===============================================================================]]
function LogName()
	return LOG:LogName()
end

--[[=============================================================================
    LogEnabled()

    Description: 
    Identifies if logging or print has been enabled

    Parameters:
    None

    Returns:
    true if either logging or print has been enabled, false otherwise
===============================================================================]]
function LogEnabled()
	return LOG:Enabled()
end

--[[=============================================================================
    PrintEnabled()

    Description: 
    Gets the state of print output

    Parameters:
    None

    Returns:
    true if print has been enabled, false otherwise
===============================================================================]]
function PrintEnabled()
	return LOG:PrintEnabled()
end

--[[=============================================================================
    C4LogEnabled()

    Description: 
    Gets the state of logging

    Parameters:
    None

    Returns:
    true if logging has been enabled, false otherwise
===============================================================================]]
function C4LogEnabled()
	return LOG:C4LogEnabled()
end

--[[=============================================================================
    LogFatal(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Fatal(0) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogFatal(sLogText, ...)
	local status, err = pcall(TryLog, 0, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogFatal failed: " .. err)
	end
end

--[[=============================================================================
    LogError(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Error(1) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogError(sLogText, ...)
	local status, err = pcall(TryLog, 1, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogError failed: " .. err)
	end
end

--[[=============================================================================
    LogWarn(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Warn(2) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogWarn(sLogText, ...)
	local status, err = pcall(TryLog, 2, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogWarn failed: " .. err)
	end
end

--[[=============================================================================
    LogInfo(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Info(3) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogInfo(sLogText, ...)
	local status, err = pcall(TryLog, 3, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogInfo failed: " .. err)
	end
end

--[[=============================================================================
    LogDebug(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Debug(4) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogDebug(sLogText, ...)
	local status, err = pcall(TryLog, 4, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogDebug failed: " .. err)
	end
end

--[[=============================================================================
    LogTrace(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Trace(5) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogTrace(sLogText, ...)
	local status, err = pcall(TryLog, 5, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogTrace failed: " .. err)
	end
end

function dbgPrint(buf)
	if (LOG:PrintEnabled()) then
		print (buf)
	end
end

function dbgHexdump(buf)
	hexdump(buf, dbgPrint)
end

--[[=============================================================================
    c4_log unit tests
===============================================================================]]
function __test_c4_log()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	assert(LOG:LogName() == "test_c4_log", "_logName is not equal to 'test_c4_log' it is: " .. LOG:LogName())

	-- Test setting log level
	LOG:SetLogLevel("2 - Warning")
	assert(LOG:LogLevel() == 2, "_logLevel is not equal to '2' it is: " .. LOG:LogLevel())

	LOG:SetLogLevel(3)
	assert(LOG:LogLevel() == 3, "_logLevel is not equal to '3' it is: " .. LOG:LogLevel())

	-- Test enabling logs
	LOG:OutputPrint(false)
	assert(LOG:PrintEnabled() == false, "_outputPrint is not equal to 'false' it is: " .. tostring(LOG:PrintEnabled()))

	LOG:OutputC4Log(true)
	assert(LOG:C4LogEnabled() == true, "_outputC4Log is not equal to 'true' it is: " .. tostring(LOG:C4LogEnabled()))

	LOG:SetLogLevel(4)
	LogTrace("***** This is a test *****")
end

function __test_CreatTableText()
	local tTest = {}

	tTest[1] = {}
	tTest[2] = {}
	tTest[3] = 30
	tTest[4] = "Forty"
	
	LogTrace("----- tText -----")
	LogTrace(tTest)

	local tTest2 = { One = {},
					 Two = {},
					 Three = 30,
					 Four = "Forty" }
	LogTrace("----- tText2 -----")
	LogTrace(tTest2)
	
	local tTest3 = { [1] = {},
					 [2] = {},
					 [3] = 30,
					 [4] = "Forty" }
	LogTrace("----- tText3 -----")
	LogTrace(tTest3)

	local tTest4 = { [1] = {},
					 Two = {},
					 [3] = 30,
					 [4] = "Forty",
					 Five = "Fifty" }
	LogTrace("----- tText4 -----")
	LogTrace(tTest4)

	local tableText = LOG:CreateTableText(tTest4)
	LogTrace("----- tableText -----")
	LogTrace(tableText)
	
	--local tNew = {[1] = {},[3] = 30,[4] = 'Forty',Five = 'Fifty',Two = {},}
	--LogTrace(tNew)
end

function __TestCreateTableTextFormatted()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	local tButtons = {
		Name = 'heat',
		Attributes = {},
		ChildNodes = {
			[1] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '51',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Pool Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'POOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[2] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '53',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Spa Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'SPAHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[3] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '54',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {Value = 'Pool Solar Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {}
					},
					[3] = {
						Value = 'SOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				}
			}
		}
	}

	print(LOG:CreateTableTextFormatted(tButtons))
end end)
package.preload['common.c4_common'] = (function (...)
--[[=============================================================================
    ON_INIT, Timer,s and Property management functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_log"
require "lib.c4_timer"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_common = "2016.01.08"
end

--[[=============================================================================
    Create and Initialize Logging
===============================================================================]]
function ON_DRIVER_EARLY_INIT.c4_common()
	-- Create a logger
	LOG = c4_log:new("Template_c4z Change Name")
end

function ON_DRIVER_INIT.c4_common()
	-- Create Log Timer
	gC4LogTimer = c4_timer:new("Log Timer", 45, "MINUTES", OnLogTimerExpired)
end

--[[=============================================================================
    Log timer callback function
===============================================================================]]
function OnLogTimerExpired()
	LogWarn("Turning Log Mode Off (timer expired)")
	gC4LogTimer:KillTimer()
	
	C4:UpdateProperty("Log Mode", "Off")
	OnPropertyChanged("Log Mode")
end

gForceLogging = false
function ON_PROPERTY_CHANGED.LogMode(propertyValue)
	gC4LogTimer:KillTimer()
	
	if (gForceLogging) then
		LOG:OutputPrint(true)
		LOG:OutputC4Log(true)
	else
		LOG:OutputPrint(propertyValue:find("Print") ~= nil)
		LOG:OutputC4Log(propertyValue:find("Log") ~= nil)
		if (propertyValue == "Off") then
			return
		end
		
		gC4LogTimer:StartTimer()
	end
end

function ON_PROPERTY_CHANGED.LogLevel(propertyValue)
	if (gForceLogging) then
		LOG:SetLogLevel("5 - Debug")
	else
		LOG:SetLogLevel(propertyValue)
	end
end

--[[=============================================================================
    Print Template Versions
===============================================================================]]
function TemplateVersion()
	print ("\nTemplate Versions")
	print ("-----------------------")
	for k, v in pairs(TEMPLATE_VERSION) do
		print (k .. " = " .. v)
	end
	
	print ("")
end end)
package.preload['common.c4_driver_declarations'] = (function (...)
--[[=============================================================================
    Driver Declarations used to call startup routines, teardown routines, and 
    other basic functions of the drivers operation

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Template Version Table
TEMPLATE_VERSION = {}
TEMPLATE_VERSION.c4_driver_declarations = "2016.01.08"

-- Command Handler Tables
EX_CMD = {}
PRX_CMD = {}
UI_REQ = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}

--[[=============================================================================
    Tables of functions
    The following tables are function containers that are called within the
    following functions:

    OnDriverInit()
        First calls all functions contained within ON_DRIVER_EARLY_INIT table
        then calls all functions contained within ON_DRIVER_INIT table

    OnDriverLateInit()
        Calls all functions contained within ON_DRIVER_LATEINIT table

    OnDriverDestroyed()
        Calls all functions contained within ON_DRIVER_DESTROYED table

    OnPropertyChanged()
        Calls all functions contained within ON_PROPERTY_CHANGED table
===============================================================================]]
ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001 end)
----------------------------------------------------------------------------------------------
-- Template for Receiver-Tuner Driver
----------------------------------------------------------------------------------------------
require "common.c4_driver_declarations"
require "common.c4_common"
require "common.c4_init"
require "common.c4_property"
require "common.c4_command"
require "common.c4_notify"
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_utils"
require "lib.c4_timer"
require "connections"
require "actions"
require "device_specific_commands"
require "device_messages"
require "receiver_init_shared"
require "receiver_init_conditional"
require "properties"
require "proxy_commands"
require "receiver.receiver_proxy_class"
require "receiver.receiver_proxy_commands"
require "receiver.receiver_proxy_notifies"
require "minidriver"


-- This macro is utilized to identify the version string of the driver template version used.
TEMPLATE_VERSION.driver = "2014.08.14"

-- constants
COM_USE_ACK = false
COM_COMMAND_DELAY_MILLISECONDS = 250
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4


DEFAULT_PROXY_BINDINGID = 5001

RECEIVER_PROXY_BINDINGID = 5001
TUNER_PROXY_BINDINGID = 5002
NET_PROXY_BINDINGID = 5003
USBFRONT_PROXY_BINDINGID = 5016
USBREAR_PROXY_BINDINGID = 5017
BLUETOOTH_PROXY_BINDINGID = 5018


--Values must equal mod 1000 values of Output Connections
ZONE_1 = 0
ZONE_MAIN = ZONE_1
ZONE_2 = 1
ZONE_3 = 2
ZONE_4 = 3

--[[
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    Initialization Code
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--]]

function ON_DRIVER_EARLY_INIT.receiver_driver()
    --hide properties not needed by installer
    C4:SetPropertyAttribs ("Max Volume", 1)

end

function ON_DRIVER_INIT.receiver_driver()

    -- TODO: Modify tVolumeRamping to have on enrty per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {
		[0] = {state = false,mode = "",},
		[1] = {state = false,mode = "",},
		[2] = {state = false,mode = "",},	
		[3] = {state = false,mode = "",},
	}
	
    -- Create an instance of the ReceiverProxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local bProcessesDeviceMessages = true
    local bUsePulseCommandsForVolumeRamping = false
    if (Properties["Ramping Method"] == "Pulse Commands") then bUsePulseCommandsForVolumeRamping = true end
    gReceiverProxy = ReceiverProxy:new(RECEIVER_PROXY_BINDINGID, TUNER_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping, bUsePulseCommandsForVolumeRamping)
    
    --[[
	   Device Volume Range
	   TODO: edit "MIN_DEVICE_LEVEL" & "MAX_DEVICE_LEVEL" values based upon the protocol specification for volume range.
	   If zones have volume ranges that vary, convert these constants into a table indexed by the mod value of the output connection 
	   and then update all references to handle the table structure.
    --]]
    MIN_DEVICE_LEVEL = 0
    MAX_DEVICE_LEVEL = tonumber(Properties["Max Volume"]) --TX-L50, DSX-3, HT-L05, HT-D03 have volume range 0-50; TX-NR535, TX-NR636, HT-RC660 & DTR-30.6 have volume range 0-80, all others 0-100	
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    
    --Net Tune Preset Command
    gNetTunePresetCommand = {
	   ["prefix"] = "", 
	   ["preset"] = nil, 
    }
    gNetTunePresetCommandTimer = c4_timer:new("NetTunePresetCommand", 3, "SECONDS", OnNetTunePresetCommandExpired, false)
    gLastSelectedNetService = {
	   [0] = -1,
	   [1] = -1,
	   [2] = -1,
	   [3] = -1,
    }
    
end

function getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    local steps = tonumber(Properties["Volume Ramp Steps"])
    local slope = tonumber(Properties["Volume Ramp Slope"])
    local tCurve = gReceiverProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    
    return tCurve 
end

function ON_DRIVER_LATEINIT.receiver_driver()
    DRIVER_NAME = C4:GetDriverConfigInfo("name")
    LOG:SetLogName(DRIVER_NAME)
    
    gProxyNameToBindingIdMap = {}
    buildProxyNameToBindingIdMap()
    
    --v114 bug fix...
    fixMaxVolumeProperty()
    
end

function GetAirplayVersion()
    tAirplay2Models = {
	   "DRX-2.3",
	   "DRX-3.3",
	   "DRX-4.3",
	   "DRX-5.3",
	   "TXNR-595",
	   "TXNR-696",
	   "TXNR-797",
	   "TX-RZ740",
	   "TX-RZ840",
	   "VSX-534",
	   "VSX-934",
	   "VSX-834",
	   "VSX-LX104",
	   "VSX-LX304",
	   "VSX-LX504",
	   "SC-LX701",
	   "SC-LX801",
	   "SC-LX901",
	   "VSX-LX103",
	   "VSX-LX303",
	   "VSX-LX503",
	   "DRC-R1",
	   "DRX-3.2",
	   "DRX-4.2",
	   "DRX-5.2",
	   "DRX-7",
	   "DRX-R1",
	   "DSX-3",
	   "DTM-7",
	   "PR-RZ5100",
	   "TX-8270",
	   "TX-NR585",
	   "TX-NR686",
	   "TX-NR787",
	   "TX-RZ1100",
	   "TX-RZ3100",
	   "TX-RZ630",
	   "TX-RZ730",
	   "TX-RZ830",
	   "TX-RZ920",
	   "SX-S30",
	   "VSX-933",
	   "VSX-S520",
    }
    
    local version = "Airplay"
    local model = C4:GetDriverConfigInfo("model")
    for j,k in pairs(tAirplay2Models) do
	   if (k == model) then
		  version = "Airplay2"
		  return version
	   end
    end 
    
    return version
    
end

function fixMaxVolumeProperty()
    --v114 Bug fix, changed max volume property from 100 to 200
    tAffectedModels = {
	   "DRC-R1.1",
	   "DRX-4.2",
	   "DRX-5.2",
	   "DRX-7.1",
	   "DRX-R1.1",
	   "TX-NR787",
	   "TX-RZ730",
	   "TX-RZ830",
    }
    local model = C4:GetDriverConfigInfo("model")
    for j,k in pairs(tAffectedModels) do
	   if (k == model) then
		  C4:UpdateProperty("Max Volume", "200")
		  MAX_DEVICE_LEVEL = 200
	   end
    end 
end

--[[
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    Driver Code
    --=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--]]

function buildProxyNameToBindingIdMap()
  local xml = C4:GetDriverConfigInfo("proxies")
  xml = "<proxies>" .. xml .. "</proxies>"
  local t = C4:ParseXml(xml)

  for j,k in pairs(t.ChildNodes) do 
    print(k.Attributes.name, k.Attributes.proxybindingid)
    gProxyNameToBindingIdMap[k.Attributes.name] = k.Attributes.proxybindingid
  end
end

function set_proxy_visbility(idBinding, value)
    local bHide = false
    if (value == "HIDE") then
	   bHide = true
	   SetProxyVisibilityInAllRooms(idBinding, bHide, "Hiding")
    elseif (value == "SHOW") then
	   bHide = false
	   SetProxyVisibilityInAllRooms(idBinding, bHide, "Showing")
    else
	   --do nothing
    end
end

function SetProxyVisibilityInAllRooms(idBinding, bHide, sShowHide)
  idBinding = idBinding or 0
  if (idBinding == 0) then return end -- silently fail if no binding passed in.

  -- Get Bound Proxy's Device ID / Name.
  local id, name = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), idBinding))

  LogTrace(sShowHide .. " '" .. name .. "' in all rooms...")

  -- Send hide command to all rooms, for 'ALL' Navigator groups.
  for roomid, roomname in pairs(C4:GetDevicesByC4iName("roomdevice.c4i") or {}) do
    LogTrace(sShowHide .. " '" .. name .. "' in '" .. roomname .. "' room.")
    C4:SendToDevice(roomid, "SET_DEVICE_HIDDEN_STATE", {PROXY_GROUP = "ALL", DEVICE_ID = id, IS_HIDDEN = bHide})
  end 
end

function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LOG:Warn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
    local cmd
    if (gControlMethod == "Network") then
	   cmd = "!1" .. command .. "\r"
	   local sHeader = tohex('49 53 43 50 00 00 00 10 00 00 00' .. string.format('%02x',string.len(cmd)) .. '01 00 00 00')
	   cmd = sHeader .. cmd	
    elseif (gControlMethod == "Serial") then
	   cmd = "!1" .. command .. "\r"
    elseif (gControlMethod == "IR") then
	   cmd = command
    else
	   LOG:Warn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
	   return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end